/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.genome;

import java.io.IOException;
import java.io.PrintStream;
import java.util.List;

import edu.uconn.ngstools.sequences.Sequence;
import edu.uconn.ngstools.sequences.filehandlers.FastaSequencesHandler;

/**
 * Direct implementation of a Genome for an assembly like the ones published in UCSC 
 * @author Jorge Duitama
 */
public class GenomeAssembly implements Genome { 
	private FastaSequencesHandler sequencesHandler = new FastaSequencesHandler();
	/**
	 * Creates a new GenomeAssembly with the given data
	 * @param filename Name of the fasta file with the sequences making up the assembly 
	 * @throws IOException If the file can not be read
	 */
	public GenomeAssembly (String filename) throws IOException {
		sequencesHandler.loadSequences(filename);
	}
	@Override
	public char getReferenceBase(String sequenceName, int absolutePosition) {
		Sequence seq = sequencesHandler.getSequence(sequenceName);
//		System.out.println("getReferenceBase  pos - 1, seq length"+(absolutePosition-1)+" "+seq.length());
		if (absolutePosition > seq.length()) {
			System.out.println("out of range "+(absolutePosition-1));
			return 0;
		}
		if(seq!=null) {
			return seq.charAt(absolutePosition-1);
		}
		return 0;
	}

	@Override
	public List<String> getSequenceNames() {
		return sequencesHandler.getIds();
	}

	@Override
	public void saveGenome(PrintStream out, int lineLength) {
		sequencesHandler.saveSequences(out, lineLength);
	}

	@Override
	public void setReferenceBase(String sequenceName, int absolutePosition, char base) { 
		Sequence sequence = sequencesHandler.getSequence(sequenceName);
		int pos = absolutePosition-1;
		if(sequence!=null && pos >=0 && pos < sequence.length()) {
			assert sequence.isInAlphabet(base);
			char current = sequence.charAt(pos);
			if (Character.toUpperCase(current) != Character.toUpperCase(base)) {
				if(Character.isUpperCase(current)) {
					sequence.setCharAt(pos, Character.toUpperCase(base));
				} else {
					sequence.setCharAt(pos, Character.toLowerCase(base));
				}
			}
		}
	}
	/**
	 * Finds the sequence with the given name
	 * @param sequenceName Name to look for
	 * @return Sequence Object with the sequence information
	 */
	public Sequence getSequence(String sequenceName) {
		return sequencesHandler.getSequence(sequenceName);
	}

}
