package edu.uconn.ngstools.genome.contigs;

import edu.uconn.ngstools.genome.GenomicRegion;
import edu.uconn.ngstools.sequences.Sequence;

/**
 * Information for a contig mapped to a reference genome
 * @author Jorge Duitama
 *
 */
public class Contig implements GenomicRegion {
	private String id;
	private String sequenceName;
	private int start;
	private int end;
	private Sequence sequence;
	private String fastaId;
	/**
	 * Creates a new Contig with the given information
	 * @param id Id of the contig
	 * @param sequenceName Name of the sequence where the Contig is mapped
	 * @param start First position of the Contig in the sequence
	 * @param end Last position of the Contig in the sequence
	 */
	public Contig(String id, String sequenceName, int start, int end) {
		super();
		this.id = id;
		this.sequenceName = sequenceName;
		this.start = start;
		this.end = end;
	}
	/**
	 * Returns the id of the contig
	 * @return String Id of the contig
	 */
	public String getId() {
		return id;
	}
	/**
	 * Changes the id of the Contig
	 * @param id New Id
	 */
	public void setId(String id) {
		this.id = id;
	}
	/**
	 * @return String Name of the sequence where the Contig is mapped
	 */
	public String getSequenceName() {
		return sequenceName;
	}
	/**
	 * Changes the name of the sequence where the Contig is mapped
	 * @param sequenceName New Sequence name
	 */
	public void setSequenceName(String sequenceName) {
		this.sequenceName = sequenceName;
	}
	/**
	 * @return int First position of this Contig in the sequence
	 */
	public int getStart() {
		return start;
	}
	/**
	 * Changes the first position of the Contig in the sequence
	 * @param start New first position of the Contig in the sequence
	 */
	public void setStart(int start) {
		this.start = start;
	}
	/**
	 * @return int Last position of this Contig in the sequence
	 */
	public int getEnd() {
		return end;
	}
	/**
	 * Changes the last position of the Contig in the sequence
	 * @param end New last position of the Contig in the sequence
	 */
	public void setEnd(int end) {
		this.end = end;
	}
	/**
	 * Transforms coordinates relative to this contig to coordinates relative to the sequence where this contig is mapped
	 * @param relativePosition Zero based position relative to the start of this contig
	 * @return int Position in the sequence where this contig is mapped
	 */
	public int getAbsolutePosition(int relativePosition) {
		return this.getStart()+relativePosition;
	}
	/**
	 * Transforms coordinates relative to the sequence where this contig is mapped to coordinates relative to this contig 
	 * @param absolutePosition Position in the sequence where this contig is mapped
	 * @return int Zero based position relative to the start of this contig 
	 */
	public int getRelativePosition(int absolutePosition) {
		return absolutePosition-this.getStart();
	}
	/**
	 * @return Sequence Sequence of this contig
	 */
	public Sequence getSequence() {
		return sequence;
	}
	/**
	 * Changes the sequence for this contig
	 * @param sequence New sequence for this contig. It is assumed that the sequence 
	 * length is equal than the contig length
	 */
	public void setSequence(Sequence sequence) {
		this.sequence = sequence;
	}
	/**
	 * @return String Id in the fasta file
	 */
	public String getFastaId() {
		return fastaId;
	}
	/**
	 * Changes the id in the fasta file
	 * @param fastaId New id in the fasta file
	 */
	public void setFastaId(String fastaId) {
		this.fastaId = fastaId;
	}
	/**
	 * Calculates the subsequence of this contig located between the two coordinates relative to the
	 * sequence where this contig is mapped
	 * @param absoluteFirst First base of the subsequence to return
	 * @param absoluteLast Last base of the sequence to return
	 * @return String Subsequence of this contig between the given positions. Null if the interval does not make sense
	 * or if the sequence of this contig is unknown
	 */
	public String getSequence(int absoluteFirst, int absoluteLast) {
		if(absoluteFirst>absoluteLast) {
			return null;
		}
		int posStart = getRelativePosition(absoluteFirst);
		int posEnd = getRelativePosition(absoluteLast)+1;
		if(sequence!=null && posStart>=0 && posEnd>=0 && posStart < sequence.length() && posEnd <= sequence.length()) {
			return sequence.substring(posStart, posEnd);
		}
		return null;
	}
	/**
	 * Changes the base in the given position
	 * @param absolutePosition Position to change relative to the start of the sequence
	 * where this contig is mapped 
	 * @param base New base to set. If the contig sequence is masked, masking will be preserved
	 */
	public void setReferenceBase(int absolutePosition, char base) {
		int pos = getRelativePosition(absolutePosition);
		if(sequence!=null && pos >=0 && pos < sequence.length()) {
			char current = sequence.charAt(pos);
			if (Character.toUpperCase(current) != Character.toUpperCase(base)) {
				if(Character.isUpperCase(current)) {
					sequence.setCharAt(pos, Character.toUpperCase(base));
				} else {
					sequence.setCharAt(pos, Character.toLowerCase(base));
				}
			}
			
		}
	}
	/**
	 * Gets the base at the specified coordinate
	 * @param absolutePosition Position to look relative to the start of the sequence
	 * @return char base at the given position
	 */
	public char getReferenceBase(int absolutePosition) {
		int pos = getRelativePosition(absolutePosition);
		if(sequence!=null && pos >=0 && pos < sequence.length()) {
			return sequence.charAt(pos);
		}
		return 0;
	}
	
	/**
	 * @return int length of the Contig calculated as end - start + 1
	 */
	@Override
	public int length() {
		return end - start + 1;
	}
}
