package edu.uconn.ngstools.genome.contigs;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import edu.uconn.ngstools.genome.Genome;
import edu.uconn.ngstools.genome.GenomicRegionPositionComparator;
import edu.uconn.ngstools.readalignments.AlignmentsTranslator;
import edu.uconn.ngstools.readalignments.ReadAlignment;
import edu.uconn.ngstools.sequences.filehandlers.FastaSequencesHandler;


public class ContigsMapHandler implements Genome, AlignmentsTranslator {
	private Map<String,List<Contig>> contigs=new TreeMap<String,List<Contig>>();
	//List of sequence names in the same order as they appear in the contigs file
	private List<String> sequenceNames = new ArrayList<String>();
	private FastaSequencesHandler sequencesHandler = new FastaSequencesHandler();
	public ContigsMapHandler (String mapFilename) throws IOException {
		loadMap(mapFilename);
	}
	public void loadMap(String mapFilename) throws IOException {
		FileInputStream fis = new FileInputStream(mapFilename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		String seqName = null;
		while(line!=null) {
			String [] items = line.split(" |\t");
			if(!items[0].equalsIgnoreCase(seqName)) {
				seqName = items[0];
				sequenceNames.add(seqName);
			}
			Contig contig = new Contig(items[1],seqName,Integer.parseInt(items[2]),Integer.parseInt(items[3]));
			List<Contig> seqContigs = contigs.get(contig.getSequenceName());
			if(seqContigs==null) {
				seqContigs = new ArrayList<Contig>();
				contigs.put(contig.getSequenceName(), seqContigs);
			}
			seqContigs.add(contig);
			line=in.readLine();
		}
		fis.close();
		for(List<Contig> seqContigs:contigs.values()) {
			Collections.sort(seqContigs,GenomicRegionPositionComparator.getInstance());
		}
	}
	public void loadSequences (String referenceFilename) throws IOException {
		sequencesHandler.loadSequences(referenceFilename);
		for(List<Contig> seqContigs:contigs.values()) {
			for(Contig c: seqContigs) {
				int pos = sequencesHandler.getPosition(c.getId());
				c.setFastaId(sequencesHandler.getId(pos));
				c.setSequence(sequencesHandler.getSequence(pos));
			}
		}
	}
	
	@Override
	public void translateAlignment (ReadAlignment aln) {
		Contig c = getContig(aln.getSequenceName());
		if(c!=null) {
			aln.setSequenceName(c.getSequenceName());
			aln.setStart(c.getAbsolutePosition(aln.getStart()));
		}
		if(aln.getFlags()%2==1) {
			if (ReadAlignment.DEFAULT_SEQ_NAME.equals(aln.getMateSequenceName())) {
				c=null;
			} else if(!ReadAlignment.EQUAL_SEQ_NAME.equals(aln.getMateSequenceName())) {
				c = getContig(aln.getMateSequenceName());
			} 
			if(c!=null) {
				aln.setMateSequenceName(c.getSequenceName());
				aln.setMatePosition(c.getAbsolutePosition(aln.getMatePosition()));
			} else {
				aln.setMateSequenceName(ReadAlignment.DEFAULT_SEQ_NAME);
				aln.setMatePosition(0);
				aln.setInsertSize(0);
			}
		}
	}
	private Contig getContig(String fastaId) {
		for(List<Contig> seqContigs:contigs.values()) {
			for(Contig contig:seqContigs) {
				if(fastaId.equals(contig.getFastaId())) {
					return contig;
				}
			}
		}
		return null;
	}
	public Contig getContig(String sequenceName, int absolutePosition) {
		List <Contig> seqContigs = contigs.get(sequenceName);
		if(seqContigs == null) {
			return null;
		}
		for(Contig contig:seqContigs) {
			if(contig.getSequenceName().equalsIgnoreCase(sequenceName) && contig.getStart() <=absolutePosition && contig.getEnd()>=absolutePosition) {
				return contig;
			}
		}
		return null;
	}
	@Override
	public List<String> getSequenceNames() {
		return sequenceNames;
	}
	@Override
	public void setReferenceBase(String sequenceName, int absolutePosition, char base) {
		Contig c = getContig(sequenceName, absolutePosition);
		if(c!=null) {
			c.setReferenceBase(absolutePosition, base);
		} else System.err.println("Contig not found for snp: "+sequenceName+":"+absolutePosition);
	}
	@Override
	public char getReferenceBase(String sequenceName, int absolutePosition) {
		Contig c = getContig(sequenceName, absolutePosition);
		if(c!=null) {
			return c.getReferenceBase(absolutePosition);
		}
		return 0;
	}
	@Override
	public void saveGenome(PrintStream out, int lineLength) {
		sequencesHandler.saveSequences(out, lineLength);
	}
}