/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.main;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;

import edu.uconn.ngstools.genome.GenomicRegionPositionComparator;
import edu.uconn.ngstools.readalignments.ReadAlignment;
import edu.uconn.ngstools.readalignments.filehandlers.SAMReadAlignmentsHandler;

/**
 * Program to clip a given number of bases from the 5' end and from the 3' end of 
 * each alignment in a SAM file
 * @author Jorge Duitama
 */
public class AlignedReadsClipper {

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		String alignmentsFile = args[0];
		int clipping5P = Integer.parseInt(args[1]);
		int clipping3P = Integer.parseInt(args[2]);
		PrintStream out = System.out;
		Deque<ReadAlignment> alnsBuffer = new LinkedList<ReadAlignment>();
		SAMReadAlignmentsHandler samHandler = new SAMReadAlignmentsHandler();
		FileInputStream fis = new FileInputStream(alignmentsFile);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String lastSeqName = "";
		String line=in.readLine();
		while(line!=null) {
			if(line.startsWith("@")) {
				out.println(line);
			} else {
				ReadAlignment aln = samHandler.createReadAlignment(line);
				if(aln!=null) {
					if(!aln.getSequenceName().equals(lastSeqName) ) {
						saveBuffer(alnsBuffer,out,0,samHandler);
						lastSeqName = aln.getSequenceName();
					}
					int alnStart = aln.getStart();
					int clippingStart = clipping5P;
					int clippingEnd = clipping3P;
					if((aln.getFlags()/16)%2==1) {
						clippingStart = clipping3P;
						clippingEnd = clipping5P;
					}
					aln.clipAlignment(clippingStart, clippingEnd);
					alnsBuffer.add(aln);
					if(alnsBuffer.size()>=10000) {
						saveBuffer(alnsBuffer,out,alnStart,samHandler);
					}
				}
			}
			line=in.readLine();
		}
		saveBuffer(alnsBuffer,out,0,samHandler);
		fis.close();
	}

	private static void saveBuffer(Deque<ReadAlignment> alnsBuffer, PrintStream out, int alnStart,SAMReadAlignmentsHandler samHandler) throws IOException {
		List<ReadAlignment> toSave = new ArrayList<ReadAlignment>(alnsBuffer.size());
		List<ReadAlignment> toKeep = new ArrayList<ReadAlignment>();
		while(alnsBuffer.size()>0) {
			ReadAlignment aln = alnsBuffer.removeFirst();
			if(alnStart == 0 || aln.getStart()<alnStart) {
				toSave.add(aln);
			} else {
				toKeep.add(aln);
			}
		}
		//System.err.println("SORTING: "+toSave.size());
		Collections.sort(toSave,GenomicRegionPositionComparator.getInstance());
		Collections.sort(toKeep,GenomicRegionPositionComparator.getInstance());
		//System.err.println("PRINTING: "+toSave.size());
		samHandler.saveReadAlignments(toSave, out);
		for(ReadAlignment aln:toKeep) {
			alnsBuffer.add(aln);
		}
	}
}
