/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.main;

import edu.uconn.ngstools.genome.GenomicRegion;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import edu.uconn.ngstools.genome.GenomicRegionComparator;
import edu.uconn.ngstools.genome.GenomicRegionPositionComparator;
import edu.uconn.ngstools.readalignments.CommonSpanningAlignmentComparator;
import edu.uconn.ngstools.readalignments.ReadAlignment;
import edu.uconn.ngstools.readalignments.ReadIdAlignmentComparator;
import edu.uconn.ngstools.readalignments.filehandlers.SAMReadAlignmentsHandler;
import edu.uconn.ngstools.transcriptome.CCDSTranscriptome;

/**
 * Program that merge read alignments against a reference assembly with alignments 
 * against the CCDS transcripts library. By default keeps reads that do not align multiple times
 * against either reference system. If a read maps uniquely against both references, the alignment
 * is kept just if the two alignments span a common region.
 * @author Jorge Duitama
 *
 */
public class CCDSReferenceAlnsMerger {
	public static final int MAX_ALIGNMENTS=50;
	public static final String MERGE_TYPE_SOFT="SoftMerge";
	public static final String MERGE_TYPE_HARD="HardMerge";
	private String mergeType = MERGE_TYPE_HARD;
	private ReadIdAlignmentComparator comparator = new ReadIdAlignmentComparator();
	private CommonSpanningAlignmentComparator alnComparator = new CommonSpanningAlignmentComparator();
	private GenomicRegionPositionComparator trascAlnComparator = new GenomicRegionPositionComparator();
	private static PrintStream outputFileStream;
        private static int alnListMaxSize = 1000000;
        private static int mergedAlnCount;
        private static int transcriptomeAlignmentCount = 0;
        private static int genomeAlignmentCount = 0;
        private static int transcriptomeAlignedReadsCount = 0;
        private static int genomeAlignedReadsCount = 0;
        private static boolean multiple_local_aln_filtered = false;



        //sahar read from transcriptome file instead of receiving transcriptome alignments as a list
	/**
	 * Merges the given alignments with the alignments stored in the file with the given name 
	 * @param alns List of alignments to merge
	 * @param filename Name of a SAM file with alignments of the same reads as the ones in alns to be merged
	 * @throws IOException If the file can not be read
	 */
//	public List<ReadAlignment>
        public void mergeReadAlignments (String transcriptomeAlignedSam, String genomeAlignedSam, SAMReadAlignmentsHandler alnsHandlerTranscriptome) throws IOException {
		SAMReadAlignmentsHandler alnsHandlerGenome=new SAMReadAlignmentsHandler();
		//Sort the alignments by read name
//		List<ReadAlignment> sortedAlns = new ArrayList<ReadAlignment>(alns);
//		Collections.sort(sortedAlns,comparator);
		//List to store new alignments to be kept
		List <ReadAlignment> toAdd = new ArrayList<ReadAlignment>();
		//List to store alignments that need to be removed. 
		//It is more efficient to keep a list and then remove all elements at the end that remove element by element 
//		List <ReadAlignment> toRemove = new ArrayList<ReadAlignment>();
		FileInputStream gfis = new FileInputStream(genomeAlignedSam);
		FileInputStream tfis = new FileInputStream(transcriptomeAlignedSam);
		BufferedReader gAlnReader = new BufferedReader(new InputStreamReader(gfis));
		BufferedReader tAlnReader = new BufferedReader(new InputStreamReader(tfis));
		String line=gAlnReader.readLine();
		ReadAlignment galn2 = null;
		boolean gmultiple = false;
                mergedAlnCount = 0;
		//Process file to execute merging rules
		for(int i=0;line!=null;i++) {
			ReadAlignment galn = alnsHandlerGenome.createReadAlignment(line);
			if(galn!=null) {
                                genomeAlignmentCount++;
				if(galn2!=null && galn.getRead().getId().equals(galn2.getRead().getId())) {
					//Read with multiple alignments in the file
                                    //for alignments with soft clipping, a preprocessing step is done before hardmerge
                                    // to filter multiple alignments of the read based on which portion of the
                                    // read is aligned (from the cigar field)
                                        if (!multiple_local_aln_filtered)
                                            gmultiple = true;
				} else if(galn2!=null) {
					//Execute rules to merge
					hardMerge(tAlnReader, galn2, gmultiple, toAdd, alnsHandlerTranscriptome);//, toRemove);
					gmultiple=false;
                                        genomeAlignedReadsCount++;
				} 
				galn2 = galn;
			}
			if(genomeAlignmentCount%200000==0) System.out.println(genomeAlignmentCount+" Genome alignments parsed");
			line=gAlnReader.readLine();
		}
		if(galn2!=null)
                    hardMerge(tAlnReader, galn2, gmultiple, toAdd, alnsHandlerTranscriptome);//, toRemove);
                 // call one last time to process any transcriptome alignments with readsID > the last read ID in the genome SAM
                hardMerge(tAlnReader, null, false, toAdd, alnsHandlerTranscriptome);//, toRemove);
		//Remove from what was is suppose to be removed from the sorted list and remaining multiple alns
//		Collections.sort(toRemove,comparator);
		//List to store initial alignments that survive after merging
//		List<ReadAlignment> survivingAlns = new ArrayList<ReadAlignment>();
//		for(int i=0;i<sortedAlns.size();i++) {
//			ReadAlignment aln = sortedAlns.get(i);
//			boolean multiple = isMultiple(sortedAlns, i);
//			if (!multiple && Collections.binarySearch(toRemove, aln ,comparator)<0) {
//				survivingAlns.add(aln);
//			}
//		}
//		alns.clear();
//		alns.addAll(survivingAlns);
//		alns.addAll(toAdd);
		gfis.close();
                tfis.close();
                mergedAlnCount += toAdd.size();
		alnsHandlerTranscriptome.saveReadAlignments(toAdd, outputFileStream);
                System.out.println(mergedAlnCount+" merged alignments written to output file");
                //return toAdd;
	}
	/**
	 * Executes merging rules for a single read. Modifies the toAdd and toRemove lists according
	 * with the merging rules
	 * @param sortedAlns List of alignments sorted by read id
	 * @param aln2 Alignment to the second set
	 * @param multiple2 true if there were multiple alignments in the second set
	 * @param toAdd List of alignments to add to the final set
	 * @param toRemove List of alignments to remove from the final set
	 */
	private void hardMerge(BufferedReader transcriptomeSamReader, ReadAlignment genomeAln, boolean gmultiple, List<ReadAlignment> toAdd,SAMReadAlignmentsHandler alnsHandlerTranscriptome) throws IOException {//, List<ReadAlignment> toRemove) {


                transcriptomeSamReader.mark(500);
                String line=transcriptomeSamReader.readLine();
                ReadAlignment transcriptomeAln = alnsHandlerTranscriptome.createReadAlignment(line);
                List<ReadAlignment> trascriptomeReadUniqueAlns;

                // read until a valid SAM entry is found (non commented; read aligned)
                while (line != null && transcriptomeAln == null) {
                    transcriptomeSamReader.mark(500);
                    line = transcriptomeSamReader.readLine();
                    transcriptomeAln = alnsHandlerTranscriptome.createReadAlignment(line);
               }

                 // loop while input genome alignment read ID is lex. greater than the last read 
                // transcriptome alignment read ID or input genome alignment is null (all genAln have been processed)
                while (line!= null && (genomeAln == null || (transcriptomeAln.getRead().getId().compareTo(genomeAln.getRead().getId()) < 0))) {
                    transcriptomeAlignedReadsCount++;
                    transcriptomeAlignmentCount++;
                     if (!isMultiple(transcriptomeAln,transcriptomeSamReader,line.length(),alnsHandlerTranscriptome))
                         AddAlignment(transcriptomeAln,toAdd,alnsHandlerTranscriptome);
                         //toAdd.add(transcriptomeAln);
//                     trascriptomeReadUniqueAlns = getReadGenomeUniqueTranscriptomeAlns(transcriptomeAln,transcriptomeSamReader,line.length(),alnsHandlerTranscriptome);
 //                    toAdd.addAll(trascriptomeReadUniqueAlns);  // rule : transcriptome unique; genome: not mapped
                     line=transcriptomeSamReader.readLine();
                     transcriptomeAln = alnsHandlerTranscriptome.createReadAlignment(line);
                     while (line != null && transcriptomeAln == null) {
                        transcriptomeSamReader.mark(500);
                         line=transcriptomeSamReader.readLine();
                         transcriptomeAln = alnsHandlerTranscriptome.createReadAlignment(line);
                    }
                 }


                if ((transcriptomeAln != null) && (transcriptomeAln.getRead().getId().compareTo(genomeAln.getRead().getId()) == 0)) {
                    transcriptomeAlignedReadsCount++;
                    transcriptomeAlignmentCount++;
                    boolean tmultiple = isMultiple(transcriptomeAln,transcriptomeSamReader,line.length(),alnsHandlerTranscriptome);
                    if (!tmultiple && !gmultiple && (alnComparator.compare(transcriptomeAln, genomeAln)==0)) {
                        AddAlignment(transcriptomeAln,toAdd,alnsHandlerTranscriptome); // rule: genome unique; transcriptome unique; agree
                        //toAdd.add(transcriptomeAln);
//                        toAdd.add(genomeAln);
                    }
                }
                else {// input genomeAln read ID has no transcriptome alignment
                    if (!gmultiple && genomeAln != null) {
                        // rule: genome unique; transcriptome unique; agree// rule: genome unique; transcriptome unique; agree
                        AddAlignment(genomeAln,toAdd,alnsHandlerTranscriptome);
                    }
                    if (transcriptomeAln != null) {  // last read transcriptome aln has ID > that of genomeAln; go back one line
                        transcriptomeSamReader.reset();
                       //transcriptomeAlignmentCount--;
                    }
                } //else
                        //toAdd.add(genomeAln);  // rule: genome unique; transcriptome: not mapped
	}

	/**
	 * Reads all alignments in input SAM file reader that have the same ID as first Aln 
         * and returns those that are unique w.r.t. their genomic location
         * Note: Pointer to Reader is set to the first alignment with different read ID than firstAln
	 * @param firstAln: First alignment in the SAM file for the read
	 * @param trascriptomeSAMReader:  Buffered reader to a SAM file sorted by ID.
         * @param  readAheadLimitEstimate : estimate length of minimum line size in the SAM file
	 */
        private List<ReadAlignment> getReadGenomeUniqueTranscriptomeAlns(ReadAlignment firstAln, BufferedReader trascriptomeSAMReader, int readAheadLimitEstimate, SAMReadAlignmentsHandler alnsHandler) throws IOException
// Method not used anymore
        {
           List<ReadAlignment> alnList = new ArrayList<ReadAlignment>();
           List<ReadAlignment> uniqueList = new ArrayList<ReadAlignment>();
           List<Boolean> unique = new ArrayList<Boolean>();

             alnList.add(firstAln);
             unique.add(true);
             trascriptomeSAMReader.mark(readAheadLimitEstimate * 2);
             String line=trascriptomeSAMReader.readLine();
             ReadAlignment currentAln = alnsHandler.createReadAlignment(line);
             while (line != null && currentAln == null) {
                 line=trascriptomeSAMReader.readLine();
                 currentAln = alnsHandler.createReadAlignment(line);
             }
             while ((currentAln != null) && (firstAln.getRead().getId().compareTo(currentAln.getRead().getId()) == 0)) {
                    boolean currentAlnUnique = true;
                    for (int i = 0; i<alnList.size(); i++) {
                        if (currentAln.getSequenceName().compareTo(alnList.get(i).getSequenceName()) == 0)
                            if (trascAlnComparator.compare((GenomicRegion)currentAln,(GenomicRegion)alnList.get(i)) == 0) {
                                currentAlnUnique=false;
                                unique.set(i, false);
                                break;
                            }
                    }
                    if (currentAlnUnique) {
                        alnList.add(currentAln);
                        unique.add(true);
                    }
                 currentAln = null;
                 while (line != null && currentAln == null) {
                     line=trascriptomeSAMReader.readLine();
                     currentAln = alnsHandler.createReadAlignment(line);
                 }
            }
            for (int i = 0; i<alnList.size(); i++)
                  if (unique.get(i))
                      uniqueList.add(alnList.get(i));
            return uniqueList;
        }

	/**
	 * Checks if there are other alignments with the same read ID as aln in a sorted
         * SAM file. Returns true if all alignments with the same ID are mapped to the same
         * genomic location (resulting from aligments to different isoforms of the same gene
         * Note: Pointer to Reader is set to the first alignment with different read ID than aln
	 * @param aln: Alignment
	 * @param reader:  Buffered reader to a SAM file sorted by ID.
	 */
	private boolean isMultiple(ReadAlignment aln,BufferedReader reader, int readAheadLimitEstimate,SAMReadAlignmentsHandler alnsHandler) throws IOException {
            
                boolean multiple = false; 

              //for alignments with soft clipping, a preprocessing step is done before hardmerge
              // to filter multiple alignments of the read based on which portion of the
              // read is aligned (from the cigar field)
                if (multiple_local_aln_filtered)
                    return false;
                reader.mark(readAheadLimitEstimate * 2);
                String line=reader.readLine();
                ReadAlignment nextAln = alnsHandler.createReadAlignment(line);
                while ((nextAln != null) && (aln.getRead().getId().compareTo(nextAln.getRead().getId()) == 0)) {
                        transcriptomeAlignmentCount++;
                        if (aln.getSequenceName().compareTo(nextAln.getSequenceName()) == 0) //ref seq name: chr1 etc..
                            if (trascAlnComparator.compare((GenomicRegion)aln,(GenomicRegion)nextAln) == 0) 
                                multiple = true;
                    reader.mark(readAheadLimitEstimate * 2);
                    line=reader.readLine();
                    nextAln = alnsHandler.createReadAlignment(line);
                }
                reader.reset(); // point to the first alignment with read ID not equal to that of aln; ready for reading next aln
                //transcriptomeAlignmentCount--;
		return multiple;
	}

        private boolean isMultiple(List<ReadAlignment> sortedAlns, int pos) { //old
		ReadAlignment aln1 = sortedAlns.get(pos);
		boolean multiple1 = (pos < sortedAlns.size()-1 && aln1.getRead().getId().equals(sortedAlns.get(pos+1).getRead().getId()));
		multiple1 = multiple1 || (pos>0 && aln1.getRead().getId().equals(sortedAlns.get(pos-1).getRead().getId()) );
		return multiple1;
	}
	/**
	 * Prints headers for the output file
	 * @param filename SAM file to get the SQ headers
	 * @param out Stream where headers will be printed
	 * @throws IOException If the file can not be read
	 */
	public void printHeader(String filename, PrintStream out) throws IOException {
		out.println("@HD\tVN:1.0\tSO:coordinate");
		SAMReadAlignmentsHandler.copySequenceHeaders(filename, out);
		out.println("@PG\tID:"+this.getClass().getName());
	}

        private void AddAlignment(ReadAlignment aln, List<ReadAlignment> alnList,SAMReadAlignmentsHandler alnsHandler) throws IOException {
            if (alnList.size() == alnListMaxSize) {
		alnsHandler.saveReadAlignments(alnList, outputFileStream);
                mergedAlnCount += alnList.size();
                alnList.clear();
		System.out.println(mergedAlnCount+" merged alignments written to output file");
            }
            alnList.add(aln);
        }

	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		CCDSReferenceAlnsMerger merger = new CCDSReferenceAlnsMerger();
		SAMReadAlignmentsHandler alnsHandlerTranscriptome=new SAMReadAlignmentsHandler();
		alnsHandlerTranscriptome.setMultipleAlnsPerRead(true);
		int i=0;
		while(i<args.length && args[i].charAt(0)=='-') {
			if("-t".equalsIgnoreCase(args[i])) {
				i++;
				merger.mergeType = args[i];
			}
                        if ("-l".compareTo(args[i]) == 0)
                            multiple_local_aln_filtered = true;
			i++;
		}
//		String coordsHandler = args[i];
//		i++;
		String file1 = args[i];
		i++;
		String file2 = args[i];
		i++;
		String outFile = args[i];
//		System.out.println("Loading Transcriptome Map..");
//		CCDSTranscriptome transcriptome = new CCDSTranscriptome(coordsHandler);
//		transcriptome.setCoordsTranslationOffset(1);
//		alnsHandlerTranscriptome.setAlignmentsTranslator(transcriptome);
//sahar read one alignment at a time
		//System.out.println("Loading: "+file1);
		//List<ReadAlignment> alns = alnsHandlerCCDS.loadReadAlignments(file1);
		//System.out.println("Loaded: "+alns.size()+" alignments.");
		//System.out.println("Merging: "+file2);
                outputFileStream = new PrintStream(outFile);
		merger.printHeader(file2,outputFileStream);
    		//List<ReadAlignment> alns =
                merger.mergeReadAlignments(file1, file2,alnsHandlerTranscriptome);
//end sahar
		System.out.println("Merged process finished.");
                System.out.println("Input genome aligned reads: "+genomeAlignedReadsCount);
                System.out.println("Input transcriptome aligned reads: "+transcriptomeAlignedReadsCount);
                System.out.println("Input genome alignments: "+genomeAlignmentCount);
                System.out.println("Input transcriptome alignments: "+transcriptomeAlignmentCount);
                System.out.println("Final Alignments: "+mergedAlnCount);

//		System.out.println("Sorting merged alignments");
//		Collections.sort(alns,new GenomicRegionComparator(transcriptome.getSequenceNames()));
//		System.out.println("Saving merged alignments");
//		PrintStream out = new PrintStream(outFile);
//		merger.printHeader(file2,out);
//		alnsHandlerTranscriptome.saveReadAlignments(alns, out);
		outputFileStream.close();
	}
}
