/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.main;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;

import edu.uconn.ngstools.genome.Genome;
import edu.uconn.ngstools.genome.GenomeAssembly;
import edu.uconn.ngstools.readalignments.ReadAlignment;
import edu.uconn.ngstools.readalignments.filehandlers.SAMReadAlignmentsHandler;
import edu.uconn.ngstools.transcriptome.CCDSTranscriptome;

/**
 * Program that takes a set of alignments and a reference genome and counts the number of 
 * mismatches with the reference for each read position from 5' to 3' end
 * @author Jorge Duitama
 */
public class ReadPositionStatistics {
	private long [] mismatches;
	private int totalReads = 0;
	private Genome genome;
	public void processFile(String filename) throws IOException {
		FileInputStream fis = new FileInputStream(filename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		SAMReadAlignmentsHandler handler = new SAMReadAlignmentsHandler();
		ReadAlignment aln = handler.createReadAlignment(in);
		for(int i=0;aln!=null;i++) {
			String read = aln.getRead().toString();
			if(!ReadAlignment.DEFAULT_SEQ_NAME.equals(aln.getSequenceName())) {
				boolean reverse = (aln.getFlags()/16)%2==1;
				boolean readInRegion = false;
				for(int j=0;j<read.length();j++) {
					char baseRead = Character.toUpperCase(read.charAt(j));
					if(baseRead!='N') {
						int pos = aln.getSequencePosition(j);
						if(pos>=0) {
							char baseRef = Character.toUpperCase(genome.getReferenceBase(aln.getSequenceName(), pos));
							if(baseRef!=0 ) {
								readInRegion = true;
								if(baseRead !=baseRef) {
									int posStats = j;
									if(reverse) posStats = read.length()-1-j;
									mismatches[posStats]++;		
								}
							}
						}
					}
				}
				if(readInRegion) {
					totalReads++;
				}
			}
			aln = handler.createReadAlignment(in);
		}
		fis.close();
	}
	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		ReadPositionStatistics stats = new ReadPositionStatistics();
		String mapFilename = null;
		int i=0;
		while(i<args.length && args[i].charAt(0)=='-') {
			if("-mapFile".equalsIgnoreCase(args[i])) {
				i++;
				mapFilename = args[1];
			}
			i++;
		}
		String reference = args[i];
		i++;
		String alnsFile = args[i];
		i++;
		int readLength = Integer.parseInt(args[i]);
		i++;
		if(mapFilename!=null) {
			stats.genome = new CCDSTranscriptome(mapFilename,reference);
		} else {
			stats.genome = new GenomeAssembly(reference);
		}
		stats.mismatches = new long [readLength];
		for(int j=0;j<stats.mismatches.length;j++) {
			stats.mismatches [j] =0;
		}
		stats.processFile(alnsFile);
		for(int j=0;j<stats.mismatches.length;j++) {
			System.out.println(""+(j+1)+" "+stats.mismatches [j]);
		}
		System.out.println();
		System.out.println("Reads: "+stats.totalReads);

	}

}
