/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.main;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Deque;
import java.util.LinkedList;
import java.util.List;

import edu.uconn.ngstools.genome.Genome;
import edu.uconn.ngstools.genome.GenomeAssembly;
import edu.uconn.ngstools.readalignments.ReadAlignment;
import edu.uconn.ngstools.readalignments.filehandlers.SAMReadAlignmentsHandler;
import edu.uconn.ngstools.sequences.DNASequence;
import edu.uconn.ngstools.sequences.DegenerateSequence;
import edu.uconn.ngstools.transcriptome.CCDSTranscriptome;
import edu.uconn.ngstools.variants.CalledSNV;
import edu.uconn.ngstools.variants.CountsHelper;
import edu.uconn.ngstools.variants.PileupRecord;
import edu.uconn.ngstools.variants.SNVSFilter;
import edu.uconn.ngstools.variants.filehandlers.CalledSNVSFileHandler;
import edu.uconn.ngstools.variants.filehandlers.PileupFileHandler;

/**
 * Program that implements the Single Nucleotide Variants (SNV) detection and genotyping 
 * algorithm SNVQ. It can process a pileup file or an alignments file with its corresponding
 * reference and calls SNVs according with a bayesian model based on quality scores of allele
 * calls 
 * @author Jorge Duitama
 *
 */
public class SNVSFinder {
	public static final String CALLING_METHOD_SNVQ = "snvq";
	public static final String CALLING_METHOD_INFILE = "infile";
	public static final double DEF_HETEROZYGOSITY_RATE = 0.001;
	private String callingMethod=CALLING_METHOD_SNVQ;
	private DNASequence dummySequence = new DNASequence();
	private SAMReadAlignmentsHandler samHandler = new SAMReadAlignmentsHandler();
	private CountsHelper countsHelper=new CountsHelper();
	private String querySeq=null;
	private int absoluteFirst = 0;
	private int absoluteLast = 0;
	private boolean allVariantCalls = false;
	private double heterozygosityRate = DEF_HETEROZYGOSITY_RATE;
        private boolean strandCounts = false;
	
	private List<CalledSNV> snvs = new ArrayList<CalledSNV>();
	private CalledSNVSFileHandler snvsFileHandler = new CalledSNVSFileHandler();
	private PrintStream out = null;
	private SNVSFilter filter = new SNVSFilter();
	
	/**
	 * Creates a default instance of the SNVSFinder
	 */
	public SNVSFinder() {
		super();
	}
	/**
	 * Calls SNVs given an alignments file inside the given reference genome.
	 * @param alignmentsFile SAM file with alignments to the given reference genome
	 * @param genome Object to look for reference bases
	 * @throws IOException If the file can not be read
	 */
	public void findSNVS(String alignmentsFile, Genome genome) throws IOException {
		FileInputStream fis = new FileInputStream(alignmentsFile);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		//Queue to keep all alignments spanning the current genomic location 
		//or alignments that were already read and may be used later
//		Deque<ReadAlignment> alnsQueue = new LinkedList<ReadAlignment>();
                List<ReadAlignment> alns = new ArrayList<ReadAlignment>();

		//Genomic location to test
		String seq = null;
		int snvPosition=0;
//sahar debugging uncalled SNVs
//		System.out.println("in findSNVS");
//                setAllVariantCalls(true);
		ReadAlignment aln = samHandler.createReadAlignment(in);
		if(querySeq!=null) {
			//Look for first relevant alignment
//			System.out.println("in if querySep!= null questSeq = "+querySeq);
			while(aln!=null) {
				if(querySeq.equalsIgnoreCase(aln.getSequenceName()) && aln.getEnd()>=absoluteFirst) {
                                        alns.add(aln);
//					alnsQueue.add(aln);
					break;
				}
				aln = samHandler.createReadAlignment(in);
			}
			if(aln == null) {
				return;
			}
			seq = querySeq;
			snvPosition = absoluteFirst;
		} else if(aln!=null) {
			seq = aln.getSequenceName();
			snvPosition = aln.getStart();
//			System.out.println("in if aln!= null.. seq="+seq+" pos = "+snvPosition);
			//alnsQueue.add(aln);
                        alns.add(aln);
		} else {
			return;
		}
		boolean readsRemaining = true;
                // Get current time
//                long start = System.currentTimeMillis();
 //               long currentTime;
                int LastAlignmentForPositionIndex = 0;
		while(true) {
//                    if (snvPosition%1000 == 0) {
                // Get elapsed time in milliseconds
//                    currentTime=System.currentTimeMillis();
//                    System.err.println("ElapsedTime = "+((currentTime-start)/1000F));
//                    start = currentTime;
//                    }

//			System.err.println("In main loop "+seq+" "+snvPosition);
			if(querySeq!=null && !querySeq.equalsIgnoreCase(seq)) {
				break;
			}
			if(querySeq!=null && snvPosition > absoluteLast) {
				break;
			}
			//Drop from the queue alignments that will not be needed anymore and include new ones from the Reader
//			readsRemaining = updateQueue(alnsQueue,seq,snvPosition,in,readsRemaining);
			 LastAlignmentForPositionIndex = updateList(alns,seq,snvPosition,in,LastAlignmentForPositionIndex);
//			if(alnsQueue.size()==0) {
			if(alns.size()==0) {
				break;
			}
// this is now done inside updateList
//			List<ReadAlignment> alns = getAlignments(alnsQueue,seq,snvPosition);
//			if(alns.size()==0) {
			ReadAlignment firstAln  = alns.get(0);
                        // if there is a gap between snvPosition and start of first alignment on list; skip to start of first alignment on list
                        if (!firstAln.getSequenceName().equalsIgnoreCase(seq) ||  firstAln.getStart() > snvPosition) {
//                        System.err.println("ADVANCING");
//                        System.err.println("aln data: start, cigar, end");
//                       System.err.println(firstAln.getStart());
//                       System.err.println(firstAln.getSequenceName());
//                       System.err.println(firstAln.getCigar());
//                        System.err.println(firstAln.getEnd());
				seq = firstAln.getSequenceName();
				snvPosition = firstAln.getStart();
                                continue; // need to go back to updateList
			}
			//Calls a SNV for a single genomic location
//        		System.err.println("alnsQueue.size() "+alnsQueue.size());
//           		System.err.println("alns.size() "+alns.size());

			callSNV(seq,snvPosition,alns,genome,LastAlignmentForPositionIndex);
			snvPosition++;
//			System.err.println("In main loop after position++"+seq+" "+snvPosition);
		}
	}
	//Saves a SNV call either in a stream or in a list to be retrieved
	private void saveSNV(CalledSNV snv) {
		if(snv!=null && filter.passFilter(snv)) {
			if(out!=null) {
				snvsFileHandler.saveSNV(snv, out,strandCounts);
			} else {
				snvs.add(snv);
			}
		}
	}
	//Updates the alignments queue. Drop from the queue alignments that will not be needed anymore and include new ones from the Reader
//	private boolean updateQueue(Deque<ReadAlignment> alnsQueue, String seq, int snvPosition, BufferedReader in, boolean readsRemaining) throws IOException {
	private int updateList(List<ReadAlignment> alns, String seq, int snvPosition, BufferedReader in, int positionLastAlignmentIndex) throws IOException {
		//remove alignments before the current position
		for (int i = 0; i < alns.size(); i++) {
                    ReadAlignment aln = alns.get(i);
		    if(aln.getSequenceName().equalsIgnoreCase(seq) &&  aln.getEnd() < snvPosition) {
//                        System.err.println("REMOVING");
//                        System.err.println("In remove loop i "+i);
//                        System.err.println("aln data: start, cigar, end");
//                        System.err.println(aln.getStart());
//                        System.err.println(aln.getCigar());
//                        System.err.println(aln.getEnd());
				alns.remove(i);
                                i--;
                                positionLastAlignmentIndex--;
                    }
		}
//                positionLastAlignmentIndex = Math.max(positionLastAlignmentIndex, 0);
                for (int i = positionLastAlignmentIndex; i > -1 && i < alns.size() ; i++) {
                    ReadAlignment aln = alns.get(i);
                    if(aln.getSequenceName().equalsIgnoreCase(seq) && aln.getStart()<=snvPosition) {
                        positionLastAlignmentIndex = i;
                    }
                    else
                        break;
                }
                ReadAlignment aln;
                if (alns.size() > 0) {
                    aln = alns.get(alns.size() - 1);
                    if(!aln.getSequenceName().equalsIgnoreCase(seq) || aln.getStart() > snvPosition) // next chromosome or start > current position
                        return positionLastAlignmentIndex;
                }
		aln = samHandler.createReadAlignment(in);
                while (aln != null) {
//                        System.err.println("ADDING");
//                        System.err.println("aln data: start, cigar, end");
//                        System.err.println(aln.getStart());
//                        System.err.println(aln.getCigar());
//                        System.err.println(aln.getEnd());
                    alns.add(aln);
                    if(aln.getSequenceName().equalsIgnoreCase(seq) && aln.getStart()<=snvPosition)
                        positionLastAlignmentIndex = alns.size() - 1;
                    else {
//                        System.err.println("STOP ADDING ADDING");
                        break;
                    }
		aln = samHandler.createReadAlignment(in);
            }
            return positionLastAlignmentIndex;

	}
	//Gets from the queue relevant alignments for the given genomic location
	private List<ReadAlignment> getAlignments(Deque<ReadAlignment> alnsQueue, String seq, int snvPosition) {
		List<ReadAlignment> answer = new ArrayList<ReadAlignment>();
		for(ReadAlignment aln:alnsQueue) {
			if(aln.getSequenceName().equalsIgnoreCase(seq) && aln.getStart()<=snvPosition && aln.getEnd()>= snvPosition) {
				answer.add(aln);
			}
		}
		return answer;
	}
	/**
	 * Tries to call a SNV for a single genomic location from the given alignments
	 * @param seq Sequence name of the genomic location
	 * @param position One based position of the genomic location
	 * @param alns List of alignments spanning the given location 
	 * @param genome Object to query the reference base
	 */
	public void callSNV(String seq, int position, List<ReadAlignment> alns, Genome genome,int LastAlignmentForPositionIndex) {
		char refBase=genome.getReferenceBase(seq, position);
		if(refBase == 0) {
			return;
		}
		startCountsHelper(seq,position);
		for(int i = 0; i < LastAlignmentForPositionIndex; i++) {
                        ReadAlignment aln = alns.get(i);
			int readPos  = aln.getReadPosition(position);
			if(readPos >=0) {
				char nextBase = Character.toUpperCase(aln.getRead().charAt(readPos));
				char nextCharScore = '+';
				String qualScores = aln.getRead().getQualityScores();
				if(qualScores!=null && qualScores.length()> readPos) {
					nextCharScore = qualScores.charAt(readPos);
				}
				countsHelper.updateCounts(nextBase, nextCharScore, aln.getMappingQuality(),aln.getFlags());
			}
		}
		CalledSNV csnv = callSNV(seq, position, refBase);
		saveSNV(csnv);
	}
	//Starts the counts helper according with this heterozygous rate
	private void startCountsHelper(String seqName, int position) {
		if(heterozygosityRate>0) {
			countsHelper.startCounts(heterozygosityRate);
		} else {
			countsHelper.startCounts();
		}
	}
	/**
	 * Calls SNVs from the information provided in the pileup file
	 * @param pileupFile File in pileup format with locations to test
	 * @throws IOException If the file can not be read
	 */
	public void findSNVS(String pileupFile) throws IOException {
		PileupFileHandler pileupHandler = new PileupFileHandler();
		FileInputStream fis = new FileInputStream(pileupFile);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		while(line!=null) {
			//Loads record and tries to call a SNV for the single location
			PileupRecord record = pileupHandler.loadPileupRecord(line);
			String seq = record.getSequenceName();
			int position = record.getPosition();
			if(querySeq == null || (querySeq.equalsIgnoreCase(seq) && position>=absoluteFirst && position <=absoluteLast)) {
				CalledSNV snv = callSNV(record);
				saveSNV(snv);
			}
			line=in.readLine();
		}
		fis.close();
	}
	/**
	 * Tries to call a SNV from the given pileup record
	 * @param record Record with information to call a SNV
	 * @return CalledSNV Call information after running the SNV calling method
	 */
	public CalledSNV callSNV(PileupRecord record) {
		calculateCounts(record);
		CalledSNV csnv = null;
		if(CALLING_METHOD_INFILE.equals(callingMethod)) {
			csnv = callSNVInfile(record);
		} else {
			csnv = callSNV(record.getSequenceName(),record.getPosition(),record.getRefBase());
		} 
		return csnv;
	}
	/**
	 * Just reads the genotype information in the pileup record and calls a SNV accordingly
	 * @param record Pileup record with the locus and genotype information
	 * @return CalledSNV Call consistent with the record or null if the genotype is homozygous
	 * reference and the allVariantCalls flag is turned off
	 */
	public CalledSNV callSNVInfile(PileupRecord record) {
		char upRefBase = Character.toUpperCase(record.getRefBase());
		char genDeg = Character.toUpperCase(record.getGenotype());
		double prob = record.getGenotypeProbability();
		double probRef = record.getRefProbability();
		double countRef = countsHelper.getCount(upRefBase);
		if(countRef == countsHelper.getTotalCount()) {
			return null;
		}
		if(!allVariantCalls && upRefBase == genDeg) {
			return null;
		}
		double [] counts = countsHelper.getCounts();
		String genotype = ""+genDeg+""+genDeg;
		char altBase = genDeg;
		if(!DNASequence.isInAlphabeth(genDeg)) {
			genotype = DegenerateSequence.getExtendedBases(genDeg);
			if(genotype.length()!=2) {
				System.err.println("WARN: Unknown genotype for SNP at "+record.getSequenceName()+":"+record.getPosition());
				return null;
			}
			if(genotype.charAt(0)!=upRefBase) {
				altBase = genotype.charAt(0);
			} else {
				altBase = genotype.charAt(1);
			}
		} else if (altBase == upRefBase) {
			int index2 = getMaxIndex(counts,dummySequence.getAlphabetIndex(upRefBase));
			altBase = dummySequence.getAlphabetCharacter(index2);
		}
		CalledSNV csnp = new CalledSNV(record.getSequenceName(), record.getPosition(), record.getRefBase(), altBase, genotype, countRef,countsHelper.getCount(altBase),prob);
		csnp.setRefProbability(probRef);
		csnp.setAllCounts(Arrays.copyOf(counts,counts.length));
//                csnp.setpAllCounts(Arrays.copyOf(countsHelper.getpCounts(),countsHelper.getpCounts().length));
//                csnp.setnAllCounts(Arrays.copyOf(countsHelper.getnCounts(),countsHelper.getnCounts().length));
		return csnp;
	}
	//Calculates counts from the pileup information
	private void calculateCounts(PileupRecord record) {
		startCountsHelper(record.getSequenceName(),record.getPosition());
		int j=0;
		String calls = record.getAlleleCalls();
		String scores = record.getQualityScores();
		//System.out.println("Calls: "+calls+" Scores: "+scores);
		for(int i=0;i<calls.length()&& j<scores.length();i++) {
			char nextBase = Character.toUpperCase(calls.charAt(i));
			if(nextBase == '.' || nextBase == ',') {
				nextBase = Character.toUpperCase(record.getRefBase());
			}
			//Needed to ignore annotations different than bases
			int index = dummySequence.getAlphabetIndex(nextBase);
			if(index>=0) {
				char nextCharScore = scores.charAt(j);
				//System.out.println("Base: "+nextBase+". Score: "+nextCharScore);
				countsHelper.updateCounts(nextBase, nextCharScore, (short)255);
				j++;
			} else if (nextBase == 'N') {
				j++;
			}
		}
		//countsHelper.printProbs(countsHelper.getPosteriorProbabilities());
	}
	//PRE: Counts are already calculated in the countsHelper object 
	private CalledSNV callSNV(String seq, int position, char refBase) {
		char upRefBase = Character.toUpperCase(refBase);
		CalledSNV snv = null;
		if(countsHelper.getTotalCount()==0) {
			return snv;
		}
		if(!DNASequence.isInAlphabeth(upRefBase)) {
			refBase = 'N';
		}
		//Calculate homozygous and heterozygous probabilities from contributions
		snv = callSNVQ(seq,position,upRefBase);
		if (snv!=null) {
			//Put back lowercase if needed
			snv.setRefBase(refBase);
			double [] counts = Arrays.copyOf(countsHelper.getCounts(),countsHelper.getCounts().length);
			snv.setAllCounts(counts);
			double [] pcounts = Arrays.copyOf(countsHelper.getpCounts(),countsHelper.getpCounts().length);
			snv.setpAllCounts(pcounts);
			double [] ncounts = Arrays.copyOf(countsHelper.getnCounts(),countsHelper.getnCounts().length);
			snv.setnAllCounts(ncounts);
		}
		return snv;
	}
	//Calls the SNVQ algorithm from the counts stored in the counts helper object   sssssssssssssssssssssss
	private CalledSNV callSNVQ(String seq, int position, char refBase ) {
		int indexRef = dummySequence.getAlphabetIndex(refBase);
		double [] counts = countsHelper.getCounts();
		double [][] postProbs = countsHelper.getPosteriorProbabilities();
		double countRef = 0;
		double refProb = 0;
		double probMax = postProbs[0][0];
		int indexI=0,indexJ=0;
		if(indexRef>=0) {
			countRef = counts[indexRef];
			refProb = postProbs[indexRef][indexRef];
			//Prefer the reference in case of ties
			probMax = refProb;
			indexI = indexRef;
			indexJ = indexRef;
		}
		if(countRef == countsHelper.getTotalCount()) {
			return null;
		}
		for(int i=0;i<postProbs.length;i++) {
			for(int j=i;j<postProbs[0].length;j++) {
				//Differences of less than 0.01 are considered equal
				if (postProbs[i][j] > probMax+0.01) {
					probMax = postProbs[i][j];
					indexI = i;
					indexJ = j;
				}
			}
		}
		
		int indexAlt=0;
		char altBase=0;
		String genotype=null;
		double posteriorProbability=1;
		boolean homoRef =false;
		if (indexI!=indexJ) {
			if(indexI!=indexRef && indexJ!=indexRef) {
				//Triallelic SNV
				char thirdBase;
				if(postProbs[indexI][indexI]>postProbs[indexJ][indexJ]+0.01) {
					indexAlt = indexI;
					thirdBase = dummySequence.getAlphabetCharacter(indexJ);
				} else {
					indexAlt = indexJ;
					thirdBase = dummySequence.getAlphabetCharacter(indexI);
				}
				altBase = dummySequence.getAlphabetCharacter(indexAlt);
				genotype = ""+altBase+""+thirdBase;
			} else {
				//Call heterozygous SNV
				if(indexRef!=indexI ) {
					indexAlt = indexI;
				} else {
					indexAlt = indexJ;
				}
				altBase = dummySequence.getAlphabetCharacter(indexAlt);
				genotype = ""+refBase+""+altBase;
			}
			posteriorProbability = postProbs[indexI][indexJ];
		} else if(indexRef!=indexI ) {
			//Homozygous non reference
			indexAlt = indexI;
			altBase = dummySequence.getAlphabetCharacter(indexAlt);
			genotype = ""+altBase+""+altBase;
			posteriorProbability = postProbs[indexAlt][indexAlt];
		} else {
			//Homozygous reference
			indexAlt = getMaxIndex(counts,indexRef);
			altBase = dummySequence.getAlphabetCharacter(indexAlt);
			genotype = ""+refBase+""+refBase;
			posteriorProbability = postProbs[indexRef][indexRef];
			homoRef = true;
		}
		
		if(!allVariantCalls && homoRef) {
			return null;
		}
		//Creates a homozygous reference call
		CalledSNV csnp = new CalledSNV(seq,position,refBase,altBase,genotype,countRef,counts[indexAlt],posteriorProbability);
		csnp.setRefProbability(refProb);

                //strand specific counts  (Sahar)
                csnp.setpCountReference(countsHelper.getpCount(refBase));
                csnp.setnCountReference(countsHelper.getnCount(refBase));
                csnp.setpCountAlternative(countsHelper.getpCount(altBase));
                csnp.setnCountAlternative(countsHelper.getnCount(altBase));
		return csnp;
	}
	//Gets the index with the maximum from the given values excluding the given index
	private int getMaxIndex(double[] values,int excludeIndex) {
		int index = 0;
		if(excludeIndex == 0) {
			index=1;
		}
		for (int i=0;i<values.length;i++) {
			if(i!=excludeIndex && values[i]>values[index]) {
				index = i;
			}
		}
		return index;
	}
	/**
	 * @return boolean true if all loci with variant calls are reported, false otherwise
	 */
	public boolean isAllVariantCalls() {
		return allVariantCalls;
	}
	/**
	 * Sets the allVariantCalls flag
	 * @param allVariantCalls true if all loci with variant calls must be reported, false otherwise
	 */
	public void setAllVariantCalls(boolean allVariantCalls) {
		this.allVariantCalls = allVariantCalls;
	}
	/**
	 * @return double Apriori heterozygous rate
	 */
	public double getHeterozygosityRate() {
		return heterozygosityRate;
	}
	/**
	 * Changes the apriori heterozygous rate
	 * @param heterozygosityRate New rate
	 */
	public void setHeterozygosityRate(double heterozygosityRate) {
		this.heterozygosityRate = heterozygosityRate;
	}
	/**
	 * @return List<CalledSNV> Called SNVs
	 */
	public List<CalledSNV> getSnvs() {
		return snvs;
	}
	/**
	 * Changes the snvs filter
	 * @param filter SNVs filter
	 */
	public void setFilter(SNVSFilter filter) {
		this.filter = filter;
	}
	/**
	 * 
	 * @param args
	 * @throws Exception
	 */
	public static void main (String [] args) throws Exception{
		SNVSFinder finder = new SNVSFinder();
		String referenceFile=null;
		String mapFile=null;
		int i=0;
		while(i<args.length && args[i].charAt(0)=='-') {
			if("-method".equalsIgnoreCase(args[i])) {
				i++;
				finder.callingMethod = args[i];
			} else if("-querySeq".equalsIgnoreCase(args[i])) {
				i++;
				finder.querySeq = args[i];
			} else if("-start".equalsIgnoreCase(args[i])) {
				i++;
				finder.absoluteFirst = Integer.parseInt(args[i]);
			} else if("-end".equalsIgnoreCase(args[i])) {
				i++;
				finder.absoluteLast = Integer.parseInt(args[i]);
			} else if("-h".equalsIgnoreCase(args[i])) {
				i++;
				finder.heterozygosityRate = Double.parseDouble(args[i]);
			} else if("-allVariantCalls".equalsIgnoreCase(args[i])) {
				finder.allVariantCalls = true;
				finder.filter.setKeepLowerCaseRef(true);
			} else if("-keepLowerCaseRef".equalsIgnoreCase(args[i])) {
				finder.filter.setKeepLowerCaseRef(true);
			} else if("-minAltCoverage".equalsIgnoreCase(args[i])) {
				i++;
				finder.filter.setMinAltCoverage(Double.parseDouble(args[i]));
			} else if("-maxAltCoverage".equalsIgnoreCase(args[i])) {
				i++;
				finder.filter.setMaxAltCoverage(Double.parseDouble(args[i]));
			} else if("-minProbability".equalsIgnoreCase(args[i])) {
				i++;
				finder.filter.setMinProbability(Double.parseDouble(args[i]));
			} else if("-ccdsMap".equalsIgnoreCase(args[i])) {
				i++;
				mapFile = args[i];
			} else if("-ccdsRef".equalsIgnoreCase(args[i])) {
				i++;
				referenceFile = args[i];
			} else if("-reference".equalsIgnoreCase(args[i])) {
				i++;
				referenceFile = args[i];
			} else if("-strand".equalsIgnoreCase(args[i])) {
                                finder.strandCounts=true;
                        }else {
				System.err.println("Unrecognized option: "+args[i]);
				return;
			}
			i++;
		}
		
		String inFile = args[i];
		i++;
		finder.out = System.out;
                finder.snvsFileHandler.printHeader(finder.out,finder.strandCounts);
		if(mapFile != null && referenceFile!=null) {
			finder.findSNVS(inFile,new CCDSTranscriptome(mapFile, referenceFile));
		} else if(referenceFile!=null) {
			finder.findSNVS(inFile,new GenomeAssembly(referenceFile));
		} else {
			finder.findSNVS(inFile);
		}
		
		finder.out.close();
	}
}
