/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.readalignments;

import edu.uconn.ngstools.sequences.DNAMaskedSequence;


/**
 * Information from a raw read. A read is implemented as a DNAMaskedSequence with
 * an id and optional quality scores. This allows to use the space saving features
 * of custom sequences
 * @author Jorge Duitama
 */
public class Read extends DNAMaskedSequence {
	private byte [] id;
	private byte [] qualityScores;
	/**
	 * Creates a new read with the given information
	 * @param id of the new read
	 * @param sequence DNA sequence of the read
	 * @param qualityScores Phred score for each base in the read
	 */
	public Read(String id, String sequence, String qualityScores) {
		super(sequence);
		this.setId(id);
		this.setQualityScores(qualityScores);
	}
	/**
	 * @return String Id of the read
	 */
	public String getId() {
		if(id!=null) {
			return new String(id);
		}
		return null;
	}
	/**
	 * Changes the id of the read
	 * @param id New read id
	 */
	public void setId(String id) {
		if(id == null) { 
			this.id = null;
		} else {
			this.id = id.getBytes();
		}
		
	}
	/**
	 * @return String Quality scores of the read 
	 */
	public String getQualityScores() {
		if(qualityScores != null) {
			return new String(qualityScores);
		} else {
			return null;
		}
	}
	/**
	 * Changes the quality scores
	 * @param qualityScores New quality scores
	 */
	public void setQualityScores(String qualityScores) {
		if(qualityScores == null) {
			this.qualityScores = null;
		} else {
			this.qualityScores = qualityScores.getBytes();
		}
	}
	/**
	 * Creates a new read with the same id but reverse complemented bases
	 * @return Read new read
	 */
	public Read createReverseComplement() {
		StringBuffer sb = new StringBuffer(getQualityScores());
		sb.reverse();
		return new Read(getId(), getReverseComplement(), sb.toString());
	}	
}