/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.readalignments;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import edu.uconn.ngstools.genome.GenomicRegion;

/**
 * Class that stores read alignments to a reference genome
 * The internal representation is consistent with the SAM format
 * @author Jorge Duitama
 *
 */
public class ReadAlignment implements GenomicRegion {
	public static final String DEFAULT_SEQ_NAME="*";
	public static final String EQUAL_SEQ_NAME="=";
	private Read read;
	private String sequenceName=DEFAULT_SEQ_NAME;
	private int start;
	private int end; //Calculated from CIGAR. Always greater than start
	private short mappingQuality;
	private short flags;
	private String cigar;
	private String mateSequenceName=DEFAULT_SEQ_NAME;
	private int matePosition=0;
	private int insertSize=0;
	/**
	 * Creates a new alignment with the given data
	 * @param read Object with the raw read information
	 * @param sequenceName Name of the sequence where the read is aligned
	 * @param start Position in the sequence corresponding with the first base in the clipped read
	 * @param cigar Sequence describing the alignment (See SAM format documentation)
	 * @param mappingQuality Quality score of the alignment
	 */
	public ReadAlignment(Read read, String sequenceName, int start, String cigar, short mappingQuality) {
		super();
		this.setRead(read);
		this.setSequenceName(sequenceName);
		this.setMappingQuality(mappingQuality);
		this.start = start;
		this.setCigar(cigar);
	}
	/**
	 * @return Read Object with the raw read information
	 */
	public Read getRead() {
		return read;
	}
	/**
	 * Changes the raw read information
	 * @param read New read information
	 */
	public void setRead(Read read) {
		this.read = read;
	}
	/**
	 * @return String Name of the sequence where the read is aligned
	 */
	public String getSequenceName() {
		return sequenceName;
	}
	/**
	 * Changes the name of the sequence where the read is aligned
	 * @param sequenceName New sequence name
	 */
	public void setSequenceName(String sequenceName) {
		this.sequenceName = sequenceName;
	}
	/**
	 * @return int Position in the sequence corresponding with the first base in the clipped read
	 */
	public int getStart() {
		return start;
	}
	/**
	 * Changes the position in the sequence corresponding with the first base in the clipped read.
	 * The end position is updated according with the cigar
	 * @param start New start
	 */
	public void setStart(int start) {
		assert start >=0;
		this.start = start;
		updateEnd();
	}
	/**
	 * @return int Position in the sequence corresponding with the last base in the clipped read
	 */
	public int getEnd() {
		return end;
	}
	/**
	 * @return short Quality score of the alignment
	 */
	public short getMappingQuality() {
		return mappingQuality;
	}
	/**
	 * Changes the quality score of the alignment
	 * @param mappingQuality New quality score of the alignment
	 */
	public void setMappingQuality(short mappingQuality) {
		this.mappingQuality = mappingQuality;
	}
	/**
	 * @return String Sequence describing the alignment (See SAM format documentation)
	 */
	public String getCigar() {
		return cigar;
	}
	/**
	 * Changes the sequence describing the alignment. The end position is updated according with the
	 * start position
	 * @param cigar New cigar
	 */
	public void setCigar(String cigar) {
		this.cigar = cigar;
		updateEnd();
	}
	/**
	 * @return short Number made up of information about the read in the form of flags (See SAM format documentation)
	 */
	public short getFlags() {
		return flags;
	}
	/**
	 * Changes the flags information for this alignment
	 * @param flags New flags
	 */
	public void setFlags(short flags) {
		this.flags = flags;
	}
	/**
	 * @return String Sequence name where the mate was mapped
	 */
	public String getMateSequenceName() {
		return mateSequenceName;
	}
	/**
	 * Changes the sequence name where the mate was mapped
	 * @param mateSequenceName New sequence name
	 */
	public void setMateSequenceName(String mateSequenceName) {
		this.mateSequenceName = mateSequenceName;
	}
	/**
	 * @return int position on the mate sequence where the mate was mapped
	 */
	public int getMatePosition() {
		return matePosition;
	}
	/**
	 * Changes the position on the mate sequence where the mate was mapped
	 * @param matePosition New position
	 */
	public void setMatePosition(int matePosition) {
		this.matePosition = matePosition;
	}
	/**
	 * @return int Inferred insert size (See: SAM spec documentation)
	 */
	public int getInsertSize() {
		return insertSize;
	}
	/**
	 * Changes the inferred insert size
	 * @param insertSize New insert size
	 */
	public void setInsertSize(int insertSize) {
		this.insertSize = insertSize;
	}
	/**
	 * Translates the position relative to the sequence where the read is aligned to a zero 
	 * based position relative to the start of the read 
	 * @param sequencePosition position relative to the sequence where the read is aligned.
	 * @return int Corresponding zero based position relative to the start of the read or 
	 * a negative number if the given position can not be mapped to a read position
	 */
	public int getReadPosition(int sequencePosition) {
		if(sequencePosition < start) {
			return -60;
		}
		if(sequencePosition > end) {
			return -62;
		}
		int readPos = 0;
		int seqPos = start;
		List<Integer> eventIndexes = getCigarEventIndexes(cigar);
		int lastEventIndex = -1;
		for(int i:eventIndexes) {
			char event = cigar.charAt(i);
			int eventLength = Integer.parseInt(cigar.substring(lastEventIndex+1, i));
			lastEventIndex = i;
			int nextSeqPos = getNextSeqPos(seqPos, event, eventLength);
			int nextReadPos = getNextReadPos(readPos, event, eventLength);
			if(seqPos <= sequencePosition && sequencePosition < nextSeqPos) {
				int dif = sequencePosition - seqPos;
				if(nextReadPos - readPos > dif) {
					return readPos + dif;
				} else if (event=='D') {
					return -45;
				} else {
					return -42;
				}
			}
			seqPos = nextSeqPos;
			readPos = nextReadPos;
		}
		return -62;
	}
	/**
	 * Calculates the read base of the position relative to the sequence where the read is aligned 
	 * @param sequencePosition position relative to the sequence where the read is aligned.
	 * @return char Base call in the corresponding position of the read or one of the following
	 * characters depending the case:
	 * &lt; if the position lies before the alignment
	 * &gt; if the position lies after the alignment
	 * -    if the position is called in the read as a deletion
	 * *    if the position is padded in the alignment
	 */
	public char getReadBase(int sequencePosition) {
		int pos = getReadPosition(sequencePosition);
		if(pos < 0) {
			return (char)(-pos);
		}
		return read.charAt(pos);
	}
	/**
	 * Translates the position relative to the start of the read to 
	 * a position relative to the sequence where the read is aligned  
	 * @param readPosition Zero based position in the read
	 * @return int Corresponding position relative to the sequence where the read is aligned or
	 * a negative number if the given position was not aligned to the reference
	 */
	public int getSequencePosition(int readPosition) {
		if(readPosition < 0) {
			return -60;
		}
		if(readPosition > read.length()) {
			return -62;
		}
		int readPos = 0;
		int seqPos = start;
		List<Integer> eventIndexes = getCigarEventIndexes(cigar);
		int lastEventIndex = -1;
		for(int i:eventIndexes) {
			char event = cigar.charAt(i);
			int eventLength = Integer.parseInt(cigar.substring(lastEventIndex+1, i));
			lastEventIndex = i;
			int nextSeqPos = getNextSeqPos(seqPos, event, eventLength);
			int nextReadPos = getNextReadPos(readPos, event, eventLength);
			if(readPos <= readPosition && readPosition < nextReadPos) {
				int dif = readPosition - readPos;
				if(nextSeqPos - seqPos > dif) {
					return seqPos + dif;
				} else if (event=='I') {
					return -43;
				} else {
					return -42;
				}
			}
			seqPos = nextSeqPos;
			readPos = nextReadPos;
		}
		return -62;
	}
	/**
	 * computes the cigar element character corresponding to the input read position
	 * @param readPosition Zero based position in the read
	 * @return char the cigar element corresponding to the input read position
         *          or '!' if readPosition is out of range
	 */
	public char getCigarElem(int readPosition) {
		if(readPosition < 0) {
			return '!';
		}
		if(readPosition > read.length()) {
			return '!';
		}
		int readPos = 0;
		int seqPos = start;
		List<Integer> eventIndexes = getCigarEventIndexes(cigar);
		int lastEventIndex = -1;
		for(int i:eventIndexes) {
			char event = cigar.charAt(i);
			int eventLength = Integer.parseInt(cigar.substring(lastEventIndex+1, i));
			lastEventIndex = i;
			int nextSeqPos = getNextSeqPos(seqPos, event, eventLength);
			int nextReadPos = getNextReadPos(readPos, event, eventLength);
			if(readPos <= readPosition && readPosition < nextReadPos) {
                                return event;
			}
			seqPos = nextSeqPos;
			readPos = nextReadPos;
		}
		return '!';
	}
	/**
	 * Reverses a given cigar
	 * @param cigar Cigar to reverse
	 * @return String A new cigar with the events reversed
	 */
	public static String getReverseCigar(String cigar) {
		StringBuffer answer = new StringBuffer();
		List<Integer> eventIndexes = getCigarEventIndexes(cigar);
		Collections.reverse(eventIndexes);
		int nextEventIndex = cigar.length()-1;
		for(int i:eventIndexes) {
			if(nextEventIndex!=i) {
				answer.append(cigar.substring(i+1, nextEventIndex+1));
				nextEventIndex = i;	
			}
		}
		//Process first event
		answer.append(cigar.substring(0, nextEventIndex+1));
		return answer.toString();
	}
	/**
	 * Clips bases left and right in this alignment according with the given parameters.
	 * Preserves clipped bases in the sequence (soft clipping)
	 * @param clippingStart Number of bases to clip in the left part of the alignment
	 * @param clippingEnd Number of bases to clip in the right part of the alignment 
	 */
	public void clipAlignment(int clippingStart,int clippingEnd) {
		List<Integer> eventIndexes = ReadAlignment.getCigarEventIndexes(cigar);
		int l = eventIndexes.size();
		int [] readLengths = new int[l];
		boolean [] saveEvent = new boolean[l];
		char firstEvent=0,lastEvent=0;
		int lastEventIndex = -1;
		//Translate cigar
		for(int k=0;k<l;k++) {
			int i = eventIndexes.get(k);
			char event = cigar.charAt(i);
			int eventLength = Integer.parseInt(cigar.substring(lastEventIndex+1, i));
			lastEventIndex = i;
			readLengths[k] =  ReadAlignment.getNextReadPos(0, event, eventLength);
			saveEvent[k] = true;
			if(k==0) firstEvent = event;
			if(k==l-1) lastEvent = event;
		}
		//Modify forward
		int firstReadClippedPos = clippingStart;
		int ignoreLength = clippingStart;
		int k=0;
		if(firstEvent=='S') {
			readLengths[0]+=clippingStart;
			firstReadClippedPos = readLengths[0];
			k++;
		}
		for(;k<l && ignoreLength > 0 ;k++) {
			if(ignoreLength<readLengths[k]) {
				readLengths[k] -= ignoreLength;
				ignoreLength = 0;
				saveEvent[k] = true;
			} else {
				ignoreLength -= readLengths[k];
				saveEvent[k] = false;
				if(ignoreLength == 0) {
					for(k++;k<l && readLengths[k]==0;k++) saveEvent[k] = false;
				}
			}
		}
		//Modify backward
		ignoreLength = clippingEnd;
		k=l-1;
		if(lastEvent == 'S') {
			readLengths[k]+=clippingEnd;
			saveEvent[k] = true;
			k--;	
		}
		for(;k>=0 && ignoreLength > 0;k--) {
			if(ignoreLength<readLengths[k]) {
				readLengths[k] -= ignoreLength;
				ignoreLength = 0;
				saveEvent[k] = true;
			} else {
				ignoreLength -= readLengths[k];
				saveEvent[k] = false;
				if(ignoreLength == 0) {
					for(k--;k>=0 && readLengths[k]==0;k--) saveEvent[k] = false;
				}
			}
		}
		//Build final cigar
		StringBuffer newCigar = new StringBuffer();
		lastEventIndex = -1;
		
		//Translate cigar
		if(firstEvent!='S' && clippingStart > 0) {
			newCigar.append(""+clippingStart+"S");
		}
		for(k=0;k<l;k++) {
			int i = eventIndexes.get(k);
			char event = cigar.charAt(i);
			int eventLength = Integer.parseInt(cigar.substring(lastEventIndex+1, i));
			lastEventIndex = i;
			if(saveEvent[k]) {
				if(readLengths[k] > 0 && readLengths[k]!= eventLength) {
					newCigar.append(""+readLengths[k]+""+event);
				} else {
					newCigar.append(""+eventLength+""+event);
				}
			}
		}
		if(lastEvent!='S' && clippingEnd > 0) {
			newCigar.append(""+clippingEnd+"S");
		}
		int newStart = -1;
		for (;newStart<0;firstReadClippedPos++) {
			newStart = getSequencePosition(firstReadClippedPos); 
		}
		start = newStart;
		cigar = newCigar.toString();
		updateEnd();
	}
	/**
	 * Gets the next position in the sequence given a start position a cigar event and its length
	 * @param seqPos Position where the event happens
	 * @param event Type of event
	 * @param length Length of the event
	 * @return int Next position. seqPos + length if the event spans bases in the reference or seqPos otherwise   
	 */
	public static int getNextSeqPos(int seqPos, char event, int length) {
		if(event=='D' || event=='M' || event == 'N' ) {
			return seqPos + length;
		}
		return seqPos;
	}
	/**
	 * Gets the next position in the read given a start position a cigar event and its length
	 * @param readPos Position where the event happens
	 * @param event Type of event
	 * @param length Length of the event
	 * @return int Next position. readPos + length if the event spans bases in the read or readPos otherwise
	 */
    public static int getNextReadPos(int readPos, char event, int length) {
    	if(event=='I' || event=='M' || event == 'S' ) {
			return (short)(readPos + length);
		}
		return readPos;
	}
    /**
     * Calculates a list of indexes in the cigar where the events are located
     * @param cigar Alignment descriptor to be processed
     * @return List<Integer> Indexes in cigar of the events
     */
	public static List<Integer> getCigarEventIndexes(String cigar) {
		List<Integer> indexes = new ArrayList<Integer>();
		if(DEFAULT_SEQ_NAME.equals(cigar)) {
			return indexes;
		}
		for(int i=0;i<cigar.length();i++) {
			if(!Character.isDigit(cigar.charAt(i))) {
				indexes.add(i);
			}
		}
		return indexes;
	}
	private void updateEnd() {
		if(DEFAULT_SEQ_NAME.equals(cigar)) {
			end = 0;
		}
		int seqPos = start;
		List<Integer> eventIndexes = getCigarEventIndexes(cigar);
		int lastEventIndex = -1;
		for(int i:eventIndexes) {
			char event = cigar.charAt(i);
			int eventLength = Integer.parseInt(cigar.substring(lastEventIndex+1, i));
			lastEventIndex = i;
			seqPos = getNextSeqPos(seqPos, event, eventLength);
		}
		end = seqPos-1;
	}
	@Override
	public int length() {
		return end - start + 1;
	}

    /**
     * Calculates a the index of the first aligned (not soft clipped)
     * base in a read
     */        public int getIndexOfFirstAlignedBase() {
            int indexOfFirstAlignedBase;
            List<Integer> eventIndexes = getCigarEventIndexes(cigar);
            if (cigar.charAt(eventIndexes.get(0)) == 'S')
                 indexOfFirstAlignedBase = Integer.parseInt(cigar.substring(0, eventIndexes.get(0)));
            else
                indexOfFirstAlignedBase = 0;

            return indexOfFirstAlignedBase;

        }
	
     /**
     * Calculates a the index of the last aligned (not soft clipped)
     * base in a read
     */
        public int getIndexOfLastAlignedBase() {
            int indexOfLastAlignedBase;
            List<Integer> eventIndexes = getCigarEventIndexes(cigar);
                int lastEvent = eventIndexes.size() - 1;
//		  System.out.println(cigar);
//		  System.out.println(eventIndexes.get(lastEvent)-1);
//		  System.out.println(eventIndexes.get(lastEvent-1)+1);
                if (cigar.charAt(eventIndexes.get(lastEvent)) == 'S')
                    indexOfLastAlignedBase = read.length() - 1 - Integer.parseInt(cigar.substring(eventIndexes.get(lastEvent-1)+1,eventIndexes.get(lastEvent)));
                else
                    indexOfLastAlignedBase = read.length() - 1;

            return indexOfLastAlignedBase;

        }

     /**
     * Computes the validity of an alignment based on its number of aligned (not soft clipped)
     * bases compared to some input thresholds
     * @param int lower threshold of number of aligned bases
     * @param double upper threshold for the percentage of soft clipped bases to the total read length
     * @return true if the alignment is valid and false otherwise
     */
        public boolean validAlignment(int mappableBasesThreshold,double clippingThreshold) {
            int readLength= read.length();
            int softClippedBasesCount=0;
   	    int i = 0;
	    int n = cigar.length();
            do {
			int k = i;
			while (i < n && Character.isDigit(cigar.charAt(i))) {
				++i;
			}
			if (i == n) {
				throw new IllegalArgumentException(
						"Malformed CIGAR string: " + cigar);
			}
	                //System.out.println("k = "+k);
	                //System.out.println("i - 1 = "+(i-1));
			int elemLen = Integer.parseInt(cigar.substring(k,i), 10);
			char cigarCode = cigar.charAt(i);
			if (cigarCode == 'S')
                                softClippedBasesCount+=elemLen;
			++i;
		} while (i < n);
// 	     System.out.println("readLength "+readLength+" softClippedBasesCount "+softClippedBasesCount);
//	     System.out.println("ratio "+(((double)softClippedBasesCount)/(double)readLength)+" clippingThreshold "+clippingThreshold);
            if ((readLength - softClippedBasesCount) < mappableBasesThreshold || ((double)softClippedBasesCount)/(double)readLength > clippingThreshold) {
//		  System.out.println("false" );
                return false;
	     }
            else {
//		  System.out.println("true" );
                return true;

	    }        
    }

}
