/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.readalignments.filehandlers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import edu.uconn.ngstools.readalignments.AlignmentsTranslator;
import edu.uconn.ngstools.readalignments.CommonSpanningAlignmentComparator;
import edu.uconn.ngstools.readalignments.Read;
import edu.uconn.ngstools.readalignments.ReadAlignment;

/**
 * Handles io operations for Alignments in SAM format
 * @author Jorge Duitama
 *
 */
public class SAMReadAlignmentsHandler {
	private List<String> sequenceNames;
	private AlignmentsTranslator alignmentsTranslator;
	private boolean loadNotMapped=false;
	private boolean multipleAlnsPerRead = false;
	/**
	 * Loads all read alignments from a SAM file with the given name. If an alignments
	 * translator is set up, alignments are translated before being added to the final list 
	 * @param filename Name of the file to load
	 * @return List<ReadAlignment> List of read alignments present in the file
	 * @throws IOException If the file can not be read
	 */
	public List<ReadAlignment> loadReadAlignments(String filename) throws IOException {
		CommonSpanningAlignmentComparator alnComp = new CommonSpanningAlignmentComparator();
		List<ReadAlignment> alns = new ArrayList<ReadAlignment>();
		FileInputStream fis = new FileInputStream(filename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		ReadAlignment aln = createReadAlignment(in);
		List<ReadAlignment> lastAlns = new ArrayList<ReadAlignment>();
		for(int i=0;aln!=null;i++) {
			if(!multipleAlnsPerRead) alns.add(aln);
			else if(lastAlns.size()>0 && !lastAlns.get(0).getRead().getId().equals(aln.getRead().getId())) {
				alns.addAll(lastAlns);
				lastAlns.clear();
				lastAlns.add(aln);
			} else {
				boolean add = true;
				for(ReadAlignment aln2:lastAlns) {
					if(alnComp.compare(aln, aln2)==0) {
						add=false;
						break;
					}
				}
				if(add) lastAlns.add(aln);
			}
			aln = createReadAlignment(in);
		}
		if(multipleAlnsPerRead) alns.addAll(lastAlns);
		fis.close();
		return alns;
	}
	private String lastSeqId=null;
	/**
	 * Creates a read alignment using the information contained in the given line
	 * @param line String with the alignment information in SAM format
	 * @return ReadAlignment Object with the alignment information or null if the
	 * line is empty, corresponds to a comment or the read is not mapped and
	 * the flag to keep unmapped reads is not turned on
	 */
	public ReadAlignment createReadAlignment(String line) {
		if (line == null)
			return null;
		if(line.charAt(0)=='@')  {
			return null;
		}
		String [] items = line.split("\t");
//temp fix for testing existing sam files with spaces; changed back to \t and script for inserting /1 and /2  cahnged to sep by tabs
//		String [] items = line.split(" ");
		if(!loadNotMapped && ReadAlignment.DEFAULT_SEQ_NAME.equals(items[2])) {
			return null;
		}
		String id = items[0];
		Read r = new Read(id,items[9].toUpperCase(),items[10]);
		if(!items[2].equalsIgnoreCase(lastSeqId)) {
			lastSeqId = items[2];
		}
		ReadAlignment aln = new ReadAlignment(r,lastSeqId,Integer.parseInt(items[3]),items[5],Short.parseShort(items[4]));
		aln.setFlags(Short.parseShort(items[1]));
		if(!ReadAlignment.DEFAULT_SEQ_NAME.equals(items[6])) {
			aln.setMateSequenceName(items[6]);
			aln.setMatePosition(Integer.parseInt(items[7]));
			aln.setInsertSize(Integer.parseInt(items[8]));
		}
		if(alignmentsTranslator!=null) {
			alignmentsTranslator.translateAlignment(aln);
// to be removed when conversion for hardmerge is removed
			if (aln == null)
				return null;
			if(!loadNotMapped && ReadAlignment.DEFAULT_SEQ_NAME.equals(aln.getSequenceName())) {
				return null;
			}
		}
		return aln;
	}
	/**
	 * Creates a read alignment with the next line of the given buffer which contains information
	 * to create an alignment
	 * @param in Reader to get the next alignment information from
	 * @return ReadAlignment Object with the alignment information or null if there are no more lines 
	 * in the file
	 * @throws IOException If the Reader can not be read
	 */
	public ReadAlignment createReadAlignment(BufferedReader in) throws IOException {
		ReadAlignment aln = null;
		String line = "";
		while(line!=null && aln==null) {
			line = in.readLine();
			if(line!=null) {
				try {
					aln = createReadAlignment(line);
				} catch(Exception e) {
					System.err.println("WARN: Format error for line: "+line+". Message: "+e.getMessage());
					e.printStackTrace();
				}
			}
		}
		return aln;
	}
	/**
	 * Writes the given alignments in the given stream
	 * @param alns Alignments to save
	 * @param out Stream to write the alignments information in SAM format
	 */
	public void saveReadAlignments(Collection<ReadAlignment> alns, PrintStream out) {
		for(ReadAlignment aln: alns) {
			saveReadAlignment(aln, out);
		}
		out.flush();
	}


	/**
	 * Writes the information of the given alignment in the given stream
	 * @param aln Alignment to save
	 * @param out Stream to write the alignment information in SAM format
	 */
	public void saveReadAlignment(ReadAlignment aln, PrintStream out) {
		Read r = aln.getRead();
		out.print(r.getId()+"\t"+aln.getFlags()+"\t"+aln.getSequenceName()+"\t"+aln.getStart()+"\t"+aln.getMappingQuality());
		out.println("\t"+aln.getCigar()+"\t"+aln.getMateSequenceName()+"\t"+aln.getMatePosition()+"\t"+aln.getInsertSize()+"\t"+r.toString()+"\t"+r.getQualityScores());
	}

        /**
	 * Writes the given alignments in the given stream
	 * @param alns Alignments to save
	 * @param out Stream to write the alignments information in SAM format
	 */
	public void saveReadsFromAlignments(Collection<ReadAlignment> alns, PrintStream out) {
		for(ReadAlignment aln: alns) {
			saveReadFromAlignment(aln, out);
		}
		out.flush();
	}
	/**
	 * Writes the information of the given alignment in the given stream
	 * @param aln Alignment to save
	 * @param out Stream to write the alignment information in SAM format
	 */
	public void saveReadFromAlignment(ReadAlignment aln, PrintStream out) {
		Read r = aln.getRead();
		out.println("@"+r.getId());
                out.println(r.toString());
                out.println("+");
                out.println(r.getQualityScores());
	}
	/**
	 * Copy the sequence headers from a SAM file to the given stream
	 * @param filename Name of the SAM file to get the sequence headers
	 * @param out Stream to write the headers
	 * @throws IOException If the file can not be read
	 */
	public static void copySequenceHeaders(String filename, PrintStream out) throws IOException {
		FileInputStream fis = new FileInputStream(filename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		while(line!=null) {
			if(line.trim().length()>0 && !line.startsWith("@")) {
				break;
			}
			if(line.startsWith("@SQ")) {
				out.println(line);
			}
			line=in.readLine();
		}
		fis.close();
	}
	/**
	 * Loads the sequence names from the SQ headers of the given SAM file
	 * @param filename Name of the SAM file to get the sequence headers
	 * @throws IOException If the file can not be read
	 */
	public void loadSequenceNames(String filename) throws IOException {
		sequenceNames = new ArrayList<String>();
		FileInputStream fis = new FileInputStream(filename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		while(line!=null) {
			if(line.trim().length()>0 && !line.startsWith("@")) {
				break;
			}
			if(line.startsWith("@SQ")) {
				String [] items = line.split("\t");
				for(int i=0;i<items.length;i++) {
					if(items[i].startsWith("SN:")) {
						sequenceNames.add(items[i].substring(3));
						break;
					}
				}
			}
			line=in.readLine();
		}
		fis.close();
	}
	/**
	 * @return List<String> Names of the sequences
	 */
	public List<String> getSequenceNames() {
		return sequenceNames;
	}
	/**
	 * Changes the configures sequence names
	 * @param sequenceNames New Sequence names
	 */
	public void setSequenceNames(List<String> sequenceNames) {
		this.sequenceNames = sequenceNames;
	}
	/**
	 * @return AlignmentsTranslator Translator for loaded alignments. Null if no translator is set
	 */
	public AlignmentsTranslator getAlignmentsTranslator() {
		return alignmentsTranslator;
	}
	/**
	 * Changes the translator for loaded alignments
	 * @param alignmentsTranslator new translator
	 */
	public void setAlignmentsTranslator(AlignmentsTranslator alignmentsTranslator) {
		this.alignmentsTranslator = alignmentsTranslator;
	}
	/**
	 * @return true if not mapped reads are loaded. false otherwise
	 */
	public boolean isLoadNotMapped() {
		return loadNotMapped;
	}
	/**
	 * Sets the loadNotMapped flag 
	 * @param loadNotMapped trfue if not mapped reads should be loaded. false otherwise
	 */
	public void setLoadNotMapped(boolean loadNotMapped) {
		this.loadNotMapped = loadNotMapped;
	}
	/**
	 * @return boolean true if the files to load contain multiple aligns per read, false otherwise
	 */
	public boolean isMultipleAlnsPerRead() {
		return multipleAlnsPerRead;
	}
	/**
	 * Sets the multipleAlnsPerRead flag
	 * @param multipleAlnsPerRead true if the files to load will contain multiple aligns per read, 
	 * false otherwise
	 */
	public void setMultipleAlnsPerRead(boolean multipleAlnsPerRead) {
		this.multipleAlnsPerRead = multipleAlnsPerRead;
	}
}
