/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.sequences;

import java.util.Arrays;
import java.util.Random;

/**
 * Implements the Sequence interface minimizing memory consumption no matter 
 * which is the underlying alphabet
 * @author Jorge Duitama
 *
 */
public abstract class AbstractSequence implements Sequence {
	private static Random random = new Random();
	int [] sequence= new int [0];
	private byte lastHashSize=0;
	
	/**
	 * Returns the segment of this between the given indexes.
	 * @param beginIndex Zero based beginning index. Inclusive. 0<= beginIndex < this.length()
	 * @param endIndex Zero based ending index. Exclusive. 0<= endIndex <= this.length()
	 * @return String Subsequence between the given indexes as a String
	 * @see java.lang.String
	 */
	public String substring(int beginIndex, int endIndex) {
		int l = length();
		if(beginIndex < 0 || beginIndex >= l) {
			throw new StringIndexOutOfBoundsException(beginIndex);
		}
		if(endIndex<0 || endIndex > l) {
			throw new StringIndexOutOfBoundsException(endIndex);
		}
		int maxHashSize = getMaxHashSize();
		int relStart = beginIndex/maxHashSize;
		int relEnd = endIndex/maxHashSize;
		if (endIndex % maxHashSize > 0) {
			relEnd++;
		}
		StringBuffer answer = new StringBuffer();
		for(int i=relStart;i<relEnd;i++) {
			int hashSize = maxHashSize;
			if(i == sequence.length-1) {
				 hashSize = this.lastHashSize;
			} 
			String s = String.valueOf(getSequence(sequence[i],hashSize));
			int firstPos = maxHashSize*i;
			int endPos = maxHashSize*i + hashSize ;
			int firstSPos = 0;
			int endSPos = s.length();
			if(firstPos < beginIndex) {
				firstSPos = (beginIndex-firstPos);
			}
			if(endPos > endIndex) {
				endSPos = endIndex-firstPos; 
			}
			answer.append(s.substring(firstSPos, endSPos));
		}
		return answer.toString();
	}
	/**
	 * Makes this sequence equals to the given String
	 * @param sequence Sequence with data to copy
	 */
	public CharSequence subSequence (int start, int end) {
		return substring(start, end);
	}
	/**
	 * Returns the character at the given position
	 * @param position Zero based position to look for
	 * @return char Character at the given position
	 */
	public char charAt(int position) {
		String s = substring(position,position+1);
		return s.charAt(0);
	}
	/**
	 * Sets the given character at the given index
	 * @param index Zero based position to modify
	 * @param ch Character to set at index
	 */
	public void setCharAt(int position, char c) {
		if(getAlphabetIndex(c)<0) return;
		if(position < 0 || position>=this.length()) throw new StringIndexOutOfBoundsException(position);
		int maxHashSize = getMaxHashSize();
		int relPos = position/maxHashSize;
		int subPos = position%maxHashSize;
		int size;
		if(relPos < sequence.length-1) {
			size = maxHashSize;
		} else {
			size = this.lastHashSize;
		}
		char subSeq [] = getSequence(sequence[relPos], size);
		if(subSeq[subPos] != c) {
			subSeq[subPos] = c;
			int hash = getHash(String.valueOf(subSeq), 0, subSeq.length);
			sequence[relPos]=hash;
		}
	}
	@Override
	public String toString() {
		return substring(0, length());
	}
	/**
	 * Makes this sequence equals to the given String
	 * @param sequence Sequence with data to copy
	 */
	public void setSequence(String sequence) {
		this.sequence = new int [0];
		this.append(sequence);
	}
	/**
	 * Append the given String at the end of this
	 * @param sequence String to append
	 */
	public void append(String sequence) {
		if(sequence.length()==0) return;
		int newLength = length()+sequence.length();
		int maxHashSize = getMaxHashSize();
		int nHashNumbers = newLength/maxHashSize;
		if(newLength%maxHashSize>0) nHashNumbers++;
		
		//Generate junction number
		StringBuffer junction = new StringBuffer();
		if(lastHashSize>0 && lastHashSize < maxHashSize) {
			junction.append(getSequence(this.sequence[this.sequence.length-1], lastHashSize));
		}
		
		int firstPosAppend = 0;
		if(nHashNumbers == this.sequence.length) {
			junction.append(sequence);
			this.sequence[this.sequence.length-1] = getHash(junction.toString(), 0, junction.length());
			this.lastHashSize = (byte)junction.length();
			return;
		} else if (junction.length()>0) {
			firstPosAppend = maxHashSize-junction.length();
			junction.append(sequence.substring(0, firstPosAppend));
		}
		int [] newSequence = Arrays.copyOf(this.sequence, nHashNumbers);
		if(junction.length()>0) {
			newSequence[this.sequence.length-1] = getHash(junction.toString(), 0, junction.length());
		}
		
		int j = firstPosAppend;
		for(int i=this.sequence.length;i<nHashNumbers-1;i++) {
			newSequence[i]= getHash(sequence,j,j+maxHashSize);
			j+=maxHashSize;
		}
		int lastStart = j;
		this.lastHashSize = (byte)(sequence.length() - lastStart);
		newSequence[nHashNumbers-1]= getHash(sequence,lastStart,sequence.length());
		this.sequence = newSequence;
	}
	@Override
	public int length() {
		if(sequence.length==0) return 0;
		return (sequence.length-1)*getMaxHashSize()+lastHashSize;
	}
	/**
	 * Return the maximum number of characters encoded in a single number
	 * @return
	 */
	protected int getMaxHashSize() {
		return 32/getBitsPerCharacter();
	}
	/**
	 * Returns the number corresponding with a suitable size substring of the given sequence 
	 * @param seq Sequence to calculate hash
	 * @param start Zero based first position
	 * @param end Zero based last position
	 * @return int Numnber representing the substring of seq between first and last, both included
	 */
	private int getHash(String seq, int start, int end) {
		long number =0;
		int alpSize = getAlphabetSize();
		for(int i=start;i<end;i++) {
			number*=alpSize;
			int index = getAlphabetIndex(seq.charAt(i));
			if(index <0) {
				index = random.nextInt(alpSize);
			}	
			number+=index;
		}
		return (int)(number+Integer.MIN_VALUE);
	}
	/**
	 * Gets the sequence corresponding withb the given hash and the given size
	 * @param number Hash number to decode
	 * @param size Size of the sequence to return
	 * @return char[] Decoded String as a char array
	 */
	private char [] getSequence(int number, int size) {
		char [] answer = new char[size];
		int alpSize = getAlphabetSize();
		long absoluteNumber = (long)number-(long)Integer.MIN_VALUE;
		for(int i=0;i<size;i++) {
			int nextDigit = (int)(absoluteNumber%alpSize);
			int index = size-i-1;
			answer[index] = getAlphabetCharacter(nextDigit);
			absoluteNumber = absoluteNumber/alpSize;
		}
		return answer;
	}
	/**
	 * Returns the character of the alphabet with the given index in the alphabet String.
	 * @param index Zero based index to look for in the alphabet
	 * @return char Character at the given index
	 */
	public char getAlphabetCharacter(int index) {
		String alp = getAlphabet();
		if(index>=0 && index<alp.length()) {
			return alp.charAt(index);
		}
		return 0;
	}
	/**
	 * Gets the index for the given character in the alphabet.
	 * Default implementation performs a linear search on the alphabet
	 * @param ch character to look for
	 * @return in Zero based index or -1 if the character is not in the alphabet
	 */
	public int getAlphabetIndex(char ch) {
		String alp = getAlphabet();
		return alp.indexOf(ch);
	}
	/**
	 * Tells if the character is in the alphabet
	 * @param ch character to look for
	 * @return boolean true if ch is in the alphabet, false otherwise
	 */
	public boolean isInAlphabet(char ch) {
		return getAlphabetIndex(ch)>=0;
	}
	/**
	 * Returns the size of the alphabet
	 * @return int Size of the alphabet
	 */
	public int getAlphabetSize() {
		return getAlphabet().length();
	}
	/**
	 * Gets the minimum number of bits needed to represent uniquely every character of
	 * the alphabet 
	 * @return int minimum number of bits needed to represent uniquely every character of
	 * the alphabet
	 */
	protected int getBitsPerCharacter() {
		return (int)(Math.log10(getAlphabetSize())/Math.log10(2))+1;
	}
	
}
