/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.sequences.filehandlers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import edu.uconn.ngstools.sequences.DNAMaskedSequence;
import edu.uconn.ngstools.sequences.Sequence;

/**
 * General handler for sequences in fasta format. Implements methods for load sequences and
 * retrieve information about them. Keeps sequences in memory using the Sequence class to
 * optimize space
 * @author Jorge Duitama
 *
 */
public class FastaSequencesHandler {
	private List<String> ids = new ArrayList<String>();
	private List<Sequence> sequences = new ArrayList<Sequence>();
	private Class<?> sequenceType = DNAMaskedSequence.class;
	/**
	 * Loads the sequences present in the given filename
	 * @param filename Name of the fasta file where sequences must be loaded 
	 * @throws IOException If the file can not be read
	 */
	public void loadSequences(String filename) throws IOException {
		FileInputStream fis = new FileInputStream(filename);
		//Open buffer
		BufferedReader in = new BufferedReader(new InputStreamReader (fis));
		//Read sequences
		String line=in.readLine();
		StringBuilder buffer = new StringBuilder();
		Sequence nextSequence;
		try {
			nextSequence = (Sequence)sequenceType.newInstance();
		} catch (Exception e) {
			throw new IOException(e);
		}
		for(int i=0; line!=null;i++) {
			if(line.startsWith(">")) {
				if(buffer.length()>0) {
					nextSequence.append(buffer.toString());
					sequences.add(nextSequence);
					buffer.delete(0, buffer.length());
					try {
						nextSequence = (Sequence)sequenceType.newInstance();
					} catch (Exception e) {
						throw new IOException(e);
					}
				}
				String id = line.substring(1);
				ids.add(id);
				//System.err.println("Loading sequence: "+id);
				line = in.readLine();
			}
			buffer.append(line);
			if(buffer.length() > 10000000) {
				nextSequence.append(buffer.toString());
				buffer.delete(0, buffer.length());
			}
			line=in.readLine();
		}
		if(buffer.length()>0) {
			nextSequence.append(buffer.toString());
			sequences.add(nextSequence);
		}
		fis.close();
	}
	private String lastSearchId = null;
	private int lastPos=-1;
	/**
	 * Returns the position where the first sequence matching the given search key is located  
	 * @param searchId Id to search. An id matches searchId if id is equal to searchId
	 * (ignoring case) or if id contains searchId surrounded by the pipe character
	 * @return int position of the first sequence matching the given key or -1 if no sequence
	 * matches this id.
	 */
	public int getPosition(String searchId) {
		if(searchId.equals(lastSearchId)) {
			return lastPos;
		}
		lastSearchId = searchId;
		lastPos = -1;
		for(int i=0;i<ids.size();i++) {
			String id = ids.get(i);
			if(id.equalsIgnoreCase(searchId)) {
				lastPos = i;
				break;
			}
			if(id.contains("|"+searchId+"|")) {
				lastPos = i;
				break;
			}
		}
		return lastPos;
	}
	/**
	 * Returns the first sequence with an id matching the given search key  
	 * @param searchId Id to search. An id matches searchId if id is equal to searchId
	 * (ignoring case) or if id contains searchId surrounded by the pipe character
	 * @return String First sequence matching the given key
	 */
	public Sequence getSequence(String searchId) {
		int pos = getPosition(searchId);
		if(pos < 0) return null;
		return sequences.get(pos);
	}
	/**
	 * Returns the sequence id at the given position
	 * @param position Position to look for
	 * @return String Sequence id
	 */
	public String getId (int position) {
		return ids.get(position);
	}
	/**
	 * Returns the sequence at the given position
	 * @param position Position to look for
	 * @return Sequence Sequence at the given location
	 */
	public Sequence getSequence(int position) {
		return sequences.get(position);
	}
	/**
	 * @return int Number of sequences
	 */
	public int countSequences() {
		return ids.size();
	}
	/**
	 * @return Class<?> Type of the sequences that will be loaded
	 */
	public Class<?> getSequenceType() {
		return sequenceType;
	}
	/**
	 * New type for sequences that will be loaded
	 * @param sequenceType New sequence type
	 */
	public void setSequenceType(Class<?> sequenceType) {
		this.sequenceType = sequenceType;
	}
	/**
	 * Adds to the handler the sequence with the given inforamtion
	 * @param seqId Id of the new sequence
	 * @param seq New sequence
	 */
	public void addSequence(String seqId, Sequence seq) {
		ids.add(seqId);
		sequences.add(seq);
	}
	/**
	 * Dump all sequences in the given print stream
	 * @param out Stream to print the sequences
	 * @param lineLength Number of bases per line
	 */
	public void saveSequences(PrintStream out,int lineLength) {
		int i=0;
		for(Sequence seq: sequences) {
			out.println(">"+ids.get(i));
			for(int j=0;j<seq.length();j+=lineLength) {
				out.println(seq.substring(j, Math.min(seq.length(), j+lineLength)));
			}
			i++;
		}
	}
	/**
	 * @return List<String> List of sequence ids
	 */
	public List<String> getIds() {
		return ids;
	}
}
