/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.transcriptome;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import edu.uconn.ngstools.genome.Genome;
import edu.uconn.ngstools.genome.GenomicRegionPositionComparator;
import edu.uconn.ngstools.readalignments.AlignmentsTranslator;
import edu.uconn.ngstools.readalignments.Read;
import edu.uconn.ngstools.readalignments.ReadAlignment;
import edu.uconn.ngstools.sequences.DNASequence;
import edu.uconn.ngstools.sequences.Sequence;
import edu.uconn.ngstools.variants.SNV;

/**
 * Implements a transcriptome that handles sequence names and coordinates in a genomic fashion.
 * This calls is useful to transform coordinates relative to a transcript library to coordinates
 * relative to the corresponding genomic assembly. 
 * @author Jorge Duitama
 */
public abstract class AbstractTranscriptome implements Transcriptome, Genome, AlignmentsTranslator {
	public static final int MIN_ALIGNMENT_SIZE=4;
	Map<String, Gene> genesMap = new TreeMap<String, Gene>();
	//Transcripts indexed by id
	private Map<String, Transcript> transcriptsMap = new TreeMap<String, Transcript>();
	//Exons sorted by absolute position and indexed by chromosome
	private Map<String,List<Exon>> exonsMap = new TreeMap<String,List<Exon>>();
	//List of sequence names in the same order as they appear in the transcripts file
	private List<String> sequenceNames = new ArrayList<String>(); 
	//Offset for coordinates translation
	private int coordsTranslationOffset = 0;
	//Tells if exons must be sorted
	private boolean exonsSorted = true;
	//Data load helper methods
	@Override
	public void addGene(Gene g) {
		genesMap.put(g.getId(), g);
	}
	@Override
	public Gene getGene (String id) {
		return genesMap.get(id);
	}
	@Override
	public Exon getExon(Gene gene, int exonStart, int exonEnd) {
		for(Transcript t:gene.getTranscripts()) {
			for(Exon e:t.getExonsByAbsolute(exonStart, exonEnd)) {
				if(e.getStart()==exonStart && e.getEnd()==exonEnd) {
					return e;
				}
			}
		}
		return null;
	}
	@Override
	public void addExon(Exon exon) {
		List<Exon> chrExons = exonsMap.get(exon.getSequenceName());
		if(chrExons==null) {
			chrExons = new ArrayList<Exon>();
			exonsMap.put(exon.getSequenceName(), chrExons);
		}
		chrExons.add(exon);
		exonsSorted = false;
		
	}
	@Override
	public void addTranscript (Transcript t) {
		if(!sequenceNames.contains(t.getSequenceName())) {
			sequenceNames.add(t.getSequenceName());
		}
		if(transcriptsMap.get(t.getId()) != null) {
			//System.err.println("WARN: Repeated Transcript id: "+transcript.getId()+ ". Skipping");
		} else {
			transcriptsMap.put(t.getId(), t);
			t.getGene().addTranscript(t);
		}
	}
	/**
	 * Assigns the given sequence to the transcript with the given id
	 * @param transcriptId Id of the transcript to assign
	 * @param sequence Sequence to associate with the transcript
	 */
	public void fillTranscript(String transcriptId, Sequence sequence) {
		Transcript transcript = transcriptsMap.get(transcriptId);
		if(transcript!=null) {
			transcript.setCDNASequence(sequence);
		} else {
			System.err.println("WARN: Transcript not found with id: "+transcriptId);
		}
	}
	//Binary search for the exon that spans the desired position
	private int binarySearch(List<Exon> exons, int position) {
		int beginIndex = 0;
		int endIndex = exons.size();
		while(beginIndex<endIndex-1) {
			int midIndex = (beginIndex+endIndex) /2;
			Exon exon = exons.get(midIndex);
			if(exon.getStart()>position ) {
				endIndex = midIndex;
			} else {
				beginIndex = midIndex;
				if (exon.getStart()<= position && exon.getEnd()>=position) {
					break;
				}
			}	
		}
		//Go back a little due to overlapping exons
		for (int i=Math.max(0, beginIndex-5);i<beginIndex ;i++) {
			Exon exon = exons.get(i);
			if (exon.getStart()<= position && exon.getEnd()>=position) {
				return i;
			}
		}
		
		return beginIndex;
	}
	
	/**
	 * @return List<String> Names of the sequences in the transcriptome
	 */
	public List<String> getSequenceNames() {
		return sequenceNames;
	}
	@Override
	public List<Exon> getAllExons() {
		sortExons();
		List <Exon> allExons = new ArrayList<Exon>();
		for(List<Exon> chrExons: exonsMap.values()) {
			allExons.addAll(chrExons);
		}
		return allExons;
	}
	@Override
	public List<Exon> getExons(String sequenceName, int absolutePosition) {
		sortExons(); 
		List<Exon> exons = exonsMap.get(sequenceName);
		List<Exon> answer = new ArrayList<Exon>();
		if(exons!=null) {
			int index = binarySearch(exons, absolutePosition);
			for(int i=index;i<exons.size();i++) {
				Exon e = exons.get(i);
				if(e.getStart()<=absolutePosition && e.getEnd()>=absolutePosition) {
					answer.add(e);
				} else if(e.getStart()>absolutePosition) {
					break;
				}
			}
		}
		
		return answer;
	}
	@Override
	public List<Exon> getExons (String sequenceName, int absoluteStart, int absoluteEnd) {
		sortExons();
		List<Exon> exons = exonsMap.get(sequenceName);
		List <Exon> answer = new ArrayList<Exon>();
		if(exons!=null) {
			int index = binarySearch(exons, absoluteStart);
			for(int i=index;i<exons.size();i++) {
				Exon e = exons.get(i);
				if(e.getStart()<=absoluteEnd && e.getEnd()>=absoluteStart) {
					answer.add(e);
				} else if (absoluteEnd < e.getStart()) {
					break;
				}
			}
		}
		return answer;
	}
	private void sortExons() {
		if(!exonsSorted) {
			for(List<Exon> exons:exonsMap.values()) {
				Collections.sort(exons,GenomicRegionPositionComparator.getInstance());
			}
		}
		
		exonsSorted = true;
	}
	@Override
	public boolean isNonSynonymous(String seqName, int position) {
		List<Exon> posExons = getExons(seqName, position); 
		if(posExons.size()==0) {
			return false;
		}
		for(Exon e:posExons) {
			for(Transcript t:e.getGene().getTranscripts()) {
				int pos = t.getRelativeTranscriptPosition(position);
				if(pos>=0) {
					return pos%3!=2;
				}
			}
		}
		return false;
	}
	/**
	 * Cuts the transcript id from the id used in a fasta file
	 * @param sequenceName Name of the sequence in the fasta file
	 * @return String Transcript id in the map file
	 */
	protected String getTranscriptId(String sequenceName) {
		int index = sequenceName.indexOf('|');
		return sequenceName.substring(0,index);
	}
	/**
	 * Gets the transcript given the fasta sequence name.
	 * @param sequenceName Name of the transcript in the fasta file
	 * @return Transcript Object in the transcripts map corresponding with the given name
	 * or null if it is not found 
	 */
	protected Transcript getTranscriptBySeqName(String sequenceName) {
// Sahar the mRNA files Jorge creates have genome version and chr ID attached to transcript name while the ones used in isoem pipeline don't
              //String id = getTranscriptId(sequenceName);
		String id = new String(sequenceName);
//Sahar end
//		System.err.println("Input: seqName "+sequenceName+" "+sequenceName.length());
//		System.err.println("getTransscriptID "+getTranscriptId(sequenceName)+" "+getTranscriptId(sequenceName).length());
		return transcriptsMap.get(id);
	}
	/**
	 * Translate to genomic coordinates the given SNV, which is assumed to have
	 * transcript relative coordinates
	 * @param snv SNV to modify
	 */
	public void translateSNV (SNV snv) {
		int relPosSNP = snv.getPosition()-coordsTranslationOffset;
		Transcript t = getTranscriptBySeqName(snv.getSequenceName());
		if(t!=null) {
			snv.setSequenceName(t.getSequenceName());
			snv.setPosition(t.getAbsolutePosition(relPosSNP));
			//Flip alleles if the transcript is reverse
			if(t.isReverse()) {
				snv.setRefBase(DNASequence.getComplement(snv.getRefBase()));
				snv.setAltBase(DNASequence.getComplement(snv.getAltBase()));
			}
		}
	}
	/**
	 * @return int Offset of coordinates relative to the transcripts
	 */
	public int getCoordsTranslationOffset() {
		return coordsTranslationOffset;
	}
	/**
	 * Changes the offset for coordinates relative to the transcripts
	 * @param coordsTranslationOffset new offset
	 */
	public void setCoordsTranslationOffset(int coordsTranslationOffset) {
		this.coordsTranslationOffset = coordsTranslationOffset;
	}
	@Override
	public void translateAlignment (ReadAlignment aln) {
		int relAlnStart = aln.getStart() - coordsTranslationOffset;
		int relAlnEnd = aln.getEnd() - coordsTranslationOffset;
		Transcript t = getTranscriptBySeqName(aln.getSequenceName());
//sahar debugging exception
		if (t == null) {
// to be removed when conversion for hardmerge is removed
			System.err.println("Input: seq Name not found in transciptome map"+aln.getSequenceName()+" "+aln.getSequenceName().length());
			aln = null;
			return;
		}
		int clippingStart =0;
		if(relAlnStart<0) {
			clippingStart = -relAlnStart;
			relAlnStart =0;
		}
		int clippingEnd = 0;
		int tLength = t.length();
		if(relAlnEnd >=tLength) {
			clippingEnd = relAlnEnd-tLength+1;
			relAlnEnd = tLength-1;
		}
		
		List<Exon> exons = t.getExons(relAlnStart,relAlnEnd);
		int pos1 = t.getAbsolutePosition(relAlnStart);
		int pos2 = t.getAbsolutePosition(relAlnEnd);
		if(pos1<0 || pos2<0 || clippingStart + clippingEnd > aln.length()/3) {
			aln.setSequenceName(t.getSequenceName());
			aln.setStart(1);
			aln.setMappingQuality((short)0);
			return;
		}
		if(clippingStart>0 || clippingEnd>0) {
			System.err.println("WARN: Clipping CIGAR for alignment: "+aln.getRead().getId()+". Transcript: "+t.getId()+". Length: "+t.length()+". Last pos: "+relAlnEnd+". Exons: "+t.getExons().size());
			aln.clipAlignment(clippingStart, clippingEnd);
		}
		//System.out.println("RelAlnStart: "+relAlnStart+". transcriptRelPosS: "+transcriptRelPosS+". Pos1: "+pos1+". Pos2: "+pos2);
		int start = Math.min(pos1, pos2);
		String cigar = aln.getCigar();
		Read r = aln.getRead();
		//If t is in the negative strand, create reverse complement read and reverse cigar and exons list
		if(t.isReverse()) {
			r = aln.getRead().createReverseComplement();
			cigar = ReadAlignment.getReverseCigar(aln.getCigar());
			Collections.reverse(exons);
			//Reverse flag
			if((aln.getFlags()/16)%2==1) aln.setFlags((short)(aln.getFlags()-16));
			else aln.setFlags((short)(aln.getFlags()+16));
		}
		cigar = getSplicedCigar(cigar,exons,start);
		aln.setRead(r);
		aln.setSequenceName(t.getSequenceName());
		aln.setStart(start);
		aln.setCigar(cigar);
		//Mate pair info
		if(aln.getFlags()%2==1) {
			if (ReadAlignment.DEFAULT_SEQ_NAME.equals(aln.getMateSequenceName())) {
				t=null;
			} else if(!ReadAlignment.EQUAL_SEQ_NAME.equals(aln.getMateSequenceName())) {
				t = getTranscriptBySeqName(aln.getMateSequenceName());
			}
			int mateAbsPos = t.getAbsolutePosition(aln.getMatePosition()-coordsTranslationOffset);
			if(mateAbsPos<0) {
				t=null;
			}
			if(t!=null) {
				aln.setMateSequenceName(t.getSequenceName());
				aln.setMatePosition(mateAbsPos);
				if(t.isReverse()) {
					if((aln.getFlags()/32)%2==1) aln.setFlags((short)(aln.getFlags()-32));
					else aln.setFlags((short)(aln.getFlags()+32));
				}
			} else {
				aln.setMateSequenceName(ReadAlignment.DEFAULT_SEQ_NAME);
				aln.setMatePosition(0);
				aln.setInsertSize(0);
			}
		}
	}
	/**
	 * Calculates a spliced cigar assuming that the given cigar spans the given exons from the
	 * given start
	 * @param cigar Alignment descriptor relative to the given exons
	 * @param exons List of exons that the cigar spans
	 * @param absoluteStart First base of the alignment 
	 * @return String new cigar spliced to be consistent with genomic coordinates
	 */
	public String getSplicedCigar(String cigar, List<Exon> exons, int absoluteStart) {
		StringBuffer answer = new StringBuffer();
		List<Integer> eventIndexes = ReadAlignment.getCigarEventIndexes(cigar);
		int j = 0;
		int lastEventIndex = -1;
		int readPos = 0;
		int seqPos = absoluteStart;
		for(int k=0;k<eventIndexes.size();k++) {
			int i = eventIndexes.get(k);
			char event = cigar.charAt(i);
			int eventLength = Integer.parseInt(cigar.substring(lastEventIndex+1, i));
			lastEventIndex = i;
			int nextSeqPos = ReadAlignment.getNextSeqPos(seqPos, event, eventLength);
			int nextReadPos = ReadAlignment.getNextReadPos(readPos, event, eventLength);
			if(nextSeqPos > seqPos) {
				Exon e = exons.get(j);
				while(nextSeqPos > e.getEnd()) {
					//Split the event into the current exon and the next one (if any)
					int remExonLength = e.getEnd()-seqPos+1;
					answer.append(""+remExonLength+""+event);
					seqPos=e.getEnd()+1;
					j++;
					if(j<exons.size()) {
						Exon nextExon = exons.get(j);
						int intronLength = nextExon.getStart()-e.getEnd()-1;
						answer.append(""+intronLength+"N");
						nextSeqPos +=intronLength;
						e = nextExon;
						seqPos=e.getStart();
					} else {
						if(nextSeqPos > seqPos) {
							System.err.println("Check splice for reads in gene: "+e.getGene().getName()+" starting at: "+absoluteStart+". Exons:"+exons.size()+". Seqpos: "+seqPos+". NextSeqPos: "+nextSeqPos+". ReadPos:"+readPos+". NextReadPos:"+nextReadPos+". Event:"+event+". Event length: "+eventLength);
						}
						break;	
					}
				} 
				if(nextSeqPos>seqPos) {
					int nextLength = nextSeqPos - seqPos;
					answer.append(""+nextLength+""+event);
				}
			} else {
				answer.append(""+eventLength+""+event);
			}
			seqPos = nextSeqPos;
			readPos = nextReadPos;
		}
		return answer.toString();
	}
	@Override
	public char getReferenceBase (String seqName, int absolutePosition) {
		List<Exon> exons = getExons(seqName, absolutePosition); 
		if(exons.size()>0) {
			Gene g = exons.get(0).getGene();
			for(Transcript t:g.getTranscripts()) {
				char base = t.getReferenceBase(absolutePosition);
				if(base!=0) {
					return base;
				}
			}
		}
		return 0;
	}
	@Override
	public void setReferenceBase(String seqName, int absolutePosition, char base) {
		List<Exon> posExons = getExons(seqName, absolutePosition); 
		for(Exon e:posExons) {
			for(Transcript t: e.getGene().getTranscripts()) {
				t.setReferenceBase(absolutePosition, base);
			}
		}
	}
}
