/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.transcriptome;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import edu.uconn.ngstools.sequences.DNASequence;
import edu.uconn.ngstools.sequences.filehandlers.FastaSequencesHandler;
/**
 * Implementation of an AbstractTranscriptome that loads the transcripts information
 * from a CCDS file. Uses a configurable prefix for sequence names to make them consistent
 * with other reference assemblies as the ones in UCSC 
 * @author Jorge Duitama
 *
 */
public class CCDSTranscriptome extends AbstractTranscriptome {
	//Object with reference sequences
	private FastaSequencesHandler sequencesHandler = new FastaSequencesHandler();
	//Prefix for sequenceNames
	private String seqNamePrefix = "chr";
	/**
	 * Creates a CCDSTranscriptome with the given map file
	 * @param ccdsFilename Name of the file with the transcriptome description
	 * @throws IOException If the file can not be read
	 */
	public CCDSTranscriptome(String ccdsFilename) throws IOException {
		loadMap(ccdsFilename);
	}
	/**
	 * Creates a CCDSTranscriptome with the given files
	 * @param ccdsFilename Name of the file with the transcriptome description
	 * @param cdnaFilename Name of the fasta file with the transcript sequences
	 * @throws IOException If either file can not be read
	 */
	public CCDSTranscriptome(String ccdsFilename, String cdnaFilename) throws IOException {
		loadMap(ccdsFilename);
		loadSequences(cdnaFilename);
	}
	/**
	 * Loads the genes, transcripts and exons information from the given map file
	 * @param filename Name of the file with the transcriptome description
	 * @throws IOException If the file can not be read
	 */
	public void loadMap(String filename) throws IOException {
		FileInputStream fis = new FileInputStream(filename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		
		String seqName = null;
		while(line!=null) {
			if(line.charAt(0)!='#') {
				String [] items = line.split("\t");
				boolean reverse = items[6].charAt(0)=='-';
				if(!items[5].startsWith("Withdrawn") && !items[7].equals("-") ) {
					Gene gene =  getGene(items[3]);
					if(gene == null) {
						String newSeqName = seqNamePrefix+items[0]; 
						if(!newSeqName.equals(seqName)) {
							seqName = newSeqName;
						}
						gene = new Gene(seqName,items[3],items[2]);
						addGene(gene);
					}
					//Add 1 to all starts and ends because CCDS positions are zero indexed
					
					String exonsStr = items[9].substring(1,items[9].length()-1);
					String [] items2 = exonsStr.split(", ");
					List<Exon> tExons = new ArrayList<Exon>();
					for(int i=0;i<items2.length;i++) {
						int index2 = items2[i].indexOf('-');
						int exonStart = Integer.parseInt(items2[i].substring(0,index2))+1;
						int exonEnd = Integer.parseInt(items2[i].substring(index2+1))+1;
						Exon exon=getExon(gene, exonStart, exonEnd);
						if(exon==null) {
							exon = new Exon(gene,exonStart,exonEnd);
							addExon(exon);
						}
						tExons.add(exon);
					}
					Transcript transcript = new Transcript(gene,items[4],items[5],Integer.parseInt(items[7])+1,Integer.parseInt(items[8])+1,reverse,tExons);
					addTranscript(transcript);
				}
			}
			line=in.readLine();
		}
		fis.close();
	}
	/**
	 * Loads the transcripts sequences from the given fasta file 
	 * @param filename Name of the fasta file with the transcript sequences
	 * @throws IOException If the file can not be read
	 */
	public void loadSequences(String filename) throws IOException {
		sequencesHandler.setSequenceType(DNASequence.class);
		sequencesHandler.loadSequences(filename);
		List <String> ids = sequencesHandler.getIds();
		for(int i=0;i <ids.size();i++ ) {
			String nextId = getTranscriptId(ids.get(i));
			fillTranscript(nextId, sequencesHandler.getSequence(i));
		}
	}
	@Override
	public void saveGenome(PrintStream out, int lineLength) {
		sequencesHandler.saveSequences(out, lineLength);
	}
}
