/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.transcriptome;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import edu.uconn.ngstools.genome.GenomicRegion;
import edu.uconn.ngstools.genome.GenomicRegionPositionComparator;
import edu.uconn.ngstools.sequences.DNASequence;
import edu.uconn.ngstools.sequences.Sequence;
/**
 * Information for a set of exons in the same gene making a valid transcript
 * @author Jorge Duitama
 */
public class Transcript implements GenomicRegion {
	private Gene gene;
	private String id;
	private String status;
	private int start;
	private int end; //Always greater or equal than start
	private boolean reverse;
	private Sequence cdnaSequence;
	private String proteinSequence;
	private List<Exon> exons=new ArrayList<Exon>();
	
	/**
	 * Creates a new transcript with the given information
	 * @param gene Gene where the transcript is located
	 * @param id Id of the transcript
	 * @param status Tells if the transcript has been confirmed
	 * @param start First position of the transcript in the sequence where the gene is located
	 * @param end Last position of the transcript in the sequence where the gene is located
	 * @param reverse Tells if the transcription is done in the forward or in the reverse strand
	 * @param exons List of exons making up the transcript
	 */
	public Transcript(Gene gene, String id, String status, int start, int end, boolean reverse, List<Exon> exons) {
		super();
		this.setGene(gene);
		this.setId(id);
		this.setStatus(status);
		this.setStart(start);
		this.setEnd(end);
		this.setReverse(reverse);
		this.setExons(exons);
	}
	/**
	 * Changes the list of exons making up the transcript
	 * @param exons New list of exons making up the transcript
	 */
	public void setExons(List<Exon> exons) {
		this.exons = new ArrayList<Exon>(exons);
		Collections.sort(this.exons,GenomicRegionPositionComparator.getInstance());
	}
	/**
	 * Calculates the position relative to the start of the transcript given the position relative to the sequence
	 * @param absolutePosition Position relative to the sequence where the transcript is located
	 * @return int Zero based position relative to the start of the transcript or -1 if the given 
	 * position does not belong to the transcript
	 */
	public int getRelativeTranscriptPosition (int absolutePosition) {
		int answer = 0;
		ArrayList<Exon> list = new ArrayList<Exon>(exons);
		if(reverse) {
			Collections.reverse(list);
		}
		for(Exon e:list) {
			if(reverse && e.getEnd()< absolutePosition) {
				return -1;
			} else if (!reverse && e.getStart()> absolutePosition) {
				return -1;
			}
			if(e.getStart()<=absolutePosition && e.getEnd()>=absolutePosition) {
				if(reverse) {
					answer+= (e.getEnd()-absolutePosition);
				} else {
					answer+= (absolutePosition-e.getStart());
				}
				return answer;
			}
			answer+= e.length();
		}
		return -1;
	}
	/**
	 * Calculates the position relative to the sequence given the position relative to the start of the transcript
	 * @param relativeTranscriptPosition Zero based Position relative to the start of the transcript
	 * @return int Position relative to the sequence or -1 if the given position is negative or larger
	 * or equal than the transcript length 
	 */
	public int getAbsolutePosition(int relativeTranscriptPosition) {
		ArrayList<Exon> list = new ArrayList<Exon>(exons);
		if(reverse) {
			Collections.reverse(list);
		}
		for(Exon e:list) {
			if(relativeTranscriptPosition<e.length()) {
				if(this.reverse) {
					return e.getEnd() - relativeTranscriptPosition;
				} else {
					return e.getStart() + relativeTranscriptPosition;
				}
			} else {
				relativeTranscriptPosition-=e.length();
			}
		}
		return -1;
	}
	/**
	 * Finds the reference base for the given position
	 * @param absolutePosition Position relative to the sequence where the transcript is located
	 * @return char Reference base in the given position or 0 if the given position
	 * does not belong to the transcript
	 */
	public char getReferenceBase (int absolutePosition) {
		int pos = getRelativeTranscriptPosition(absolutePosition);
		if( cdnaSequence!=null && pos>=0 && pos < length()) {
			char refBase = cdnaSequence.charAt(pos);
			if(reverse) {
				refBase = DNASequence.getComplement(refBase);
			}
			return refBase;
		}
		return 0;
	}
	/**
	 * Changes the reference base for the given position
	 * @param absolutePosition Position relative to the sequence where the transcript is located
	 * @param base New base to set in the given position. No changes are done if the given
	 * position does not belong to the transcript
	 */
	public void setReferenceBase(int absolutePosition, char base) {
		int pos = getRelativeTranscriptPosition(absolutePosition);
		if( cdnaSequence!=null && pos>=0 && pos < length()) {
			if(reverse) {
				base = DNASequence.getComplement(base);
			}
			cdnaSequence.setCharAt(pos, base);
		}
	}
	/**
	 * Returns the Exon spanning the given relative position 
	 * @param relativeTranscriptPosition Zero based position relative to the start of the transcript
	 * @return Exon spanning the given position or null if the given position is invalid
	 */
	public Exon getExon (int relativeTranscriptPosition) {
		if(relativeTranscriptPosition<0) return null;
		ArrayList<Exon> list = new ArrayList<Exon>(exons);
		if(reverse) {
			Collections.reverse(list);
		}
		for(Exon e:list) {
			if(relativeTranscriptPosition<e.length()) {
				return e;
			} else {
				relativeTranscriptPosition-=e.length();
			}
		}
		return null;
	}
	/**
	 * Returns the Exon spanning the given position 
	 * @param absolutePosition Position relative to the sequence where the transcript is located
	 * @return Exon spanning the given position or null if the given position does not belong 
	 * to the transcript
	 */
	public Exon getExonByAbsolutePosition (int absolutePosition) {
		for(Exon e:exons) {
			if(e.getStart()<=absolutePosition && e.getEnd()>=absolutePosition) {
				return e;
			}
		}
		return null;
	}
	/**
	 * Returns a list of exons overlapping the region delimited by the given start and end 
	 * @param absoluteStart First position relative to the sequence where the transcript is located
	 * @param absoluteEnd Last position relative to the sequence where the transcript is located
	 * @return List<Exon> Exons in the transcript spanning the interval defined by the given coordinates
	 */
	public List<Exon> getExonsByAbsolute(int absoluteStart, int absoluteEnd) {
		ArrayList<Exon> answer = new ArrayList<Exon>();
		for(Exon e:exons) {
			if(e.getStart()<=absoluteEnd && e.getEnd() >= absoluteStart ) {
				answer.add(e);
			}
		}
		return answer;
	}
	/**
	 * @return int Length of the transcript calculated as the sum of the exons making up the transcript
	 */
	public int length() {
		int length = 0;
		for(Exon e:exons) {
			length+=e.length();
		}
		return length;
	}
	/**
	 * Returns a list of exons overlapping the region delimited by the given start and end
	 * @param transcriptRelativeStart Zero based first position relative to the start of the transcript
	 * @param transcriptRelativeEnd Zero based last position relative to the start of the transcript
	 * @return List<Exon> Exons in the transcript spanning the interval defined by the given coordinates
	 */
	public List<Exon> getExons(int transcriptRelativeStart, int transcriptRelativeEnd) {
		ArrayList<Exon> list = new ArrayList<Exon>(exons);
		ArrayList<Exon> answer = new ArrayList<Exon>();
		if(reverse) {
			Collections.reverse(list);
		}
		int exonStart=0;
		for(Exon e:list) {
			int nextExonStart = exonStart+e.length();
			if(transcriptRelativeStart<nextExonStart && transcriptRelativeEnd >= exonStart ) {
				answer.add(e);
			} 
			exonStart=nextExonStart;
		}
		return answer;
	}
	/**
	 * @return String Name of the sequence where the transcript is mapped
	 */
	public String getSequenceName() {
		return gene.getSequenceName();
	}
	/**
	 * @return String Id of the transcript
	 */
	public String getId() {
		return id;
	}
	/**
	 * Changes the id of the transcript
	 * @param id New id
	 */
	public void setId(String id) {
		this.id = id;
	}
	/**
	 * @return String Status of the transcript
	 */
	public String getStatus() {
		return status;
	}
	/**
	 * Changes the status of the trasncript
	 * @param status New status of the transcript
	 */
	public void setStatus(String status) {
		this.status = status;
	}
	/**
	 * @return int First position of the transcript in the sequence where the gene is located
	 */
	public int getStart() {
		return start;
	}
	/**
	 * Changes the start of the trasncript
	 * @param start New start
	 */
	public void setStart(int start) {
		this.start = start;
	}
	/**
	 * @return int Last position of the transcript in the sequence where the gene is located
	 */
	public int getEnd() {
		return end;
	}
	/**
	 * Changes the end of the trasncript
	 * @param end New end
	 */
	public void setEnd(int end) {
		this.end = end;
	}
	/**
	 * @return boolean true if the transcript is transcribed from the reverse strand. False otherwise
	 */
	public boolean isReverse() {
		return reverse;
	}
	/**
	 * Changes the strand of the transcript
	 * @param reverse true if the transcript is transcribed from the reverse strand. False otherwise
	 */
	public void setReverse(boolean reverse) {
		this.reverse = reverse;
	}
	/**
	 * @return Sequence the sequence of the transcript as it is translated into a protein
	 */
	public Sequence getCDNASequence() {
		return cdnaSequence;
	}
	/**
	 * Changes the CDNA sequence
	 * @param sequence New sequence
	 */
	public void setCDNASequence(Sequence sequence) {
		cdnaSequence = sequence;
	}
	/**
	 * @return String the protein sequence after translation
	 */
	public String getProteinSequence() {
		return proteinSequence;
	}
	/**
	 * Changes the protein sequence
	 * @param proteinSequence New protein sequence
	 */
	public void setProteinSequence(String proteinSequence) {
		this.proteinSequence = proteinSequence;
	}
	/**
	 * Gets the translation to protein for the CDNA sequence of this transcript according with the trasnlator
	 * @param translator Object that translates CDNA sequences into protein sequences
	 * @return String Protein sequence after translation
	 */
	public String getProteinSequence(ProteinTranslator translator) {
		return translator.getProteinSequence(this.getCDNASequence().toString());
	}
	/**
	 * @return List<Exon> List of exons making up the transcript sorted by position in the sequence
	 */
	public List<Exon> getExons() {
		return exons;
	}
	/**
	 * @return Gene the gene to which the transcript belongs
	 */
	public Gene getGene() {
		return gene;
	}
	/**
	 * Changes the gene for this transcript
	 * @param gene New gene
	 */
	public void setGene(Gene gene) {
		this.gene = gene;
	}
}
