/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.transcriptome;

import java.util.List;
/**
 * Interface to describe functions performed by a Transcriptome
 * @author Jorge Duitama
 */
public interface Transcriptome {
	/**
	 * Adds the given gene to the transcriptome
	 * @param g New gene
	 */
	public void addGene(Gene g);
	/**
	 * Adds a new transcript to the transcriptome
	 * @param t New transcript
	 */
	public void addTranscript (Transcript t);
	/**
	 * Adds a new exon to the transcriptome
	 * @param exon New exon
	 */
	public void addExon(Exon exon);
	/**
	 * Retrieves the gene with the given id
	 * @param id to look for
	 * @return Gene with the given id or null if no gene is found
	 */
	public Gene getGene (String id);
	/**
	 * Retrieves the exon matching the given criteria
	 * @param gene Gene to which the exon belongs
	 * @param exonStart One based first position of the exon
	 * @param exonEnd One based last position of the exon
	 * @return Exon meeting the given criteria or null if no exon is found
	 */
	public Exon getExon(Gene gene, int exonStart, int exonEnd);
	/**
	 * @return List<Exon> List with all exons in the transcriptome
	 */
	public List<Exon> getAllExons();
	/**
	 * Retrieves all exons spanning a given genomic location
	 * @param seqName Name of the sequence to look for 
	 * @param absolutePosition One based position to look for
	 * @return List<Exon> Exons spanning the genomic coordinate defined by the input
	 * parameters
	 */
	public List<Exon> getExons(String seqName, int absolutePosition);
	/**
	 * Retrieves all exons spanning the region delimited by the given criteria
	 * @param seqName Name of the sequence to look for
	 * @param absoluteStart One based first position to look for
	 * @param absoluteEnd One based last position to look for
	 * @return List<Exon> Exons spanning the genomic region defined by the input
	 * parameters
	 */
	public List<Exon> getExons (String seqName, int absoluteStart, int absoluteEnd);
	/**
	 * Tells if the given genomic location is non synonymous according with the transcriptome
	 * @param seqName Name of the sequence to look for
	 * @param position One based position to look for
	 * @return boolean true if the genomic coordinate is non synonymous, false otherwise
	 */
	public boolean isNonSynonymous(String seqName, int position);
}
