/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.variants;

/**
 * Information for a SNV called for an individual
 * @author Jorge Duitama
 *
 */
public class CalledSNV extends SNV {
	private String genotype; //String of length two with the genotype
	private double countReference = 0;
	private double countAlternative = 0;
	private double probability = 1;
	private double refProbability=0;
	private double [] allCounts;
//stand specific counts (Sahar)
	private double pcountReference = 0;
	private double pcountAlternative = 0;
	private double [] pallCounts;
	private double ncountReference = 0;
	private double ncountAlternative = 0;
	private double [] nallCounts;
	
	/**
	 * Creates a new CalledSNV with the given information
	 * @param sequenceName Name of the sequence where the SNV is located
	 * @param absolutePosition Position in the sequence where the SNp is located
	 * @param refBase Base present in the reference sequence
	 * @param altBase Base representing the minor allele 
	 * @param genotype Sequence of alleles representing the called genotype for 
	 * an individual in the given coordinates
	 * @param countReference Number of times that the reference allele was observed
	 * @param countAlternative Number of times that the alternative allele was observed
	 * @param probability Probability that the given genotype is the true one for an individual
	 */
	public CalledSNV(String sequenceName, int position, char refBase,
			char altBase, String genotype, double countReference, double countAlternative, double probability) {
		super(sequenceName, position, refBase, altBase);
		this.setGenotype(genotype);
		this.setCountReference(countReference);
		this.setCountAlternative(countAlternative);
		this.setProbability(probability);
	}

	/**
	 * Creates a new CalledSNV with the given information
	 * @param sequenceId Id of the sequence where the SNV is located
	 * @param relativePosition Zero based position in the sequence
	 * @param refBase Base present in the reference sequence
	 * @param altBase Base representing the minor allele 
	 * @param genotype Sequence of alleles representing the called genotype for 
	 * an individual in the given coordinates
	 */
	public CalledSNV(String sequenceName, int position, char refBase, char altBase, String genotype) {
		super(sequenceName, position, refBase, altBase);
		this.setGenotype(genotype);
	}
	/**
	 * Returns the called genotype
	 * @return String called genotype as a sequence of alleles
	 */
	public String getGenotype() {
		return genotype;
	}

	/**
	 * Changes the called genotype
	 * @param genotype Sequence of alleles representing the called genotype for 
	 * an individual in the given coordinates
	 */
	public void setGenotype(String genotype) {
		this.genotype = genotype;
	}

	/**
	 * @return int Number of times that the reference allele was observed
	 */
	public double getCountReference() {
		return countReference;
	}

 //strand specific counts  (Sahar)

	/**
	 * @return int Number of times that the reference allele was observed on the +ve strand
	 */
	public double getpCountReference() {
		return pcountReference;
	}

	/**
	 * @return int Number of times that the reference allele was observed on the -ve strand
	 */
	public double getnCountReference() {
		return ncountReference;
	}

        /**
	 * Sets the number of times that the reference allele was observed
	 * @param countReference New number of times that the reference allele was observed
	 */
	public void setCountReference(double countReference) {
		this.countReference = countReference;
	}
//strand specific counts  (Sahar)

        /**
	 * Sets the number of times that the reference allele was observed on the +ve strand
	 * @param countReference New number of times that the reference allele was observed
	 */
	public void setpCountReference(double countReference) {
		this.pcountReference = countReference;
	}

        /**
	 * Sets the number of times that the reference allele was observed on the -ve strand
	 * @param countReference New number of times that the reference allele was observed
	 */
	public void setnCountReference(double countReference) {
		this.ncountReference = countReference;
	}


        /**
	 * @return int Number of times that the alternative allele was observed
	 */
	public double getCountAlternative() {
		return countAlternative;
	}
//strand specific counts  (Sahar)

        /**
	 * @return int Number of times that the alternative allele was observed on the +ve strand
	 */
	public double getpCountAlternative() {
		return pcountAlternative;
	}

        /**
	 * @return int Number of times that the alternative allele was observed on the -ve strand
	 */
	public double getnCountAlternative() {
		return ncountAlternative;
	}

	/**
	 * Sets the number of times that the alternative allele was observed
	 * @param countAlternative New number of times that the alternative allele was observed
	 */
	public void setCountAlternative(double countAlternative) {
		this.countAlternative = countAlternative;
	}
//strand specific counts  (Sahar)
	/**
	 * Sets the number of times that the alternative allele was observed on the +ve strand
	 * @param countAlternative New number of times that the alternative allele was observed
	 */
	public void setpCountAlternative(double countAlternative) {
		this.pcountAlternative = countAlternative;
	}

	/**
	 * Sets the number of times that the alternative allele was observed on the -ve strand
	 * @param countAlternative New number of times that the alternative allele was observed
	 */
	public void setnCountAlternative(double countAlternative) {
		this.ncountAlternative = countAlternative;
	}

	/**
	 * @return double Probability that this genotype is the true one for an individual
	 */
	public double getProbability() {
		return probability;
	}

	/**
	 * Sets the probability that this genotype is the true one for an individual
	 * @param probability New probability that this genotype is the true one for an individual
	 */
	public void setProbability(double probability) {
		this.probability = probability;
	}
	
	/**
	 * @return double Probability of having the reference base for this locus
	 */
	public double getRefProbability() {
		return refProbability;
	}

	/**
	 * Changes the probability of having the reference base for this locus
	 * @param refProbability New probability
	 */
	public void setRefProbability(double refProbability) {
		this.refProbability = refProbability;
	}

	/**
	 * @return double [] Number of times that each possible allele appears
	 */
	public double[] getAllCounts() {
		return allCounts;
	}

	/**
	 * @return double [] Number of times that each possible allele appears on the +ve strand
	 */
	public double[] getpAllCounts() {
		return pallCounts;
	}

	/**
	 * @return double [] Number of times that each possible allele appears on the -ve strand
	 */
	public double[] getnAllCounts() {
		return nallCounts;
	}


	/**
	 * @param allCounts New number of times that each possible allele appears
	 */
	public void setAllCounts(double[] allCounts) {
		this.allCounts = allCounts;
	}

	/**
	 * @param allCounts New number of times that each possible allele appears on +ve strand
	 */
	public void setpAllCounts(double[] allCounts) {
		this.pallCounts = allCounts;
	}

	/**
	 * @param allCounts New number of times that each possible allele appears on -ve strand
	 */
	public void setnAllCounts(double[] allCounts) {
		this.nallCounts = allCounts;
	}

	/**
	 * Tells if the genotype is homozygous
	 * @return boolean true if both alleles in the called genotype are equal, false otherwise
	 */
	public boolean isHomozygous() {
		return genotype!=null && genotype.length()>=2 && genotype.charAt(0) == genotype.charAt(1);
	}
	/**
	 * Tells if the genotype is homozygous for the reference allele
	 * @return boolean true if both alleles in the called genotype are equal and equal to the reference, false otherwise
	 */
	public boolean isHomozygousReference() {
		char upRef = Character.toUpperCase(this.getRefBase());
		return isHomozygous() && genotype.charAt(0) == upRef;
	}
	/**
	 * Tells if the genotype is heterozygous
	 * @return boolean true if both alleles in the called genotype are different, false otherwise
	 */
	public boolean isHeterozygous() {
		return genotype!=null && genotype.length()>=2 && genotype.charAt(0) != genotype.charAt(1);
	}
	/**
	 * Tells if the SNV is triallelic according with its genotype
	 * @return boolean true if both alleles in the called genotype are different each other and from the reference, false otherwise
	 */
	public boolean isTriallelic() {
		return isHeterozygous() && genotype.indexOf(getRefBase())<0;
	}

}
