/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.variants;

import edu.uconn.ngstools.sequences.DNASequence;
/**
 * Class that stores allele counts and genotype probabilities and offer methods to calculate 
 * posterior probabilities from conditionals 
 * @author Jorge Duitama
 */
public class CountsHelper {
	private double totalCount=0;
	private double [] counts;
	private double ptotalCount=0;
	private double [] pcounts;
	private double ntotalCount=0;
	private double [] ncounts;
	private double [][] logConditionalProbs;
	private double pAllele1=0.5;
	private String alphabeth = new DNASequence().getAlphabet();
	/**
	 * Creates a default counts helper
	 */
	public CountsHelper () {
		int bases = alphabeth.length();
		counts = new double [bases];
		pcounts = new double [bases];
		ncounts = new double [bases];
		logConditionalProbs = new double [bases][bases];
		startCounts();
	}
	/**
	 * Starts all counts to zero
	 */
	public void startCounts () {
		totalCount=0;
		ptotalCount=0;
		ntotalCount=0;
		for(int i=0;i<logConditionalProbs.length;i++) {
			counts[i] = 0;
			pcounts[i] = 0;
			ncounts[i] = 0;
			for(int j=0;j<logConditionalProbs[0].length;j++) {
				logConditionalProbs[i][j] = 0;
			}
		}
	}
	/**
	 * Start counts based on the given heterozygosity rate
	 * @param hetRate Prior probability of a heterozygous genotype
	 */
	public void startCounts( double hetRate) {
		startCounts();
		double logHetero = Math.log10(hetRate/6);
		double logHomo = Math.log10((1-hetRate)/4);
		for(int i=0;i<logConditionalProbs.length;i++) {
			logConditionalProbs[i][i] += logHomo;
			for(int j=0;j<logConditionalProbs[0].length;j++) {
				if(i!=j) {
					logConditionalProbs[i][j]+= logHetero;
				}
				
			}
		}	
	}
	/**
	 * Updates counts and conditional probabilities for the given allele call.
         * Keeps track of strand specific counts
	 * @param base New allele call to count
	 * @param charQualityScore Quality score of the allele call in Phred+33 scale
	 * @param mappingQuality Quality score of the mapping in Phred scale
	 */
	public void updateCounts (char base, char charQualityScore, short mappingQuality, short Flags) {
		double readProb = 1;
		if(mappingQuality<30) {
			readProb = (1 - Math.pow(10.0, (-mappingQuality/10.0)));
		}
		//System.out.println(base + " "+(3*errorProb)+" "+readProb);
		int index = alphabeth.indexOf(base);
		if(index>=0) {
                    if ((Flags & 0x10) == 0) {
                            ptotalCount+=readProb;
                            pcounts[index]+=readProb;
                    }
                    else {
                            ntotalCount+=readProb;
                            ncounts[index]+=readProb;
                    }

                }
                updateCounts (base,charQualityScore,mappingQuality);

        }
	/**
	 * Updates counts and conditional probabilities for the given allele call
	 * @param base New allele call to count
	 * @param charQualityScore Quality score of the allele call in Phred+33 scale
	 * @param mappingQuality Quality score of the mapping in Phred scale
	 */
	public void updateCounts (char base, char charQualityScore, short mappingQuality) {
		int qualScore = (int)(charQualityScore-33);
		if(qualScore<=0) {
			qualScore = 1;
		}
		double readProb = 1;
		double errorProb = Math.pow(10.0, (-qualScore/10.0));
		double successProb = (1 - errorProb);
		errorProb = errorProb/3;
		if(mappingQuality<30) {
			readProb = (1 - Math.pow(10.0, (-mappingQuality/10.0)));
		}
		//System.out.println(base + " "+(3*errorProb)+" "+readProb);
		int index = alphabeth.indexOf(base);
		if(index>=0) {
			totalCount+=readProb;
			counts[index]+=readProb;
			if(successProb > errorProb) {
				double term = pAllele1*(1-4*errorProb);
				for(int i=0;i<logConditionalProbs.length;i++) {
					double errorCont = readProb*Math.log10(errorProb);
					if(i==index) {
						logConditionalProbs[i][i] += readProb*Math.log10(successProb); 
					} else {
						logConditionalProbs[i][i] += errorCont;
					}
					for(int j=0;j<logConditionalProbs[0].length;j++) {
						if(i!=j) {
							if(i==index) {
								logConditionalProbs[i][j] += readProb*Math.log10(errorProb+term); 
							} else if (j==index) {
								logConditionalProbs[i][j] += readProb*Math.log10(successProb-term);
							} else {
								logConditionalProbs[i][j] += errorCont;
							}
						}
					}
				}
			}
		}
	}
	/**
	 * Calculates the posterior probability of each genotype
	 * @return double [][] Squared matrix of probabilities
	 */
	public double [][] getPosteriorProbabilities () {
		int bases = alphabeth.length();
		boolean allEvents = pAllele1!=0.5;
		double [] conditionals = new double [bases+bases*(bases-1)/2];
		if(allEvents) {
			conditionals = new double [bases*bases];
		}
		double [][] posteriorProb = new double [bases][bases];
		int indCond=0;
		for(int i=0;i<bases;i++) {
			conditionals[indCond] = logConditionalProbs[i][i];
			indCond++;
			for(int j=0;j<bases;j++) {
				if((j<i && allEvents) || j>i) {
					conditionals[indCond] = logConditionalProbs[i][j];
					indCond++;
				}
			}
		}
		convertToPosteriorProbabilities(conditionals);
		indCond=0;
		for(int i=0;i<bases;i++) {
			posteriorProb[i][i] = conditionals[indCond];
			indCond++;
			for(int j=0;j<bases;j++) {
				if((j<i && allEvents) || j>i) {
					posteriorProb[i][j] = conditionals[indCond];
					indCond++;
				}
			}
		}
		return posteriorProb;
	}
	
	private void convertToPosteriorProbabilities (double [] logCondProbs) {
		//Calculate max prob
		double logMax = 1;
		for(int i=0;i<logCondProbs.length;i++) {
			if(logMax > 0 || logMax < logCondProbs[i]) {
				logMax = logCondProbs[i];
			}
		}
		//Calculate total probability and contribution of each possible event
		double totalProb = 0;
		for(int i=0;i<logCondProbs.length;i++) { 
			logCondProbs[i] -= logMax;
			if(logCondProbs[i] < -20) {
				logCondProbs[i] = 0.0;
			} else {
				logCondProbs[i] = Math.pow(10.0, logCondProbs[i]);
			}
			totalProb+=logCondProbs[i];
		}
		//Normalize
		for(int i=0;i<logCondProbs.length;i++) { 
			logCondProbs[i] = logCondProbs[i]/totalProb;
		}
	}
	/**
	 * Prints the given probabilities. 
	 * @param probs Probabilities matrix to print
	 * @param logs If true, it assumes that the given matrix have log probabilities and then
	 * calculates exp(10,probs[i][j]) before printing
	 */
	public void printProbs (double [][] probs,boolean logs) {
		for(int i=0;i<probs.length;i++) { 
			for(int j=0;j<probs[0].length;j++) {
				if(logs) {
					System.out.print(" "+Math.pow(10.0, probs[i][j]));
				} else {
					System.out.print(" "+probs[i][j]);
				}
				
			}
			System.out.println();
		}
	}
	/**
	 * @return double [] Counts for each possible allele
	 */
	public double[] getCounts() {
		return counts;
	}
	/**
	 * Gets the count for the given allele
	 * @param c Allele to look for
	 * @return double Count for the given allele
	 */
	public double getCount(char c) {
		int index = alphabeth.indexOf(c);
		if(index>=0) {
			return counts[index];
		}
		return 0;
	}
	/**
	 * @return double Total count of alleles
	 */
	public double getTotalCount() {
		return totalCount;
	}

	/**
	 * @return double [] Counts for each possible allele on the +ve strand
	 */
       	public double[] getpCounts() {
		return pcounts;
	}
	/**
	 * Gets the count for the given allele on the +ve
	 * @param c Allele to look for
	 * @return double Count for the given allele
	 */
	public double getpCount(char c) {
		int index = alphabeth.indexOf(c);
		if(index>=0) {
			return pcounts[index];
		}
		return 0;
	}
	/**
	 * @return double Total count of alleles on the +ve strand
	 */
	public double getpTotalCount() {
		return ptotalCount;
	}

	/**
	 * @return double [] Counts for each possible allele on the -ve strand
	 */
       	public double[] getnCounts() {
		return ncounts;
	}
	/**
	 * Gets the count for the given allele on the -ve
	 * @param c Allele to look for
	 * @return double Count for the given allele
	 */
	public double getnCount(char c) {
		int index = alphabeth.indexOf(c);
		if(index>=0) {
			return ncounts[index];
		}
		return 0;
	}
	/**
	 * @return double Total count of alleles on the -ve strand
	 */
	public double getnTotalCount() {
		return ntotalCount;
	}

}
