/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.variants;

import edu.uconn.ngstools.genome.GenomicRegion;

/**
 * Class to store a pileup record information
 * Implements the GenomicRegion interface by like a SNV
 * @author Jorge Duitama
 */
public class PileupRecord implements GenomicRegion {
	private String sequenceName;
	private int position;
	private char refBase;
	private short maxQualScore;
	private String alleleCalls;
	private String qualityScores;
	private char genotype;
	private double genotypeProbability;
	private double refProbability=0;
	/**
	 * Creates a pileup record with the given information
	 * @param sequenceName Name of the reference sequence
	 * @param position Position in the reference sequence
	 * @param refBase Base present in the reference sequence in the given position
	 * @param alleleCalls Allele calls for the given locus in pileup format
	 * @param qualityScores Quality scores of the allele calls
	 */
	public PileupRecord(String sequenceName, int position, char refBase, String alleleCalls, String qualityScores) {
		super();
		this.sequenceName = sequenceName;
		this.position = position;
		this.refBase = refBase;
		this.alleleCalls = alleleCalls;
		this.qualityScores = qualityScores;
	}
	@Override
	public int getEnd() {
		return position;
	}
	@Override
	public String getSequenceName() {
		return sequenceName;
	}
	/**
	 * Changes the reference sequence name
	 * @param sequenceName New sequence name
	 */
	public void setSequenceName(String sequenceName) {
		this.sequenceName = sequenceName;
	}
	@Override
	public int getStart() {
		return position;
	}
	@Override
	public int length() {
		return 1;
	}
	/**
	 * @return int Position in the reference sequence
	 */
	public int getPosition() {
		return position;
	}
	/**
	 * Changes the position in the reference sequence
	 * @param position New position
	 */
	public void setPosition(int position) {
		this.position = position;
	}
	/**
	 * @return char Base present in this reference sequence and position
	 */
	public char getRefBase() {
		return refBase;
	}
	/**
	 * Changes the base present in this reference sequence and position
	 * @param refBase New base
	 */
	public void setRefBase(char refBase) {
		this.refBase = refBase;
	}
	/**
	 * @return short Maximum mapping quality score for this locus 
	 */
	public short getMaxQualScore() {
		return maxQualScore;
	}
	/**
	 * Changes the maximum mapping quality score
	 * @param maxQualScore New maximum mapping quality score
	 */
	public void setMaxQualScore(short maxQualScore) {
		this.maxQualScore = maxQualScore;
	}
	/**
	 * @return String Allele calls in pileup format
	 */
	public String getAlleleCalls() {
		return alleleCalls;
	}
	/**
	 * Changes the allele calls in pileup format
	 * @param alleleCalls New allele calls
	 */
	public void setAlleleCalls(String alleleCalls) {
		this.alleleCalls = alleleCalls;
	}
	/**
	 * @return String Quality scores of allele calls
	 */
	public String getQualityScores() {
		return qualityScores;
	}
	/**
	 * Changes the quality scores of allele calls
	 * @param qualityScores New quality scores
	 */
	public void setQualityScores(String qualityScores) {
		this.qualityScores = qualityScores;
	}
	/**
	 * @return char Genotype for this locus as a degenerate base
	 */
	public char getGenotype() {
		return genotype;
	}
	/**
	 * Changes the genotype for this locus
	 * @param genotype New genotype as a degenerate base
	 */
	public void setGenotype(char genotype) {
		this.genotype = genotype;
	}
	/**
	 * @return double Posterior probability of the genotype
	 */
	public double getGenotypeProbability() {
		return genotypeProbability;
	}
	/**
	 * Changes the posterior probability of the genotype
	 * @param genotypeProbability New probability
	 */
	public void setGenotypeProbability(double genotypeProbability) {
		this.genotypeProbability = genotypeProbability;
	}
	/**
	 * @return double Posterior probability of having a homozygous reference genotype 
	 */
	public double getRefProbability() {
		return refProbability;
	}
	/**
	 * Changes the posterior probability of having a homozygous reference genotype
	 * @param refProbability New probability
	 */
	public void setRefProbability(double refProbability) {
		this.refProbability = refProbability;
	}
	/**
	 * @return int Number of calls for this locus
	 */
	public int getNumCalls() {
		return qualityScores.length();
	}
}
