/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.variants;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import edu.uconn.ngstools.genome.GenomicRegionComparator;

/**
 * Class with different kinds of filters for SNVs. Implemented filters are the follwing
 * <ul>
 * <li>Minimum coverage on the alternative allele
 * <li>Maximum coverage on the alternative allele
 * <li>Minimum genotype posterior probability
 * <li>SNV must not be present in a given SNV list
 * <li>SNV reference allele must not be lower case
 * </ul>
 * The last filter is on by default but can be deactivated with the method setKeepLowerCaseRef
 * @author Jorge Duitama
 */
public class SNVSFilter {
	public static final double DEF_MIN_ALT_COVERAGE = 0;
	public static final double DEF_MAX_ALT_COVERAGE = 0;
	public static final double DEF_MIN_PROBABILITY = 0;
	private double minAltCoverage=DEF_MIN_ALT_COVERAGE;
	private double maxAltCoverage=DEF_MAX_ALT_COVERAGE;
	private double minProbability = DEF_MIN_PROBABILITY;
	private Set<SNV> snvsToExclude;
	private boolean keepLowerCaseRef = false;
	/**
	 * Creates a default filter
	 */
	public SNVSFilter() {
		
	}
	/**
	 * Creates a filter with the given information
	 * @param minAltCoverage Minimum coverage of the alternative allele
	 */
	public SNVSFilter(double minAltCoverage) {
		super();
		this.setMinAltCoverage(minAltCoverage);
	}
	/**
	 * Creates a filter with the given information
	 * @param minAltCoverage Minimum coverage of the alternative allele
	 * @param maxAltCoverage Maximum coverage of the alternative allele
	 */
	public SNVSFilter(double minAltCoverage, double maxAltCoverage) {
		super();
		this.setMinAltCoverage(minAltCoverage);
		this.setMaxAltCoverage(maxAltCoverage);
	}
	/**
	 * Creates a filter with the given information
	 * @param minAltCoverage Minimum coverage of the alternative allele
	 * @param maxAltCoverage Maximum coverage of the alternative allele
	 * @param minProbability Minimum genotype posterior probability
	 */
	public SNVSFilter(double minAltCoverage, double maxAltCoverage, double minProbability) {
		super();
		this.setMinAltCoverage(minAltCoverage);
		this.setMaxAltCoverage(maxAltCoverage);
		this.setMinProbability(minProbability);
	}
	/**
	 * @return double Minimum coverage of the alternative allele
	 */
	public double getMinAltCoverage() {
		return minAltCoverage;
	}
	/**
	 * Changes the minimum coverage of the alternative allele
	 * @param minAltCoverage New coverage
	 */
	public void setMinAltCoverage(double minAltCoverage) {
		this.minAltCoverage = minAltCoverage;
	}
	/**
	 * @return double Maximum coverage of the alternative allele
	 */
	public double getMaxAltCoverage() {
		return maxAltCoverage;
	}
	/**
	 * Changes the maximum coverage of the alternative allele
	 * @param maxAltCoverage New coverage
	 */
	public void setMaxAltCoverage(double maxAltCoverage) {
		this.maxAltCoverage = maxAltCoverage;
	}
	/**
	 * @return double Minimum genotype posterior probability
	 */
	public double getMinProbability() {
		return minProbability;
	}
	/**
	 * Changes the minimum genotype posterior probability
	 * @param minProbability New probability
	 */
	public void setMinProbability(double minProbability) {
		this.minProbability = minProbability;
	}
	/**
	 * @return List<SNV> SNVs to be filtered out
	 */
	public List<SNV> getSnvsToExclude() {
		return new ArrayList<SNV>(snvsToExclude);
	}
	/**
	 * Changes the SNVs to be filtered out
	 * @param snvsToExclude New SNVs
	 */
	public void setSnvsToExclude(List<SNV> snvsToExclude) {
		Set<String> seqNames = new TreeSet<String>();
		for(SNV snv:snvsToExclude) {
			if(!seqNames.contains(snv.getSequenceName())) {
				seqNames.add(snv.getSequenceName());
			}
		}
		this.snvsToExclude = new TreeSet<SNV>(new GenomicRegionComparator(new ArrayList<String>(seqNames)));
		this.snvsToExclude.addAll(snvsToExclude);
	}
	/**
	 * @return true if SNVs with lower case reference allele are not filtered, false otherwise
	 */
	public boolean isKeepLowerCaseRef() {
		return keepLowerCaseRef;
	}
	/**
	 * Changes the keep lower case reference status
	 * @param keepLowerCaseRef true f SNVs with lower case reference allele must not be filtered, 
	 * false otherwise
	 */
	public void setKeepLowerCaseRef(boolean keepLowerCaseRef) {
		this.keepLowerCaseRef = keepLowerCaseRef;
	}
	/**
	 * Creates a new list with the SNVs that pass all filters configured in this object
	 * @param snvs Listo of SNVs to filter
	 * @return List<CalledSNV> SNVs that pass the configured filters
	 */
	public List<CalledSNV> filterSNVS(List <CalledSNV> snvs) {
		List<CalledSNV> answer = new ArrayList<CalledSNV>();
		for(CalledSNV csnv: snvs) {
			if(passFilter(csnv)) {
				answer.add(csnv);
			}
		}
		return answer;
	}
	/**
	 * Tells if the given SNV passes the filters
	 * @param csnv SNV to test
	 * @return boolean true if the SNV passes the filters, false otherwise
	 */
	public boolean passFilter (CalledSNV csnv) {
		//Min coverage filter
		if(minAltCoverage != DEF_MIN_ALT_COVERAGE && csnv.getCountAlternative()<minAltCoverage) {
			return false;
		}
		//Max coverage filter
		if(maxAltCoverage != DEF_MAX_ALT_COVERAGE && csnv.getCountAlternative()>maxAltCoverage) {
			return false;
		}
		//Probability filter
		if(minProbability != DEF_MIN_PROBABILITY && csnv.getProbability()<minProbability) {
			return false;
		}
		//SNVS to exclude filter
		if(snvsToExclude!=null && snvsToExclude.contains(csnv)) {
			return false;
		}
		//Drop reference in lower case filter 
		if(!keepLowerCaseRef && Character.isLowerCase(csnv.getRefBase())) {
			return false;
		}
		return true;
	}
}
