/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.variants.filehandlers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import edu.uconn.ngstools.variants.CalledSNV;
import edu.uconn.ngstools.variants.SNV;

/**
 * Handler for i/o operations on CalledSNVs. The format includes the following columns
 * separated by tab:
 * <ul>
 * <li>Sequence name
 * <li>Position
 * <li>Reference base
 * <li>Alternative base
 * <li>Genotype as a two bases string
 * <li>Coverage of the reference allele
 * <li>Coverage of the alternative allele
 * <li>Posterior probability the genotype call as a phred score
 * <li>Posterior probability of a genotype different than homozygous reference as a phred score
 * <li>Coverage of the A allele
 * <li>Coverage of the C allele
 * <li>Coverage of the G allele
 * <li>Coverage of the T allele
 * </ul>
 * Just the first five columns are required.
 * @author Jorge Duitama
 */
public class CalledSNVSFileHandler implements SNVSFileHandler {
	private String separator = "\t";
	@Override
	public List<SNV> loadSNVS(String snvsFilename) throws IOException {
		List<SNV> answer = new ArrayList<SNV>();
		answer.addAll(loadCalledSNVS(snvsFilename));
		return answer;
	}
	/**
	 * Loads SNVs in CalledSNV format from the given file
	 * @param snvsFilename Name of the file to load the SNVs from 
	 * @return List<SNV> List of SNVs in the given file
	 * @throws IOException If the file can not be read
	 */
	public List<CalledSNV> loadCalledSNVS(String snvsFilename) throws IOException {
		List<CalledSNV> answer = new ArrayList<CalledSNV>();
		FileInputStream fis = new FileInputStream(snvsFilename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		for(int i=0;line!=null;i++) {
			
			CalledSNV snv = loadCalledSNV(line);
			answer.add(snv);
			line=in.readLine();
		}
		fis.close();
		return answer;
	}
	/**
	 * Loads a SNV with the information provided by the given String in CalledSNV format 
	 * @param line String with the information to load a CalledSNV
	 * @return CalledSNV Object with the SNV information
	 */
	public CalledSNV loadCalledSNV (String line) {
		String [] items = line.split(separator);
		CalledSNV snv = new CalledSNV(items[0],Integer.parseInt(items[1]),items[2].charAt(0),items[3].charAt(0),items[4]);
		if(items.length>5) {
			snv.setCountReference(Integer.parseInt(items[5]));
			snv.setCountAlternative(Integer.parseInt(items[6]));
			snv.setProbability(1-PhredScoreHelper.calculateProbability(Short.parseShort(items[7])));
			snv.setRefProbability(PhredScoreHelper.calculateProbability(Short.parseShort(items[8])));
		}
		if(items.length>9) {
			double [] counts = new double[4];
			counts[0] = Double.parseDouble(items[9]);
			counts[1] = Double.parseDouble(items[10]);
			counts[2] = Double.parseDouble(items[11]);
			counts[3] = Double.parseDouble(items[12]);
			snv.setAllCounts(counts);
		}
		return snv;
	}
	/**
	 * Prints the given CalledSNVs in the given Stream 
	 * @param snvs List of SNVs to print
	 * @param out Stream where SNVs will be printed
	 */
	public void saveSNVS (List<CalledSNV> snvs, PrintStream out) {
		for(SNV snv: snvs) {
			if(snv instanceof CalledSNV) {
				CalledSNV csnv = (CalledSNV)snv;
				saveSNV(csnv, out, false);
			}
		}
		out.flush();
	}

	/**
	 * Prints header with column labels for output file
	 * @param out Stream to print the SNV information
         * @param strandCounts Boolean true if strand specific counts are to be printed
	 */
        public void printHeader(PrintStream out, boolean strandCounts) {
		out.print("Chr");
		out.print(separator);
		out.print("Pos");
		out.print(separator);
		out.print("Ref");
		out.print(separator);
		out.print("Alt");
		out.print(separator);
		out.print("Gen");
		out.print(separator);
		out.print("RefCov");
		out.print(separator);
		out.print("AltCov");
		out.print(separator);
		out.print("GenQul");
		out.print(separator);
		out.print("SNVQul");
		out.print(separator);
		out.print("A's");
		out.print(separator);
		out.print("C's");
		out.print(separator);
		out.print("G's");
		out.print(separator);
		out.print("T's");
//strand specific counts
                if (strandCounts) {
                    out.print(separator);
                    out.print(separator);
                    out.print("+ve Ref Cov");
                    out.print(separator);
                    out.print("+ve Alt Cov");
                    out.print(separator);
                    out.print("+ve A Calls");
                    out.print(separator);
                    out.print("+ve C Calls");
                    out.print(separator);
                    out.print("+ve G Calls");
                    out.print(separator);
                    out.print("+ve T Calls");
                    out.print(separator);
                    out.print(separator);
                    out.print("-ve Ref Cov");
                    out.print(separator);
                    out.print("-ve Alt Cov");
                    out.print(separator);
                    out.print("-ve A Calls");
                    out.print(separator);
                    out.print("-ve C Calls");
                    out.print(separator);
                    out.print("-ve G Calls");
                    out.print(separator);
                    out.print("-ve T Calls");
                }
                out.println();
		out.flush();
	}

        

        /**
	 * Prints the given CalledSNV information in the given Stream
	 * @param csnv Object with information to print in CalledSNV format
	 * @param out Stream to print the SNV information
         * @param strandCounts Boolean true if strand specific counts are to be printed
	 */
	public void saveSNV(CalledSNV csnv, PrintStream out, boolean strandCounts) {
		out.print(csnv.getSequenceName());
		out.print(separator);
		out.print(csnv.getPosition());
		out.print(separator);
		out.print(csnv.getRefBase());
		out.print(separator);
		out.print(csnv.getAltBase());
		out.print(separator);
		out.print(csnv.getGenotype());
		out.print(separator);
		out.print(Math.round(csnv.getCountReference()));
		out.print(separator);
		out.print(Math.round(csnv.getCountAlternative()));
		out.print(separator);
		out.print(PhredScoreHelper.calculatePhredScore(1-csnv.getProbability()));
		out.print(separator);
		out.print(PhredScoreHelper.calculatePhredScore(csnv.getRefProbability()));
		if(csnv.getAllCounts()!=null) {
			double [] counts = csnv.getAllCounts();
			for(int i=0;i<counts.length;i++) {
				out.print(separator);
				out.print(Math.round(counts[i]));
			}
		}
//strand specific counts
                if (strandCounts) {
                    out.print(separator);
                    out.print(separator);
                    out.print(Math.round(csnv.getpCountReference()));
                    out.print(separator);
                    out.print(Math.round(csnv.getpCountAlternative()));
                    if(csnv.getpAllCounts()!=null) {
			double [] counts = csnv.getpAllCounts();
			for(int i=0;i<counts.length;i++) {
				out.print(separator);
				out.print(Math.round(counts[i]));
			}
                    }
                    out.print(separator);
                    out.print(separator);
                    out.print(Math.round(csnv.getnCountReference()));
                    out.print(separator);
                    out.print(Math.round(csnv.getnCountAlternative()));
                    if(csnv.getnAllCounts()!=null) {
			double [] counts = csnv.getnAllCounts();
			for(int i=0;i<counts.length;i++) {
				out.print(separator);
				out.print(Math.round(counts[i]));
			}
                    }
                }
                out.println();
		out.flush();
	}
	

}
