/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.variants.filehandlers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import edu.uconn.ngstools.genome.GenomicRegionComparator;
import edu.uconn.ngstools.sequences.DefaultSequenceNames;
import edu.uconn.ngstools.variants.CalledSNV;
import edu.uconn.ngstools.variants.SNV;

/**
 * SNVs handler that enables a broad spectrum of formats through configurable
 * parameters. Includes a small program for translation to CalledSNV format.
 * @author Jorge Duitama
 *
 */
public class DynamicSNVSFileHandler implements SNVSFileHandler { 
	private String separator = " |\t";
	private String prefixSeqName="";
	private String prefixPosition="";
	private boolean sorted=true;
	private boolean hetSNPsFile=false;
	private int indexSeqName=0;
	private int indexPosition=1;
	private int indexRefBase=2;
	private int indexAltBase=3;
	private int indexGenotype=-1;
	private List<String> sortedSequenceNames = DefaultSequenceNames.DEFAULT_SEQUENCE_NAMES;
	@Override
	public List<SNV> loadSNVS(String snvsFilename) throws IOException {
		List<SNV> answer = new ArrayList<SNV>();
		FileInputStream fis = new FileInputStream(snvsFilename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		for(int i=0;line!=null;i++) {
			String [] items = line.split(separator);
			boolean valid = true;
			String seqName = items[indexSeqName];
			String posStr = items[indexPosition];
			char refBase = items[indexRefBase].charAt(0);
			char altBase = items[indexAltBase].charAt(0);
			String genotype = null;
			if(hetSNPsFile) {
				genotype = ""+refBase+""+altBase;
			} else if(indexGenotype>=0) {
				genotype = items[indexGenotype];
			}
			if(seqName.startsWith(prefixSeqName)) {
				seqName = seqName.substring(prefixSeqName.length());
			} else {
				System.err.println("Seqname: "+seqName+" does not start with prefix: "+prefixSeqName);
				valid = false;
			}
			int pos =0;
			if(posStr.startsWith(prefixPosition)) {
				pos = Integer.parseInt(posStr.substring(prefixPosition.length()));
			} else {
				System.err.println("Position: "+posStr+" does not start with prefix: "+prefixPosition);
				valid = false;
			}
			if(valid) {
				SNV snv;
				if(genotype!=null) {
					snv = new CalledSNV(seqName, pos, refBase, altBase, genotype);
				} else {
					snv = new SNV(seqName, pos, refBase, altBase);
				}
				answer.add(snv);
			}
			
			line=in.readLine();
		}
		fis.close();
		if(!sorted) {
			GenomicRegionComparator comp = new GenomicRegionComparator(sortedSequenceNames);
			Collections.sort(answer, comp);
		}
		return answer;
	}
	/**
	 * @return String Regular expression used to separate columns in the file
	 */
	public String getSeparator() {
		return separator;
	}
	/**
	 * Changes the regular expression used to separate columns in the file
	 * @param separator New regular expression
	 */
	public void setSeparator(String separator) {
		this.separator = separator;
	}
	/**
	 * @return String Prefix to be taken out from the sequence name
	 */
	public String getPrefixSeqName() {
		return prefixSeqName;
	}
	/**
	 * Changes the prefix to be taken out from the sequence name
	 * @param prefixSeqName New prefix
	 */
	public void setPrefixSeqName(String prefixSeqName) {
		this.prefixSeqName = prefixSeqName;
	}
	/**
	 * @return String Prefix to be taken out from the position
	 */
	public String getPrefixPosition() {
		return prefixPosition;
	}
	/**
	 * Changes the prefix to be taken out from the position
	 * @param prefixPosition New prefix
	 */
	public void setPrefixPosition(String prefixPosition) {
		this.prefixPosition = prefixPosition;
	}
	/**
	 * @return boolean true if the file is sorted, false otherwise
	 */
	public boolean isSorted() {
		return sorted;
	}
	/**
	 * Changes the sorted flag
	 * @param sorted true if the file will be sorted, false otherwise
	 */
	public void setSorted(boolean sorted) {
		this.sorted = sorted;
	}
	/**
	 * @return int Column index for the sequence name 
	 */
	public int getIndexSeqName() {
		return indexSeqName;
	}
	/**
	 * Changes the column index for the sequence name
	 * @param indexSeqName New index
	 */
	public void setIndexSeqName(int indexSeqName) {
		this.indexSeqName = indexSeqName;
	}
	/**
	 * @return int Column index for the position
	 */
	public int getIndexPosition() {
		return indexPosition;
	}
	/**
	 * Changes the column index for the position
	 * @param indexPosition New index
	 */
	public void setIndexPosition(int indexPosition) {
		this.indexPosition = indexPosition;
	}
	/**
	 * @return int Column index for the reference base
	 */
	public int getIndexRefBase() {
		return indexRefBase;
	}
	/**
	 * Changes the column index for the reference base
	 * @param indexRefBase New index
	 */
	public void setIndexRefBase(int indexRefBase) {
		this.indexRefBase = indexRefBase;
	}
	/**
	 * @return int Column index for the alternative base
	 */
	public int getIndexAltBase() {
		return indexAltBase;
	}
	/**
	 * Changes the column index for the alternative base
	 * @param indexAltBase New index
	 */
	public void setIndexAltBase(int indexAltBase) {
		this.indexAltBase = indexAltBase;
	}
	/**
	 * @return int Column index for the genotype
	 */
	public int getIndexGenotype() {
		return indexGenotype;
	}
	/**
	 * Changes the column index for the genotype
	 * @param indexGenotype New index
	 */
	public void setIndexGenotype(int indexGenotype) {
		this.indexGenotype = indexGenotype;
	}
	/**
	 * @return boolean true if the file just contains heterozygous SNPs, false otherwise
	 */
	public boolean isHetSNPsFile() {
		return hetSNPsFile;
	}
	/**
	 * Changes the HetSNPsFile Flag
	 * @param hetSNPsFile true if the file just will contain heterozygous SNPs, false otherwise
	 */
	public void setHetSNPsFile(boolean hetSNPsFile) {
		this.hetSNPsFile = hetSNPsFile;
	}


	/**
	 * Small program to translate a SNVs file that can be processed with this handler to a file in 
	 * CalledSNVs format
	 * @param args
	 * @throws IOException 
	 */
	public static void main(String[] args) throws IOException {
		DynamicSNVSFileHandler handlerIn = new DynamicSNVSFileHandler();
		CalledSNVSFileHandler handlerOut = new CalledSNVSFileHandler();
		int i=0;
		while(i<args.length && args[i].charAt(0)=='-') {
			if("-separator".equalsIgnoreCase(args[i])) {
				i++;
				handlerIn.separator = " |"+args[i];
			} else if("-prefixSeqName".equalsIgnoreCase(args[i])) {
				i++;
				handlerIn.prefixSeqName = args[i];
			} else if("-prefixPosition".equalsIgnoreCase(args[i])) {
				i++;
				handlerIn.prefixPosition = args[i];
			} else if("-indexSeqName".equalsIgnoreCase(args[i])) {
				i++;
				handlerIn.indexSeqName = Integer.parseInt(args[i])-1;
			} else if("-indexPosition".equalsIgnoreCase(args[i])) {
				i++;
				handlerIn.indexPosition = Integer.parseInt(args[i])-1;
			} else if("-indexRefBase".equalsIgnoreCase(args[i])) {
				i++;
				handlerIn.indexRefBase = Integer.parseInt(args[i])-1;
			} else if("-indexAltBase".equalsIgnoreCase(args[i])) {
				i++;
				handlerIn.indexAltBase = Integer.parseInt(args[i])-1;
			} else if("-indexGenotype".equalsIgnoreCase(args[i])) {
				i++;
				handlerIn.indexGenotype = Integer.parseInt(args[i])-1;
			} else if("-sort".equalsIgnoreCase(args[i])) {
				handlerIn.sorted = false;
			} else if("-hetSNPsFile".equalsIgnoreCase(args[i])) {
				handlerIn.hetSNPsFile = true;
			} 
			i++;
		}
		List<SNV> snvs = handlerIn.loadSNVS(args[i]);
		List<CalledSNV> csnvs = new ArrayList<CalledSNV>();
		for(SNV snv:snvs) {
			if(snv instanceof CalledSNV) {
				csnvs.add((CalledSNV)snv);
			}
		}
		handlerOut.saveSNVS(csnvs, System.out);
	}

}
