/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.variants.filehandlers;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import edu.uconn.ngstools.genome.Genome;
import edu.uconn.ngstools.genome.GenomeAssembly;
import edu.uconn.ngstools.sequences.DNASequence;
import edu.uconn.ngstools.variants.CalledSNV;
import edu.uconn.ngstools.variants.SNV;

/**
 * Loads Hapmap genotypes for a given individual. Since Hapmap files contain genotypes 
 * for more than one individial, this class can be configured to load just the genotypes
 * of the individual of interest. Hapmap information also have strand problems, so a reference
 * genome is required to resolve the strand information of each locus or at least filter out 
 * the loci with strand issues. Includes a small program for translation to CalledSNV format.
 * @author Jorge Duitama
 *
 */
public class HapmapSNPSHandler implements SNVSFileHandler {
	private Genome genome;
	private String individual;

	@Override
	public List<SNV> loadSNVS(String snvsFilename) throws IOException {
		List<SNV> answer = new ArrayList<SNV>();
		
		FileInputStream fis = new FileInputStream(snvsFilename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		int index = getIndividualIndex(line,individual);
		line=in.readLine();
		for(int i=0;line!=null;i++) {
			String [] items = line.split(" |\t");
			String seqName = items[2];
			int position = Integer.parseInt(items[3]);
			char a1 = Character.toUpperCase(items[1].charAt(0));
			char a2 = Character.toUpperCase(items[1].charAt(2));
			String genotype = items[index].toUpperCase();
			char g1 = genotype.charAt(0);
			char g2 = genotype.charAt(1);
			char refBase = genome.getReferenceBase(seqName, position);
			boolean discard = refBase==0  || genotype.indexOf('N') >= 0;
			char upRefBase = Character.toUpperCase(refBase);
			boolean reverse = upRefBase!=a1 && upRefBase!=a2;
			if(a1 == DNASequence.getComplement(a2) && g1==g2) {
				discard = true;
			}
			if(!discard ) {
				if(reverse) {
					a1 = DNASequence.getComplement(a1);
					a2 = DNASequence.getComplement(a2);
					g1 = DNASequence.getComplement(g1);
					g2 = DNASequence.getComplement(g2);
					genotype = ""+g1+""+g2;
				}
				char altBase = a1;
				if(upRefBase == altBase) {
					altBase = a2;
				}
				answer.add(new CalledSNV(seqName, position, refBase, altBase, genotype));
			}
			line=in.readLine();
		}
		fis.close();
		return answer;
	}
	/**
	 * @return String Id of the individual whose genotype is going to be loaded 
	 */
	public String getIndividual() {
		return individual;
	}
	/**
	 * Changes the id of the individual whose genotype is going to be loaded
	 * @param individual
	 */
	public void setIndividual(String individual) {
		this.individual = individual;
	}
	/**
	 * Gets the column id in the file for the given individual according with the given 
	 * line
	 * @param line Line to look for 
	 * @param individual Id to look for
	 * @return int Column where the genotype of the individual is found
	 */
	public static int getIndividualIndex(String line, String individual) {
		String [] items = line.split(" |\t");
		for(int  i=0;i< items.length;i++) {
			if(individual.equals(items[i])) {
				return i;
			}
		}
		return -1;
	}
	/**
	 * Small program to transfer Hapmap genotypes in CalledSNV format
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		HapmapSNPSHandler handler = new HapmapSNPSHandler();
		CalledSNVSFileHandler csnpsHandler = new CalledSNVSFileHandler();
		handler.genome = new GenomeAssembly(args[1]);
		handler.setIndividual(args[2]);
		List <SNV> snvs = handler.loadSNVS(args[0]);
		List<CalledSNV> csnvs = new ArrayList<CalledSNV>();
		for(SNV snv:snvs) {
			if(snv instanceof CalledSNV) {
				csnvs.add((CalledSNV)snv);
			}
		}
		csnpsHandler.saveSNVS(csnvs, System.out);
	}
}
