/*******************************************************************************
 * NGSTools - Java tools for analysis of Next Generation Sequencing (NGS) data
 * Copyright 2010 Jorge Duitama
 *
 * This file is part of NGSTools.
 *
 *     NGSTools is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     NGSTools is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with NGSTools.  If not, see <http://www.gnu.org/licenses/>.
 *******************************************************************************/
package edu.uconn.ngstools.variants.filehandlers;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintStream;

import edu.uconn.ngstools.variants.PileupRecord;

/**
 * Handler for i/o operations on pileup records
 * @author Jorge Duitama
 */
public class PileupFileHandler {
	private String lastSeqName = null;
	/**
	 * Loads a record from the given reader
	 * @param in Reader to load the pileup information
	 * @return PileupRecord Record for the next line of the reader or null if no 
	 * more lines are generated
	 * @throws IOException If the Reader can not be queried
	 */
	public PileupRecord loadPileupRecord(BufferedReader in) throws IOException {
		String line = in.readLine();
		if(line!=null) {
			return loadPileupRecord(line);
		}
		return null;
	}
	/**
	 * Loads a record from the given line
	 * @param line Line with the information to load the record in pileup format
	 * @return PileupRecord Record for the given line
	 */
	public PileupRecord loadPileupRecord(String line) {
		String [] items = line.split(" |\t");
		String seq = items[0];
		if(!seq.equals(lastSeqName)) {
			lastSeqName = seq;
		}
		int position = Integer.parseInt(items[1]);
		char refBase = items[2].charAt(0);
		String calls = items[items.length-2];
		String qualities = items[items.length-1];
		
		PileupRecord record = new PileupRecord(lastSeqName, position, refBase, calls, qualities);
		record.setMaxQualScore(Short.parseShort(items[4]));
		if(items.length>=10) {
			record.setGenotype(items[3].charAt(0));
			double p = 1-PhredScoreHelper.calculateProbability(Short.parseShort(items[4]));
			double pRef = PhredScoreHelper.calculateProbability(Short.parseShort(items[5]));
			record.setGenotypeProbability(p);
			record.setRefProbability(pRef);
		}
		return record;
	}
	/**
	 * Saves the given record in the given stream
	 * @param record Object with the pileup information to store 
	 * @param out Stream where the record information should be printed in pileup format
	 */
	public void printRecord(PileupRecord record, PrintStream out) {
		short q1 = PhredScoreHelper.calculatePhredScore(1-record.getGenotypeProbability());
		short q2 = PhredScoreHelper.calculatePhredScore(record.getRefProbability());
		out.print(record.getSequenceName()+"\t");
		out.print(record.getPosition()+"\t");
		out.print(record.getRefBase()+"\t");
		out.print(record.getGenotype()+"\t");
		out.print(""+q1+"\t");
		out.print(""+q2+"\t");
		out.print(record.getMaxQualScore()+"\t");
		out.print(record.getQualityScores().length()+"\t");
		out.print(record.getAlleleCalls()+"\t");
		out.println(record.getQualityScores());	
	}
}
