package edu.uconn.bioinformatics.epitopeFinder;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.TreeMap;


public class DatabaseProteinEpitopeChecker implements ProteinEpitopeChecker {

	private List<PredictionPWM> matrices = null;
	private String [] motifs = null;
	@Override
	public List<EpitopeHit> findEpitopes(String proteinSequence) throws Exception {
		return findEpitopes(proteinSequence,0);
	}

	@Override
	public List<EpitopeHit> findEpitopes(String proteinSequence, double minScore) throws Exception {
		if(matrices == null) loadMatrices();
		List <EpitopeHit> hits = new ArrayList<EpitopeHit>();
		for(PredictionPWM matrix:matrices) {
			int length = matrix.getLength();
			for(int j=0;j <= proteinSequence.length() - length; j++) {
				String epitope = proteinSequence.substring(j,j+length);
				int score = matrix.calculateScore(epitope);
				if(score >= minScore) {
					//System.out.println("Epitope: "+epitope+". Allele: "+matrix.getMotif()+". Score: "+score);
					hits.add(new EpitopeHit(matrix.getMotif(), epitope, j, score));
				}
			}
		}
		return hits;
	}

	public void loadMatrices() throws Exception {
		matrices = new ArrayList<PredictionPWM>();
		ResourceBundle databaseBundle = ResourceBundle.getBundle("epitopeFinder");
		String driver = databaseBundle.getString("database.driver");
		String url = databaseBundle.getString("database.url");
		String user = databaseBundle.getString("database.user");
		String password = databaseBundle.getString("database.password");
		String query = "SELECT HLATYP, MATRIX, TYP FROM HLATYP, EPIPREDICTION ";
		query += "WHERE HLATYP.ID = EPIPREDICTION.REF ";
		if(motifs!=null) {
			query += " AND (";
			for(int i=0;i<motifs.length;i++) {
				if(i>0) query += " OR ";
				query += " HLATYP LIKE ? ";
			}
			query += " ) ORDER BY HLATYP";
		}
		Connection conn = null;
		PreparedStatement stmt = null;
		ResultSet rs = null;
		try {
			Class.forName(driver);
			conn = DriverManager.getConnection(url, user, password);
			stmt = conn.prepareStatement(query);
			if(motifs!=null) {
				for(int i=0;i<motifs.length;i++) {
					stmt.setString(i+1, motifs[i]);
				}
			}
			rs = stmt.executeQuery();
			while (rs.next()) {
				String motif = rs.getString(1);
				InputStream stream = rs.getBinaryStream(2);
				int length = rs.getInt(3)+8;
				short [][] matrix = new short[26][length];
				//if(motif.contains("1510")) {
					//System.out.println("Motif: "+motif+". Length: "+length);
					fillMatrix(matrix,stream);
				//}
				PredictionPWM pred = new PredictionPWM(motif, matrix);
				matrices.add(pred);
			}
		} finally {
			if(rs != null) rs.close();
			if(stmt!=null) stmt.close();
			if(conn!=null) conn.close();
		}
	}

	private void fillMatrix(short[][] matrix, InputStream stream) throws IOException {
		int i=0;
		int j=0;
		for(int k=0;true;k++) {
			short next = (short) stream.read();
			if(next < 0) return;
			if(next>=128) {
				next -= 256;
			}
			if(k<32) continue;
			if(k%4 == 1) {
				//System.out.print(" " +next);
				matrix[i][j] = next;
				i++;
				if(i==matrix.length) {
					//System.out.println();
					j++;
					i=0;
					if(j==matrix[0].length) {
						return;
					}
				}
			}
		}
		
	}

	@Override
	public double getScore(String epitope, String motif) throws Exception {
		if(matrices == null) loadMatrices();
		for(PredictionPWM matrix:matrices) {
			if(matrix.getMotif().equals(motif) && epitope.length() == matrix.getLength()) {
				return matrix.calculateScore(epitope);
			}
		}
		return 0;
	}
	
	@Override
	public void setMotifs(String motifs) {
		this.motifs = motifs.split(",");
		
	}

       	@Override
        public void setAlgorithm(String algorithm){
        }

        @Override
	public Map<String, List<EpitopeHit>> findEpitopes(Set<String> proteinSequences, double minScore) throws Exception {
		Map<String, List<EpitopeHit>> answer = new TreeMap<String, List<EpitopeHit>>();
		for(String protein: proteinSequences) {
			answer.put(protein, findEpitopes(protein, minScore));
		}
		return answer;
	}
	public static void main(String [] args) throws Exception {
		DatabaseProteinEpitopeChecker checker = new DatabaseProteinEpitopeChecker();
		checker.setMotifs(args[0]);
		String [] motifs = args[0].split(",");
		FileInputStream fis = new FileInputStream(args[1]);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		PrintStream out = System.out;
		String line=in.readLine();
		for(int i=0;line!=null;i++) {
			String [] items = line.split(" |\t");
			for(String item:items) {
				if(item.trim().length()>0) {
					out.print(item.trim()+"\t");
				}
			}
			for(int j=0;j<motifs.length;j++) {
				if(j>0) out.print("\t");
				out.print(checker.getScore(items[2], motifs[j]));
			}
			out.println();
			line=in.readLine();
		}
		out.flush();
		fis.close();
	}
}
