package edu.uconn.bioinformatics.epitopeFinder;


import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.TreeSet;

import edu.uconn.bioinformatics.fileHandlers.HaplotypesHandler;
import edu.uconn.bioinformatics.mapping.PhasedBlock;
import edu.uconn.bioinformatics.mapping.PhasedGenome;
import edu.uconn.bioinformatics.mapping.PhasedSequence;
import edu.uconn.bioinformatics.mapping.SNPSHaplotyper;
import edu.uconn.ngstools.genome.GenomicRegionPositionComparator;
import edu.uconn.ngstools.main.SNVSFinder;
import edu.uconn.ngstools.sequences.DNASequence;
import edu.uconn.ngstools.transcriptome.CCDSTranscriptome;
import edu.uconn.ngstools.transcriptome.Exon;
import edu.uconn.ngstools.transcriptome.ProteinTranslator;
import edu.uconn.ngstools.transcriptome.Transcript;
import edu.uconn.ngstools.variants.CalledSNV;
import edu.uconn.ngstools.variants.SNV;
import edu.uconn.ngstools.variants.ClassifiedSNV;
import edu.uconn.ngstools.variants.SNVSFilter;
import edu.uconn.ngstools.variants.filehandlers.CalledSNVSFileHandler;
import edu.uconn.ngstools.variants.filehandlers.SNVSFileHandler;
import java.util.HashMap;
import java.util.Iterator;

public class EpitopeFinder {
	// Static fields.
	public static final String INPUT_TYPE_PROTEIN = "protein";
	public static final String INPUT_TYPE_FASTQ = "fastq";
	public static final String INPUT_TYPE_CALLED_SNPS = "called_snps";
	public static final String INPUT_TYPE_UNSORTED_READS = "unsorted_reads";
	public static final int DEF_MIN_SCORE = 8;
	
	private CCDSTranscriptome transcriptome;
	private ProteinTranslator translator = new ProteinTranslator();
	private List<ReportRecord> report = new ArrayList<ReportRecord>();
	private ProteinEpitopeChecker checker = new DummyProteinEpitopeChecker();
	private EpitopeHitComparator hitComparator = new EpitopeHitComparator();
	private List<CalledSNV> snvs;
	private PhasedGenome phasedGenome=null;
//sahar SNV stats
        private Map<String,ClassifiedSNV> classifiedCodingRegionSNVs;// = new HashMap<String,List<ClassifiedSNV>>();
//      private List<ClassifiedSNV> classifiedCodingRegionSNVs;
        private boolean noprediction = false;
	
	//Input files
	private String inputType;
	private String snpsFilename;
	private String samFilename;
	private String ccdsFilename;
	private String cdnaFilename;

	
	//Optional parameters
	private String outFilename;
	private String snvsOutFilename;
	private String phaseOutFilename;
	private String phaseInFilename;
	private double minAltCoverage=SNVSFilter.DEF_MIN_ALT_COVERAGE;
	private double maxAltCoverage=SNVSFilter.DEF_MAX_ALT_COVERAGE;
	private double minProbability=SNVSFilter.DEF_MIN_PROBABILITY;
	private String snpsToExcludeFilename;
	private int minScore = DEF_MIN_SCORE;
	private boolean processHomozygous = false;
	private boolean processTriallelic = false;
	private int readingFrame = 0;
	
	public boolean validateParameters() {
		boolean answer =true;
		if(outFilename!=null && (ccdsFilename==null || cdnaFilename==null) ) {
			System.err.println("CCDS and CDNA files required to generate an output report");
			answer = false;
		}
		return answer;
	}
	
//	public void runProcess () throws IOException {
	public void runProcess () throws Exception {
            if(ccdsFilename!=null) transcriptome = new CCDSTranscriptome(ccdsFilename);
		if(cdnaFilename!=null) transcriptome.loadSequences(cdnaFilename);
		loadSNPS();
		if (snvsOutFilename!=null) {
			printSnvs();
		}
		loadPhasing();
		if (phaseOutFilename!=null) {
			printHaplotypes();
		}
//sahar SNV stats
                classifiedCodingRegionSNVs = new HashMap<String,ClassifiedSNV>(); //new ArrayList<ClassifiedSNV>();
		if(this.outFilename!=null) {
			try {
				findEpitopes();
			} catch (Exception e) {
				e.printStackTrace();
			}
			printReport();
                        printClassifiedSNVLists();
		}
	}
	
	
        public void printClassifiedSNVLists () throws IOException{

            int numHeterozygous=0;
            int numNonSynonymous=0;
            int numReadthru=0;
            int numEarlyStop=0;
            PrintStream out = new PrintStream(outFilename+".snvStat");
            Set keys = classifiedCodingRegionSNVs.keySet();
            Iterator i = keys.iterator();
            while(i.hasNext())
            {
               String key = (String) i.next();
               ClassifiedSNV snv = (ClassifiedSNV) classifiedCodingRegionSNVs.get(key);
                    out.print(snv.getSequenceName()+"_"+snv.getPosition()+"\t"+snv.getRefBase()+"\t"+snv.getAltBase()+"\t"+snv.getGenotype()+"\t");
                    if (snv.isHeterozygous()) {
                        numHeterozygous++;
                        out.print("HT\t");
                    }
                    else
                        out.print("HO\t");
                    if (snv.isNonSynonymous())
                        out.print("NonSyn\t");
                    else
                        out.print("Syn\t");
                    out.println(snv.getType());

                    if (snv.isHeterozygous() && snv.isNonSynonymous())
                        numNonSynonymous++;
                    if (snv.getType().equals(ClassifiedSNV.Nonsense))
                        numEarlyStop++;
                    if (snv.getType().equals(ClassifiedSNV.NonStop))
                        numReadthru++;
            }

            out.println("Number of SNVs in coding regions = "+classifiedCodingRegionSNVs.size());
            out.println("Number of heterozygous SNVs in coding regions = "+numHeterozygous);
            out.println("Number of heterozygous nonSynonymousSNVs in coding regions = "+numNonSynonymous);
            out.println("Number of mutations in STOP codon = "+numReadthru);
            out.println("Number of mutations causing early STOP codon = "+numEarlyStop);
            System.out.println("Number of SNVs in coding regions = "+classifiedCodingRegionSNVs.size());
            System.out.println("Number of heterozygous SNVs in coding regions = "+numHeterozygous);
            System.out.println("Number of heterozygous nonSynonymousSNVs in coding regions = "+numNonSynonymous);
            out.println("Number of mutations in STOP codon = "+numReadthru);
            out.println("Number of mutations causing early STOP codon = "+numEarlyStop);
            out.flush();
            out.close();
        }

	public void loadSNPS () throws IOException {
		SNVSFilter snvsFilter = new SNVSFilter(this.minAltCoverage,this.maxAltCoverage,this.minProbability);
		if(snpsToExcludeFilename!=null) {
			SNVSFileHandler snvsHandler = new CalledSNVSFileHandler();
			List<SNV> snpsToExclude = snvsHandler.loadSNVS(snpsToExcludeFilename);
			snvsFilter.setSnvsToExclude(snpsToExclude);
		}
		if(INPUT_TYPE_UNSORTED_READS.equalsIgnoreCase(inputType)){
			SNVSFinder snvsFinder = new SNVSFinder();
			snvsFinder.setFilter(snvsFilter);
			snvsFinder.findSNVS(samFilename, transcriptome);
			System.out.println("Called "+snvs.size()+" snvs"); 
		} else if (INPUT_TYPE_CALLED_SNPS.equalsIgnoreCase(inputType)) {
			CalledSNVSFileHandler snvsHandler = new CalledSNVSFileHandler();
			snvs = snvsHandler.loadCalledSNVS(snpsFilename);
			snvs = snvsFilter.filterSNVS(snvs);
			System.out.println("Loaded "+snvs.size()+" snvs");
		}
	}
	
//	public void loadPhasing() throws IOException {
	public void loadPhasing() throws Exception {
            if(INPUT_TYPE_UNSORTED_READS.equalsIgnoreCase(inputType)){
			SNPSHaplotyper haplotyper = new SNPSHaplotyper();
			haplotyper.setSnvs(snvs);
			phasedGenome = haplotyper.phaseSNPs(samFilename);
		} else if (phaseInFilename!=null) {
			HaplotypesHandler handler = new HaplotypesHandler();
			phasedGenome = handler.loadHaplotypes(snvs, phaseInFilename);
			System.out.println("Loaded phased genome. Sequences: "+phasedGenome.getNumSequences());
		}
	}
	
	public void findEpitopes() throws Exception {
//temp; set to generate snvStats and return
//        boolean noprediction = true;

            //Find candidates for report
		List<ReportRecord> candidates = new ArrayList<ReportRecord>();
		for(int i=0;i<snvs.size();i++) {
			CalledSNV snv = snvs.get(i);
			if(!processHomozygous) {
				if(snv.isHomozygous()) {
//sahar SNV stats
                                    // if this SNV will not be processed further; check if in coding region now and add it to list if so
                                    List<Exon> posExons = transcriptome.getExons(snv.getSequenceName(), snv.getPosition());
                                    if(posExons.size()!=0) {
                                        String key = snv.getSequenceName()+Integer.toString(snv.getPosition());
                                        if (!classifiedCodingRegionSNVs.containsKey(key))
//                                            classifiedCodingRegionSNVs.add(new ClassifiedSNV(snv.getSequenceName(),snv.getPosition(),snv.getRefBase(),snv.getAltBase(),snv.getGenotype()));
                                            classifiedCodingRegionSNVs.put(key,new ClassifiedSNV(snv.getSequenceName(),snv.getPosition(),snv.getRefBase(),snv.getAltBase(),snv.getGenotype()));
                                    }
                                    continue;
				}
			}
			if(!processTriallelic) {
				if(snv.isTriallelic()) {
//sahar SNV stats
                                    // if this SNV will not be processed further; check if in coding region now and add it to list of so
                                    List<Exon> posExons = transcriptome.getExons(snv.getSequenceName(), snv.getPosition());
                                    if(posExons.size()!=0) {
                                        String key = snv.getSequenceName()+Integer.toString(snv.getPosition());
                                        if (!classifiedCodingRegionSNVs.containsKey(key))
//                                            classifiedCodingRegionSNVs.add(new ClassifiedSNV(snv.getSequenceName(),snv.getPosition(),snv.getRefBase(),snv.getAltBase(),snv.getGenotype()));
                                            classifiedCodingRegionSNVs.put(key,new ClassifiedSNV(snv.getSequenceName(),snv.getPosition(),snv.getRefBase(),snv.getAltBase(),snv.getGenotype()));
                                    }
                                    continue;
				}
			}

                        List<ReportRecord> snvCandidates;
                        try {
//				List<ReportRecord> snvCandidates = getCandidates(snv,i);
                            snvCandidates = getCandidates(snv,i);
                        }
                        catch (Exception e) {
                            System.err.println("Error processing snv "+snv.getSequenceName()+" "+snv.getPosition()+" "+snv.getGenotype());
                            e.printStackTrace(System.err);
                            continue;    
                        }
			candidates.addAll(snvCandidates);
		}
//sahar added for calculating SNV statistics only
                if (noprediction)
                    return;
		//Perform epitope predictions
		Set<String> proteinSequences = new TreeSet<String>();
		for(ReportRecord record:candidates) {
			proteinSequences.add(record.getReferenceProtein());
			proteinSequences.add(record.getMutatedProtein());
		}
		Map<String,List<EpitopeHit>> hitsMap = checker.findEpitopes(proteinSequences, minScore);
		//Fill report with hits
		for(ReportRecord record:candidates) {
//                        int MutProteinSize = record.getMutatedProtein().length();
			List<EpitopeHit> refHitCandidates = hitsMap.get(record.getReferenceProtein());
			if(refHitCandidates==null) {
				refHitCandidates=new ArrayList<EpitopeHit>();
			} else {
				Collections.sort(refHitCandidates,hitComparator);
			}
			List<EpitopeHit> hitCandidates = hitsMap.get(record.getMutatedProtein());
			if(hitCandidates==null) {
				hitCandidates=new ArrayList<EpitopeHit>();
			} else {
				Collections.sort(hitCandidates,hitComparator);
			}
			//System.out.println("refHitCandidates.size(), hitCandidates.size()"+refHitCandidates.size()+" "+hitCandidates.size());
			//System.out.println("refProtein, MutProtein "+record.getReferenceProtein()+" "+record.getMutatedProtein());

			if(refHitCandidates.size() > 0  || hitCandidates.size() > 0) {
				int j=0;
				int k=0;
				while(j<refHitCandidates.size() && k < hitCandidates.size()) {
					EpitopeHit hitRef = refHitCandidates.get(j);
					EpitopeHit hitMod = hitCandidates.get(k);
					int comp = hitComparator.compare(hitRef, hitMod);
					if(comp < 0) {
//                                            if (hitMod.getEpitope().startsWith("AYCETCWE") || hitRef.getEpitope().startsWith("AYCETCWE")) {
//						      System.out.println("in comp<0");
//                                                System.out.println("j,k "+j+k);
//                                                System.out.println("Ref = "+hitRef.getEpitope()+" "+hitRef.getPosition());
//                                                System.out.println("Mut = "+hitMod.getEpitope()+" "+hitMod.getPosition());
//                                            }
						processHit(record,hitRef,true);
						j++;
					} else if (comp > 0) {
//                                            if (hitMod.getEpitope().startsWith("AYCETCWE") || hitRef.getEpitope().startsWith("AYCETCWE")) {
//						      System.out.println("in comp>0");
//                                                System.out.println("j,k "+j+k);
//                                                System.out.println("Ref = "+hitRef.getEpitope()+" "+hitRef.getPosition());
//                                                System.out.println("Mut = "+hitMod.getEpitope()+" "+hitMod.getPosition());
//                                            }
						processHit(record,hitMod,false);
						k++;
					} else if(!hitRef.getEpitope().equals(hitMod.getEpitope())){
//                                            if (hitMod.getEpitope().startsWith("AYCETCWE") || hitRef.getEpitope().startsWith("AYCETCWE"))
//                                            {
//					      System.out.println("in Ref != Mod");
//                                               System.out.println("j,k "+j+k);
//                                                System.out.println("Ref = "+hitRef.getEpitope()+" "+hitRef.getPosition());
//                                                System.out.println("Mut = "+hitMod.getEpitope()+" "+hitMod.getPosition());
//                                             }
						hitMod.setReferenceEpitope(hitRef.getEpitope());
						hitMod.setReferenceScore(hitRef.getScore());
                                                int epitopeStart=hitMod.getPosition();
                                                int epitopeEnd = epitopeStart+hitMod.getEpitope().length();
                                                // check if predicted mutant epitope contains the mutation
                                                // position returned by NetMHC is 1 based
                                                //mutation in the protein in at MAX_EPITOPE_LENGTH  on 1 based index
//                                                if (hitMod.getEpitope().startsWith("AYCETCWE"))
//                                                    System.out.println("In epitopes equal else");
                                                if (epitopeStart <= (hitMod.MAX_EPITOPE_LENGTH) && epitopeEnd >= (hitMod.MAX_EPITOPE_LENGTH))
                                                       record.addHit(hitMod);
//                                                else
//                                                        System.out.println("filtered out epitope "+hitMod.getEpitope()+" epitope start = "+epitopeStart+" epitope Enf = "+epitopeEnd);

						j++;
						k++;
					} else {
//                                            if (hitMod.getEpitope().startsWith("AYCETCWE") || hitRef.getEpitope().startsWith("AYCETCWE"))
//                                            {
//						      System.out.println("in lase else; skip epitope");
//                                                System.out.println("j,k "+j+k);
//                                                System.out.println("Ref = "+hitRef.getEpitope()+" "+hitRef.getPosition());
//                                                System.out.println("Mut = "+hitMod.getEpitope()+" "+hitMod.getPosition());
//                                               }
						j++;
						k++;
					}
				}
				while(j<refHitCandidates.size()) {
					EpitopeHit hitRef = refHitCandidates.get(j);
//						      System.out.println("j<refHitCandidates.size()");
//                                                System.out.println("j "+j);
//                                                System.out.println("Ref = "+hitRef.getEpitope()+" "+hitRef.getPosition());
					processHit(record,hitRef,true);
					j++;
				}
				while(k<hitCandidates.size()) {
					EpitopeHit hitMod = hitCandidates.get(k);
//						      System.out.println("k<refHitCandidates.size()");
//                                                System.out.println("k "+k);
//                                                System.out.println("Mut = "+hitMod.getEpitope()+" "+hitMod.getPosition());
					processHit(record,hitMod,false);
					k++;
				}
				report.add(record);
			}
		}
	}
	private List<ReportRecord> getCandidates(CalledSNV snv, int posList) throws Exception {

//sahar moved up (was futher down in this method) to eliminate all block and nearby homozygous SNV lookup; if current SNV is not in coding region
		List<ReportRecord> candidates = new ArrayList<ReportRecord>();
		List<Exon> posExons = transcriptome.getExons(snv.getSequenceName(), snv.getPosition());
		if(posExons.size()==0) {
			return candidates;
		}
//sahar end


                //Get phasing information

		//System.out.println("SNV: "+snv.getPosition());
		PhasedBlock block = null;
		List <CalledSNV> snvsZeroAllele = new ArrayList<CalledSNV>();
		List <CalledSNV> snvsOneAllele = new ArrayList<CalledSNV>();
		List <CalledSNV> triallelicZero = new ArrayList<CalledSNV>();
		List <CalledSNV> triallelicOne = new ArrayList<CalledSNV>();
                List <CalledSNV> snvsHomozygous = new ArrayList<CalledSNV>();
		if(phasedGenome!=null) {
			PhasedSequence pSeq = phasedGenome.getPhasedSequence(snv.getSequenceName());
			if(pSeq!=null) {
//                                if (snv.getPosition() == 36302817)
//                                    System.out.println("Seqname "+pSeq.getSequenceName());
				block = pSeq.findBlock(snv);

				if(block!=null) {
					List <CalledSNV> snvsInBlock = block.getSnvs();
					List <CalledSNV> snvsInPhase = block.findSNVsInPhase(snv);
					int i=0;
					int j=0;
					while(i<snvsInBlock.size() && j<snvsInPhase.size()) {
						CalledSNV blockSNV = snvsInBlock.get(i);
						CalledSNV phasedSNV = snvsInPhase.get(j);
						if(phasedSNV.getPosition() < blockSNV.getPosition()) {
							throw new RuntimeException("Error loading phased block inforamtion");
						} else if (phasedSNV.getPosition() > blockSNV.getPosition()) {
							//block snv not in phase. Add it to zero allele lists
							if(blockSNV.isTriallelic()) {
								triallelicZero.add(blockSNV);
							} else {
								snvsZeroAllele.add(blockSNV);
//                                                                System.out.println("adding NOTInphase SNV at "+blockSNV.getPosition()+" "+blockSNV.getGenotype());
							}
							i++;	
						} else {
							//Near snv in phase. Add it to one allele lists
							if(blockSNV.isTriallelic()) {
								triallelicOne.add(blockSNV);
							} else {
								snvsOneAllele.add(blockSNV);
//                                                                System.out.println("adding Inphase SNV at "+blockSNV.getPosition()+" "+blockSNV.getGenotype());
							}
							i++;
							j++;
						}
					}
					for(;i<snvsInBlock.size();i++) {
						//Remaining snvs not in phase
//                                                System.out.println("adding rest NOTInphase SNVs in block");
						CalledSNV nearSNV = snvsInBlock.get(i);
						if(nearSNV.isTriallelic()) {
							triallelicZero.add(nearSNV);
						} else {
							snvsZeroAllele.add(nearSNV);
//                                                        System.out.println("adding NOTInphase SNV at "+nearSNV.getPosition()+" "+nearSNV.getGenotype());
						}
					}
				}
			}
		}
		if(snvsOneAllele.size()==0) {
			snvsOneAllele.add(snv);
		}
//                System.out.println("number of snvsZeroAllele "+snvsZeroAllele.size());
//                System.out.println("number of snvsOneAllele "+snvsOneAllele.size());
		//Add homozygous not reference  
		List <CalledSNV> nearbySNVS = getNearbySnvs(block,posList);
		for(CalledSNV nearSNV:nearbySNVS) {
			if(nearSNV.isHomozygous()) {
				snvsZeroAllele.add(nearSNV);
				snvsOneAllele.add(nearSNV);
//sahar WT = reference + homozygous alternative
                                if(!processHomozygous)
                                    snvsHomozygous.add(nearSNV);
			}
		}
//                    System.out.println("SNVs nearby: "+nearbySNVS.size());
//                    System.out.println("SNVs zero: "+snvsZeroAllele.size());
//                    System.out.println("SNVs one: "+snvsOneAllele.size());
//                    System.out.println("Triallelic zero: "+triallelicZero.size());
//                    System.out.println("Triallelic one: "+triallelicOne.size());
//sahar move up and eliminate all block and nearby homozygous SNV lookup; if current SNV is not in coding region
//		List<ReportRecord> candidates = new ArrayList<ReportRecord>();
		List<String> cdnaSeqs = new ArrayList<String>();
//		List<Exon> posExons = transcriptome.getExons(snv.getSequenceName(), snv.getPosition());
//		if(posExons.size()==0) {
//			return candidates;
//		}
//sahar end
		for(Exon e:posExons) {
			for(Transcript t:e.getGene().getTranscripts()) {
                            if (snv.getSequenceName().contains("random")) {
                                System.out.println("SNV pn random chromosome"+t.getId());
                            }

				int relPos = t.getRelativeTranscriptPosition(snv.getPosition());
				if(relPos<0) {
					continue;
				}
                                String cDNASeqRef;
                                try {
//				String cDNASeqRef = t.getCDNASequence().toString();
				cDNASeqRef = t.getCDNASequence().toString();
                            }
                                catch (Exception Exp)
                                {
                                   //System.out.println("exon "+e.getSequenceName()+" "+e.getStart()+" "+e.getEnd());
                                   System.err.println("Error: No available cDNA for transcript "+t.getId());                                                                      
                                    //System.err.println("gene "+t.getGene().getName());
                                    //System.err.println("cDNA "+t.getCDNASequence());
                                    Exp.printStackTrace(System.err);
                                    continue;
                                }
				int beginPos = relPos-3*(EpitopeHit.MAX_EPITOPE_LENGTH-1);
				int endPos = relPos+3*EpitopeHit.MAX_EPITOPE_LENGTH;
//sahar added  mutPositionInProtein for nonSyn mutation detection
                                int mutPositionInProtein;
				if(beginPos<0) {
                                        mutPositionInProtein = (beginPos + 3*(EpitopeHit.MAX_EPITOPE_LENGTH-1))/3;
					beginPos = 0;
				}
                                else
                                    mutPositionInProtein = EpitopeHit.MAX_EPITOPE_LENGTH-1;
				beginPos-=beginPos%3;
				if(endPos>cDNASeqRef.length()) {
					endPos = cDNASeqRef.length();
				}
				endPos-=endPos%3;
				cDNASeqRef = cDNASeqRef.substring(beginPos,endPos);
//sahar WT = reference + homozygous alternative
				char [] cdnaSeqArray = cDNASeqRef.toCharArray();
				applySNVs(snvsHomozygous,t,beginPos,false,cdnaSeqArray);
                                String cDNASeqWT = new String(cdnaSeqArray);
				boolean found = false;
				for(String seq:cdnaSeqs ) {
					if(seq.equals(cDNASeqRef)) {
						found = true;
						break;
					}
				}
				if(!found) {
					cdnaSeqs.add(cDNASeqRef);
					//System.out.println("Sequence  ref: "+cDNASeqRef);
					char [] cdnaSeqZeroAllele = cDNASeqRef.toCharArray();
					applySNVs(snvsZeroAllele,t,beginPos,false,cdnaSeqZeroAllele);
					applySNVs(triallelicZero,t,beginPos,false,cdnaSeqZeroAllele);
					applySNVs(triallelicOne,t,beginPos,true,cdnaSeqZeroAllele);
					String seqZeroAllele = new String(cdnaSeqZeroAllele);
					//System.out.println("Sequence zero: "+seqZeroAllele);
					char [] cdnaSeqOneAllele = cDNASeqRef.toCharArray();
					applySNVs(snvsOneAllele,t,beginPos,false,cdnaSeqOneAllele);
					applySNVs(triallelicZero,t,beginPos,true,cdnaSeqOneAllele);
					applySNVs(triallelicOne,t,beginPos,false,cdnaSeqOneAllele);
					String seqOneAllele=new String(cdnaSeqOneAllele);
					//System.out.println("Sequence  one: "+seqOneAllele);
					String WTProtein = getProtein(cDNASeqWT, readingFrame);
					String hap2Protein = getProtein(seqZeroAllele, readingFrame);
					String hap1Protein = getProtein(seqOneAllele, readingFrame);

//					System.out.println(WTProtein);
//					System.out.println(hap2Protein);
//					System.out.println(hap1Protein);
//                                        System.out.println("mutPositionInProtein = "+mutPositionInProtein+"referenceProtein length = "+WTProtein.length());
//sahar: SNV stats
					char refCodon;
					if (WTProtein.length() <= mutPositionInProtein) //mutation in stop codon
                                            refCodon = 'X';
					else
                                            refCodon = WTProtein.charAt(mutPositionInProtein);
                                        char hap2Codon;
                                        boolean hap2SNP_nonsysn;
					boolean hap2SNP_readthru = false;
                                        boolean hap2SNP_earlyStop = false;
					if (hap2Protein.length() == WTProtein.length()) {
                                            if (refCodon == 'X') //equal length and mutation in stop codon
						hap2SNP_nonsysn=false;
                                            else {
                                                if (hap2Protein.length() <= mutPositionInProtein)
                                                    hap2Codon='X';
						else
	                                            hap2Codon=hap2Protein.charAt(mutPositionInProtein);
                                                if (hap2Codon==refCodon)
                                                    hap2SNP_nonsysn=false;
                                                else
                            	                    hap2SNP_nonsysn=true;
						}
                                        }
                                        else {
						if (hap2Protein.length() == mutPositionInProtein){ //current mutation is the cause of the early stop codon; protein stopped at this mutated codon
                                                    hap2SNP_nonsysn = true;
                                                    hap2SNP_earlyStop=true;
                                                }
                                                else
                                                    if (WTProtein.length() == mutPositionInProtein) {
							hap2SNP_nonsysn = true;
							hap2SNP_readthru = true;
						      }
                                                    else
							hap2SNP_nonsysn = false;
                                                    }


                                        char hap1Codon;
                                        boolean hap1SNP_nonsysn;
                                        boolean hap1SNP_readthru = false;
                                        boolean hap1SNP_earlyStop = false;
                                        if (hap1Protein.length() == WTProtein.length()) {
					  if (refCodon == 'X') //equal length and mutation in stop codon
						hap1SNP_nonsysn=false;
					  else {
                                                if (hap1Protein.length() <= mutPositionInProtein)
                                                    hap1Codon='X';
						else
                                                    hap1Codon=hap1Protein.charAt(mutPositionInProtein);
                                                if (hap1Codon==refCodon)
                                                    hap1SNP_nonsysn=false;
                                                else
                            	                    hap1SNP_nonsysn=true;
                                           }
                                        }
                                        else {
					  if (hap1Protein.length() == mutPositionInProtein) { //current mutation is the cause of the early stop codon; protein stopped at this mutated codon
                                                hap1SNP_nonsysn = true;
                                                hap1SNP_earlyStop = true;
                                          }
                                          else
						if (WTProtein.length() == mutPositionInProtein) {
							hap1SNP_nonsysn = true;
							hap1SNP_readthru = true;
                                                }
                                                else
							hap1SNP_nonsysn = false;
					  }
                                          if (hap1SNP_nonsysn) {
//                                        if(!referenceProtein.equals(hap1Protein)) {
						ReportRecord record = new ReportRecord(snv,t);
						record.setReferenceProtein(WTProtein);
						record.setMutatedProtein(hap1Protein);
						record.setTranscriptPosition(relPos/3+1);
						record.setNearbySnvs(nearbySNVS);
						candidates.add(record);
					}
					if(hap2SNP_nonsysn  /*!referenceProtein.equals(hap2Protein)*/&& !hap1Protein.equals(hap2Protein)) { //if mutation resutled in the same hap1 protein after translation
						ReportRecord record = new ReportRecord(snv,t);
						record.setReferenceProtein(WTProtein);
						record.setMutatedProtein(hap2Protein);
						record.setTranscriptPosition(relPos/3+1);
						record.setNearbySnvs(nearbySNVS);
						candidates.add(record);
					}
                                        ClassifiedSNV class_snv = new ClassifiedSNV(snv.getSequenceName(),snv.getPosition(),snv.getRefBase(),snv.getAltBase(),snv.getGenotype());
                                        if (hap1SNP_nonsysn || hap2SNP_nonsysn) {
                                            class_snv.setNonSynonymous(true);
                                            class_snv.setType(class_snv.Missense);
                                        }
                                        if (hap1SNP_readthru || hap2SNP_readthru)
                                              class_snv.setType(class_snv.NonStop);
                                        if (hap1SNP_earlyStop || hap2SNP_earlyStop)
                                              class_snv.setType(class_snv.Nonsense);
//                                            readthruMutations++;
                                        String key = snv.getSequenceName()+Integer.toString(snv.getPosition());
                                        if (!classifiedCodingRegionSNVs.containsKey(key))
//                                            classifiedCodingRegionSNVs.add(class_snv);
                                            classifiedCodingRegionSNVs.put(key,class_snv);
                                        else if (class_snv.isNonSynonymous()){
                                            ClassifiedSNV saved_class_snv = classifiedCodingRegionSNVs.get(key);
                                            if (class_snv.moreInterstingCategory(saved_class_snv)) {
                                                classifiedCodingRegionSNVs.remove(key);
                                                classifiedCodingRegionSNVs.put(key,class_snv);
                                            } //if more interesting

                                        } //else if nonsysnonymous

				}//if !found
			}// for each transcript
		} // for each exon
		return candidates;
	}
	private void applySNVs(List<CalledSNV> snvs, Transcript transcript, int cDNARefSubseqStart, boolean triallelicFirst, char [] cdnaSeqMod) {
		for(CalledSNV snv: snvs) { 
			applySNV(snv, transcript, cDNARefSubseqStart,triallelicFirst, cdnaSeqMod);
		}
	}
	private void applySNV(CalledSNV snv, Transcript transcript, int cDNARefSubseqStart, boolean triallelicFirst, char [] cdnaSeqMod) {
		int relPos = transcript.getRelativeTranscriptPosition(snv.getPosition())-cDNARefSubseqStart;
		if(relPos>=0 && relPos<cdnaSeqMod.length) {
			char ref = snv.getRefBase();
			char alt = snv.getAltBase();
			if(snv.isTriallelic()) {
				alt = snv.getGenotype().charAt(1);
				if(triallelicFirst) {
					alt = snv.getGenotype().charAt(0);
				}
			}
			if(transcript.isReverse()) {
				ref = DNASequence.getComplement(ref);
				alt = DNASequence.getComplement(alt);
			}
                        try{
                            if(cdnaSeqMod[relPos]!=ref) {

                                    System.err.println("Unexpected reference base "+cdnaSeqMod[relPos]+" found. Expected: "+ref+" for SNV in chromosome: "+snv.getSequenceName()+" and position:"+snv.getPosition());
//                                    System.err.println("Transcript "+transcript.getId()+" SNV position in transcript "+transcript.getRelativeTranscriptPosition(snv.getPosition()));
                                    throw new Exception();
                            }
                        }
                        catch (Exception e) {
                            e.printStackTrace();
                        }
			cdnaSeqMod[relPos]=alt;
		}
	}
	private List<CalledSNV> getNearbySnvs(PhasedBlock block, int pos) {
		Set<CalledSNV> nearbySNVs= new TreeSet<CalledSNV>(GenomicRegionPositionComparator.getInstance());
		SNV snv = snvs.get(pos);
		//Add snvs in the same block different than the snv to consider
		if(block!=null) {
//                        System.out.println("In getNearbySnvs, before first loop");
//			System.out.println("SNVs in block: "+block.getSnvs().size());
//                        System.out.println("Current SNV: "+snv.getPosition());
			for(CalledSNV snvInBlock:block.getSnvs()) {
				if(snvInBlock.getPosition()!=snv.getPosition()) {
//					System.out.println("Adding: "+snvInBlock.getPosition()+" "+snvInBlock.getGenotype());
					nearbySNVs.add(snvInBlock);
				}
			}
		}
		int distance = 3*EpitopeHit.MAX_EPITOPE_LENGTH;
		//Look back
//                System.out.println("In getNearbySnvs, before loop back");
		for(int i=pos-1;i>=0;i--) {
			CalledSNV snv2 = snvs.get(i);
			if(snv.getSequenceName().equals(snv2.getSequenceName()) && snv.getPosition()-snv2.getPosition()<=distance) {
//				System.out.println("Adding: "+snv2.getPosition()+" "+snv2.getGenotype()+" i = "+i);
				nearbySNVs.add(snv2);
			} else {
				break;
			}
		}
		//Look forward
//                System.out.println("In getNearbySnvs, before loop forward");
		for(int i=pos+1;i<snvs.size();i++) {
			CalledSNV snv2 = snvs.get(i);
			if(snv.getSequenceName().equals(snv2.getSequenceName()) && snv2.getPosition()-snv.getPosition()<=distance) {
//				System.out.println("Adding: "+snv2.getPosition()+" "+snv2.getGenotype()+" i = "+i);
				nearbySNVs.add(snv2);
			} else {
				break;
			}
		}
		List<CalledSNV> answer = new ArrayList<CalledSNV>();
		answer.addAll(nearbySNVs);
		return answer;
	}

	public void processHit (ReportRecord record, EpitopeHit hit, boolean refHit) throws Exception {
                if (hit.isAdded())
                    return;
//                if (hit.getEpitope().startsWith("AYCETCWE")) {
//                           System.out.println("In processHit refHit = "+refHit);
//                        System.out.println("In processHit epitope = "+hit.getEpitope());
//                        System.out.println("In processHit refEpitope = "+hit.getReferenceEpitope());
//                        System.out.println("In processHit position = "+hit.getPosition());
//
//                }
		if(refHit) {
			hit.setReferenceEpitope(hit.getEpitope());
			hit.setReferenceScore(hit.getScore());
			hit.setEpitope(getEpitope(record.getMutatedProtein(), hit.getPosition(), hit.getEpitope().length())); 
		} else {
			hit.setReferenceEpitope(getEpitope(record.getReferenceProtein(), hit.getPosition(), hit.getEpitope().length()));
		}
                //if (hit.getEpitope().equals("AFSDA") || hit.getEpitope().equals("AFVATQQGATV")) {
                //        System.out.println("In processHit after Change epitope = "+hit.getEpitope());
                //        System.out.println("In processHit after Change refEpitope = "+hit.getReferenceEpitope());
		//	   System.out.println();

                //}
		if(!hit.getReferenceEpitope().equals(hit.getEpitope())) {
			if(refHit) {
				if(hit.getEpitope().length()== hit.getReferenceEpitope().length()) {
					hit.setScore(checker.getScore(hit.getEpitope(), hit.getMotif()));
				} else {
					hit.setScore(0);
				}
				
			} else {
				hit.setReferenceEpitope(getEpitope(record.getReferenceProtein(), hit.getPosition(), hit.getEpitope().length()));
				if(hit.getReferenceEpitope().length()== hit.getEpitope().length()) {
					hit.setReferenceScore(checker.getScore(hit.getReferenceEpitope(), hit.getMotif()));
				}
			}
                        int epitopeStart=hit.getPosition();
                        int epitopeEnd = epitopeStart+hit.getEpitope().length();
//sahar: Fix for incorrect mutation ID associated with epitope
                        // check if predicted mutant epitope contains the mutation
                        // position returned by NetMHC is 1 based
                         //mutation in the protein is at MAX_EPITOPE_LENGTH 1 based index
                         if (epitopeStart <= (hit.MAX_EPITOPE_LENGTH) && epitopeEnd >= (hit.MAX_EPITOPE_LENGTH )) {
                            hit.setAdded(true);
                            record.addHit(hit);
                        }
  //                      else
  //                          System.out.println("filtered out epitope "+hit.getEpitope()+" epitope start = "+epitopeStart+" epitope Enf = "+epitopeEnd);
		}
	}
	public String getEpitope(String protein,int start, int length) {
		if(protein.length() > start) {
			return protein.substring(start,Math.min(protein.length(), start+length));
		}
		return "";
	}
	public void printSnvs() throws IOException{
		CalledSNVSFileHandler handler = new CalledSNVSFileHandler();
		PrintStream out = new PrintStream(snvsOutFilename);
		handler.saveSNVS(snvs, out);
		out.close();
	}
	public void printHaplotypes() throws IOException {
		HaplotypesHandler handler = new HaplotypesHandler();
		PrintStream out = new PrintStream(phaseOutFilename);
		handler.printHaplotypes(phasedGenome, out);
		out.close();
	}
	public void printReport() throws IOException {	
		PrintStream out = new PrintStream(outFilename);
//debug galaxy
//                System.out.println("In printReport. outFilename = "+outFilename);
                Map<String,List<String>> printedReportEpitopes = new HashMap<String,List<String>>(); // map of epitope seq and transcript ID

		for(ReportRecord record: report) {
//debug galaxy
//                System.out.println("In printReport ReportRecord loop");
			printedReportEpitopes = record.printRecord(out,printedReportEpitopes,minScore);
                        //System.out.println("In printReport. Size of printedReportEpitopes = "+printedReportEpitopes.size());
		}
		out.flush();
		out.close();
	}
	public void processFastq(String fastqFilename) throws Exception {
		BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(fastqFilename)));
		String line=in.readLine();
		boolean sequence = false;
		while(line!=null) {
			if(sequence ==true && line.length()>0) {
				processSequence(line);
				sequence=false;
			} else if(line.length()>0 && line.charAt(0)=='@') {
				sequence=true;
			}
			line=in.readLine();
		}
		in.close();
	}
	public void processSequence(String dnaSequence) throws Exception {
		for (int i=0;i<3;i++) {
			String protein = getProtein(dnaSequence,i);
			getHits(protein);
		}
	}
	public String getProtein(String dnaSequence, int initialPosition) throws Exception {
		return String.valueOf(translator.getProteinSequence(dnaSequence.toCharArray(),initialPosition, false));
	}
	public List<EpitopeHit> getHits(String proteinSequence) throws Exception {
		if(proteinSequence.length()>=EpitopeHit.MIN_EPITOPE_LENGTH ) {
			return checker.findEpitopes(proteinSequence,minScore);
		}
		return new ArrayList<EpitopeHit>();
	}
	public void printHits(List<EpitopeHit> hits, PrintStream out) {
		for(EpitopeHit hit: hits) {
			out.print(hit.getMotif()+"\t");
			out.print(hit.getEpitope()+"\t");
			out.print(hit.getPosition()+"\t");
			out.println(hit.getScore()+"\t");
		}
	}
	

	public static void main(String [] args) throws Exception{
		long time1 = System.currentTimeMillis();
		EpitopeFinder finder = new EpitopeFinder();
		//Load optional parameters
		String motifs=null;
                String algorithm=null;
		int i=0;
		while(i<args.length && args[i].charAt(0)=='-') {
			if("-outFilename".equalsIgnoreCase(args[i])) {
				i++;
				finder.outFilename = args[i];
			} else if("-snvsOutFilename".equalsIgnoreCase(args[i])) {
				i++;
				finder.snvsOutFilename = args[i];
			} else if("-phaseOutFilename".equalsIgnoreCase(args[i])) {
				i++;
				finder.phaseOutFilename = args[i];
			} else if("-phaseInFilename".equalsIgnoreCase(args[i])) {
				i++;
				finder.phaseInFilename = args[i];
			} else if("-minAltCoverage".equalsIgnoreCase(args[i])) {
				i++;
				finder.minAltCoverage = Double.parseDouble(args[i]);
			} else if("-maxAltCoverage".equalsIgnoreCase(args[i])) {
				i++;
				finder.maxAltCoverage = Double.parseDouble(args[i]);
			} else if("-minProbability".equalsIgnoreCase(args[i])) {
				i++;
				finder.minProbability = Double.parseDouble(args[i]);
			} else if("-snpsToExcludeFilename".equalsIgnoreCase(args[i])) {
				i++;
				finder.snpsToExcludeFilename = args[i];
			} else if("-ccdsFilename".equalsIgnoreCase(args[i])) {
				i++;
				finder.ccdsFilename = args[i];
			} else if("-cdnaFilename".equalsIgnoreCase(args[i])) {
				i++;
				finder.cdnaFilename = args[i];
			} else if("-motifs".equalsIgnoreCase(args[i])) {
				i++;
				motifs = args[i];
//sahar added options for running ANN algorithm and running EpitopeFinder to collect SNV classification statitics only
			} else if ("-algorithm".equalsIgnoreCase(args[i])){
				i++;
				algorithm = args[i];
                        } else if ("-snvStats".equalsIgnoreCase(args[i])){
//                                i++;
                                finder.noprediction = true;
                        } else if ("-minScore".equalsIgnoreCase(args[i])) {
				i++;
				finder.minScore = Integer.parseInt(args[i]);
			} else if("-readingFrame".equalsIgnoreCase(args[i])) {
				i++;
				finder.readingFrame = Integer.parseInt(args[i]);
			} else if("-processHomozygous".equalsIgnoreCase(args[i])) {
				finder.processHomozygous = true;
			} else if("-processTriallelic".equalsIgnoreCase(args[i])) {
				finder.processTriallelic = true;
			} else {
				System.err.println("Unrecognized option: "+args[i]);
				printUsage();
				return;
			}
			i++;
		}
		if(i>args.length-2) {
			printUsage();
			return;
		}
		//Load epitope checker
		ResourceBundle appBundle = ResourceBundle.getBundle("epitopeFinder");
		finder.checker = (ProteinEpitopeChecker)Class.forName(appBundle.getString("epitope.checker")).newInstance();
		if(motifs!=null) {
			finder.checker.setMotifs(motifs);
		}
                if (algorithm!=null) {
                    finder.checker.setAlgorithm(algorithm);
                }
		finder.inputType=args[i];
		i++;
		if(INPUT_TYPE_PROTEIN.equalsIgnoreCase(finder.inputType)) {
			List<EpitopeHit> hits = finder.getHits(args[i]);
			System.out.println("Number of hits: "+hits.size());
			finder.printHits(hits, System.out);
			return;
		} else if (INPUT_TYPE_FASTQ.equalsIgnoreCase(finder.inputType)) {
			finder.processFastq(args[i]);
			return;
		} else if (INPUT_TYPE_CALLED_SNPS.equalsIgnoreCase(finder.inputType)) {
			finder.snpsFilename=args[i];
		} else if (INPUT_TYPE_UNSORTED_READS.equalsIgnoreCase(finder.inputType)) {
			finder.samFilename=args[i];
		}
		if(finder.validateParameters()) finder.runProcess();
		long time2 = System.currentTimeMillis();
		System.out.println("Process finished. Total time: "+((time2-time1)/1000));
	}

	



	private static void printUsage() {
		System.err.println("TODO: Usage");
	}

	public ProteinEpitopeChecker getChecker() {
		return checker;
	}

	public void setChecker(ProteinEpitopeChecker checker) {
		this.checker = checker;
	}
	
}

class EpitopeHitComparator implements Comparator<EpitopeHit> {

	@Override
	public int compare(EpitopeHit hit1, EpitopeHit hit2) {
		if(hit1.getPosition() != hit2.getPosition()) {
			return hit1.getPosition() - hit2.getPosition();
		}
		if(hit1.getEpitope().length() != hit2.getEpitope().length()) {
			return hit1.getEpitope().length() - hit2.getEpitope().length();
		}
		return hit1.getMotif().compareTo(hit2.getMotif());
	}
	
}
