package edu.uconn.bioinformatics.mapping;

import java.util.ArrayList;
import java.util.List;

import edu.uconn.bioinformatics.haplotyper.Block;
import edu.uconn.bioinformatics.haplotyper.Fragment;
import edu.uconn.ngstools.genome.GenomicRegion;
import edu.uconn.ngstools.readalignments.ReadAlignment;
import edu.uconn.ngstools.variants.CalledSNV;

public class PhasedSequence {
	private String sequenceName;
	private List<CalledSNV> snvs;
	private List<PhasedBlock> haplotypeBlocks = new ArrayList<PhasedBlock>();
	public PhasedSequence(String sequenceName, List<CalledSNV> snvs) {
		this.setSequenceName(sequenceName);
		this.setSnvs(snvs);
	}
	public String getSequenceName() {
		return sequenceName;
	}
	public void setSequenceName(String sequenceName) {
		this.sequenceName = sequenceName;
	}
	public List<CalledSNV> getSnvs() {
		return snvs;
	}
	public void setSnvs(List<CalledSNV> snvs) {
		this.snvs = snvs;
	}
	public List<PhasedBlock> getHaplotypeBlocks() {
		return haplotypeBlocks;
	}
	public void addBlock (List<CalledSNV> blockSNVs, String haplotype) {
		haplotypeBlocks.add(new PhasedBlock(blockSNVs,haplotype));
	}
	public void setHaplotypeBlocks(List<Block> haplotypeBlocks) {
		for(Block b:haplotypeBlocks) {
			String extHaplotype = b.getHaplotype();
			List<CalledSNV> blockSNVs = new ArrayList<CalledSNV>();
			StringBuilder haplotype = new StringBuilder();
			for(int i=0;i<extHaplotype.length();i++) {
				char allele = extHaplotype.charAt(i); 
				if(allele!=Fragment.NODATACHAR) {
					blockSNVs.add(snvs.get(b.getFirstPos()+i));
					haplotype.append(allele);
				}
			}
			addBlock(blockSNVs, haplotype.toString());
		}
	}
	public int findSNVPos(int snvPos, int indexFrom) {
		for(int i=indexFrom;i<snvs.size();i++) {
			CalledSNV snv = snvs.get(i);
			if(snv.getPosition() == snvPos) {
				return i;
			} else if (snv.getPosition()> snvPos) {
				return -1;
			}
		}
		return -1;
	}
	
	public Fragment buildFragment(ReadAlignment aln) {
		int firstPos=-1;
		int lastPos=-1;
		int count = 0;
		StringBuilder extendedCalls = new StringBuilder(); 
		int snpPos = binarySearchGenomicRegion(snvs, aln.getStart());
		for(int i=snpPos;i<snvs.size()&& snvs.get(i).getPosition() <= aln.getEnd();i++) {
			CalledSNV snv = snvs.get(i);
			int readPos = aln.getReadPosition(snv.getPosition());
			if(readPos >=0) {
				char base = Character.toUpperCase(aln.getReadBase(snv.getPosition()));
				char allele = Fragment.NODATACHAR;
				if(snv.isTriallelic()) {
					if(base == snv.getGenotype().charAt(0)) {
						allele = Fragment.ALLELE1CHAR;
					} else if(base == snv.getGenotype().charAt(1)) {
						allele = Fragment.ALLELE2CHAR;
					}
				} else {
					if(base == snv.getRefBase()) {
						allele = Fragment.ALLELE1CHAR;
					} else if(base == snv.getAltBase()) {
						allele = Fragment.ALLELE2CHAR;
					}
				}
				
				if(allele != Fragment.NODATACHAR) {
					if(firstPos==-1) {
						firstPos = i;
						lastPos = i;
					}
					for(int j = lastPos+1;j<i;j++) {
						extendedCalls.append(Fragment.NODATACHAR);
					}
					extendedCalls.append(allele);
					count++;
					lastPos = i;
				}
			}
		}
		if(count>1) {
			return new Fragment(aln.getRead().getId(),firstPos,extendedCalls.toString());
		}
		return null;
	}
	public List<CalledSNV> findSNVsInPhase(CalledSNV searchSnv) {
		PhasedBlock block = findBlock(searchSnv);
		if(block!=null) {
			return block.findSNVsInPhase(searchSnv);
		}
		return null;
	}
	public PhasedBlock findBlock(CalledSNV searchSnv) {
		int index = binarySearchGenomicRegion(haplotypeBlocks, searchSnv.getPosition());
		for(int i=index;i<haplotypeBlocks.size();i++) {
			PhasedBlock b = haplotypeBlocks.get(i);
			if(b.getStart()>searchSnv.getPosition()) {
				return null;
			} else if (b.getEnd()>=searchSnv.getPosition()) {
				if(b.containsSNV(searchSnv)) {
					return b;
				}
			}
		}
		return null;
	}
	public int binarySearchGenomicRegion (List<?> regions, int search) {
		int a = 0;
		int b = regions.size();
		while(a<b-1) {
			int m = (a+b)/2;
			GenomicRegion reg = (GenomicRegion) regions.get(m);
			int pos = reg.getStart(); 
			if(pos == search) {
				return m;
			} else if (pos < search) {
				a = m;
			} else {
				b = m;
			}
		}
		return a;
		
	}
}
