package edu.uconn.bioinformatics.programs;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import edu.uconn.ngstools.genome.GenomicRegionComparator;
import edu.uconn.ngstools.readalignments.ReadAlignment;
import edu.uconn.ngstools.readalignments.filehandlers.SAMReadAlignmentsHandler;
import edu.uconn.ngstools.transcriptome.CCDSTranscriptome;
import edu.uconn.ngstools.transcriptome.Exon;


public class CCDSFilter {
	public static void main(String [] args) throws Exception {
		String snpsFile = args[0];
		String ccdsMap = args[1];
		String outFile = args[2];
		
		CCDSTranscriptome transcriptome = new CCDSTranscriptome(ccdsMap);
		//Getting exons coverage if a sam file is provided
		Map<Exon, CoverageCount> coverageExons=null;
		int numberOfReads = 0;
		if(args.length>3) {
			
			coverageExons = new TreeMap<Exon, CoverageCount>(new GenomicRegionComparator(transcriptome.getSequenceNames()));
			int [] statistics = getCoverageExonsReport(args[3], transcriptome, coverageExons );
			System.out.println("Statistics");
			for(int i=0;i<statistics.length;i++) {
				if(i>0) {
					System.out.print("\t");
				}
				System.out.print(statistics[i]);
				
			}
			System.out.println();
			numberOfReads = statistics[2];
		}
		double nMegabases = (double)numberOfReads / 1000000.0;
		FileInputStream fis = new FileInputStream(snpsFile);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		PrintStream out = new PrintStream(outFile);
		String line=in.readLine();
		while(line!=null) {
			String [] items = line.split(" |\t");
			char separator = line.charAt(items[0].length());
			String seqName = items[0];
			int position = Integer.parseInt(items[1]);
			List<Exon> posExons = transcriptome.getExons(seqName, position);
			if ( posExons.size() > 0) {
				out.print(line);
				out.print(separator);
				Exon e = posExons.get(0);
				out.print(e.getStart());
				out.print(separator);
				out.print(e.getEnd());
				
				if(coverageExons!=null) {
					CoverageCount count = coverageExons.get(e);
					double nKBasesExon = (double)e.length() / 1000.0;
					double rpkm = (double)count.getReads()/(nKBasesExon*nMegabases);
					out.print(separator);
					out.print(rpkm);
					out.print(separator);
					double baseCoverage = (double)count.getBases()/e.length();
					out.print(baseCoverage);
					out.print(separator);
					out.print(count.getReads());
					out.print(separator);
					out.print(count.getBases());
				}
				out.println();
			}
			line=in.readLine();
		}
		out.flush();
		out.close();
		if(args.length>4) {
			String separator = " ";
			out = new PrintStream(args[4]);
			for(Exon e:coverageExons.keySet()) {
				out.print(e.getSequenceName());
				out.print(separator);
				out.print(e.getStart());
				out.print(separator);
				out.print(e.getEnd());
				out.print(separator);
				CoverageCount coverage = coverageExons.get(e);
				out.print(coverage.getReads());
				out.print(separator);
				double nKBasesExon = (double)e.length() / 1000.0;
				double rpkm = (double)coverage.getReads()/(nKBasesExon*nMegabases);
				out.print(rpkm);
				out.println();
			}
			out.flush();
			out.close();
		}
		fis.close();
	}
	public static int [] getCoverageExonsReport(String filename, CCDSTranscriptome transcriptome, Map<Exon, CoverageCount> coverageExons) throws IOException {
		SAMReadAlignmentsHandler alnsHandler = new SAMReadAlignmentsHandler();
		int [] statistics = new int[4];
		for(int i=0;i<statistics.length;i++) {
			statistics[i]=0;
		}
		for(Exon e: transcriptome.getAllExons()) {
			coverageExons.put(e, new CoverageCount());
		}
		FileInputStream fis = new FileInputStream(filename);
		BufferedReader in = new BufferedReader(new InputStreamReader(fis));
		String line=in.readLine();
		for(int i=0;line!=null;i++) {
			ReadAlignment aln = alnsHandler.createReadAlignment(line);
			if(aln!=null) {
				fillReadCoverageInfo(aln, transcriptome, coverageExons, statistics);
			}
			line=in.readLine();
		}
		fis.close();
		return statistics;
	}
	private static void fillReadCoverageInfo (ReadAlignment aln, CCDSTranscriptome transcriptome, Map<Exon, CoverageCount> coverageExons, int [] statistics) {
		statistics[0]++;
		int bases = aln.getRead().length();
		statistics[1]+=bases;
		Collection<Exon> exons = transcriptome.getExons(aln.getSequenceName(), aln.getStart(), aln.getEnd());
		List<Exon> filteredExons = filterExons(exons, aln);
		if (exons.size()>0) {
			statistics[2]++;
			statistics[3]+=bases;
		}
		for(Exon e: filteredExons) {
			CoverageCount count = coverageExons.get(e);
			int overlapping = e.length();
			overlapping = Math.min(overlapping, aln.length());
			if(e.getEnd()>=aln.getStart()) {
				overlapping = Math.min(overlapping, e.getEnd()-aln.getStart()+1);
			}
			if(aln.getEnd()>=e.getStart()) {
				overlapping = Math.min(overlapping, aln.getEnd()-e.getStart()+1);
			}
			count.addRead(overlapping);
		}
	}
	private static List<Exon> filterExons(Collection<Exon> exons, ReadAlignment aln) {
		List<Exon> answer = new ArrayList<Exon>();
		for(Exon e:exons) {
			if(e.getStart() <=aln.getStart() || e.getEnd()>= aln.getEnd() ) {
				answer.add(e);
			} else if(aln.getReadPosition(e.getStart())>=0) {
				answer.add(e);
			} else if (aln.getReadPosition(e.getEnd())>=0) {
				answer.add(e);
			} else if (aln.getReadPosition((e.getEnd()-e.getStart())/2)>=0) {
				answer.add(e);
			}
		}
		return answer;
	}
}
class CoverageCount {
	private int reads=0;
	private int bases=0;
	/**
	 * @return the reads
	 */
	public int getReads() {
		return reads;
	}
	/**
	 * @return the bases
	 */
	public int getBases() {
		return bases;
	}
	public void addRead(int overlapping) {
		reads++;
		bases +=overlapping;
	}
}