package edu.uconn.bioinformatics.programs;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import edu.uconn.ngstools.sequences.DNAMaskedSequence;
import edu.uconn.ngstools.sequences.Sequence;

/**
 * Greedy Solution to the Multiple Sequence Alignment problem
 * Calculates the best sum of pairs for a star arrangement of the sequences to align
 * The total sum of pairs score of the solution given by this algorithm is guaranteed 
 * equal or less than two times the optimum solution total sum of pairs  
 * @author Jorge Duitama
 */
public class MultipleSequenceAlignment {
	//Sequences to align
	private List<Sequence> sequences;
	//List to store the answer
	private List<StringBuilder> alignedSequences;
	//Matrix used to store pairwise alignment calculations 
	private AlignmentCell alignmentMatrix [][];
	//Index in sequences of the best center of a star arrangement
	private int bestCenter;
	//Sum of pairs of the of the best center star arrangement
	private int bestSumPairs;
	
	/**
	 * Loads data from standard input and calls the algorithm
	 * @throws Exception
	 */
	public MultipleSequenceAlignment() throws Exception {
		loadData();
		alignSequences();
	}
	/**
	 * Loads data from the given fasta file and calls the algorithm
	 * @throws Exception
	 */
	public MultipleSequenceAlignment(String filename) throws Exception {
		loadDataFasta(filename);
		alignSequences();
	}
	/**
	 * Performs the algorithm steps
	 * @throws Exception
	 */
	public void alignSequences() {
		calculateBestSumPairs();
		System.out.println("best star cost: "+bestSumPairs);
		calculateAlignmentBestStar();
		int totalCost = calculateAlignmentCost();
		System.out.println("total cost: "+totalCost);
		printAlignment();
	}
	/**
	 * Prints the alignment stored in the list alignedSequences 
	 */
	public void printAlignment() {
		for(StringBuilder aligSeq : alignedSequences) {
			System.out.println(aligSeq.toString());
		}
		//Calculate consensus and 0-9 agreement 
		System.out.println("Consensus");
		StringBuilder sb = alignedSequences.get(0);
		StringBuilder consensus = new StringBuilder();
		StringBuilder disagreement = new StringBuilder();
		for(int j=0;j<sb.length();j++) {
			int hitsA = 0;
			int hitsC = 0;
			int hitsG = 0;
			int hitsT = 0;
			int hitsS = 0;
			for(StringBuilder aligSeq : alignedSequences) {
				char next = aligSeq.charAt(j);
				if(next == 'A' || next == 'a') {
					hitsA++;
				} else if(next == 'C' || next == 'c') {
					hitsC++;
				} else if(next == 'G' || next == 'g') {
					hitsG++;
				} else if(next == 'T' || next == 't') {
					hitsT++;
				} else if(next == '-') {
					hitsS++;
				}
			}
			int maxHits = hitsA;
			char charMax = 'A';
			if(hitsC > maxHits) {
				maxHits = hitsC;
				charMax = 'C';
			}
			if(hitsG > maxHits) {
				maxHits = hitsG;
				charMax = 'G';
			}
			if(hitsT > maxHits) {
				maxHits = hitsT;
				charMax = 'T';
			}
			if(hitsS > maxHits) {
				maxHits = hitsS;
				charMax = '-';
			}
			consensus.append(charMax);
			if(alignedSequences.size()!=maxHits) {
				disagreement.append(""+(1+10*(alignedSequences.size()-maxHits)/alignedSequences.size()));	
			} else {
				disagreement.append("0");
			}
			
		}
		System.out.println(consensus);
		System.out.println(disagreement);
	}
	/**
	 * Debug function. Prints the alignment Matrix
	 * @param l1 Last row to print
	 * @param l2 Last column to print
	 */
	public void printMatrix(int l1, int l2) {
		for(int i=0;i<=l1;i++) {
			for(int j=0;j<=l2;j++) {
				System.out.print(alignmentMatrix[i][j].getScore()+" ");
				System.out.print(alignmentMatrix[i][j].getCs1()+"     ");
			}
			System.out.println();
		}
	}
	/**
	 * Loads from standard input the sequences and store them in the list sequences
	 * Allocates the alignment matrix according to the largest sequence
	 * @throws Exception
	 */
	public void loadData() throws Exception{
		//Open buffer
		BufferedReader in = new BufferedReader(new InputStreamReader (System.in));
		//Read number of sequences
		String strNumber = in.readLine();
		int nSeq = Integer.parseInt(strNumber);
		//Read sequences
		sequences = new ArrayList<Sequence>();
		int maxLengthSeq = 0;
		for(int i=0; i<nSeq;i++) {
			String line=in.readLine();
			sequences.add(new DNAMaskedSequence(line));
			//Calculate largest sequence length
			if(maxLengthSeq < line.length()) {
				maxLengthSeq = line.length();
			}
		}
		//Allocate memory for the alignment matrix
		int alignmentMax = (int) (1.2*maxLengthSeq+1);
		alignmentMatrix = new AlignmentCell [alignmentMax][alignmentMax];
		for(int i=0; i<alignmentMax;i++) {
			for(int j=0; j<alignmentMax;j++) {
				alignmentMatrix[i][j] = new AlignmentCell();
			}
		}
	}
	public void loadDataFasta (String filename) throws Exception{
		//Open buffer
		BufferedReader in = new BufferedReader(new InputStreamReader (new FileInputStream(filename)));
		//Read sequences
		sequences = new ArrayList<Sequence>();
		String line=in.readLine();
		StringBuilder nextSequence = new StringBuilder();
		int maxLengthSeq = 0;
		double avgSeqLength=0;
		for(int j=0;line!=null;j++) {
			if(line.startsWith(">")) {
				if(nextSequence.length()>0) {
					sequences.add(new DNAMaskedSequence(nextSequence.toString()));
					//Calculate average and largest sequence length
					avgSeqLength+=nextSequence.length();
					if(maxLengthSeq < nextSequence.length()) {
						maxLengthSeq = nextSequence.length();
					}
					nextSequence = new StringBuilder();
				}
				line = in.readLine();
			}
			//line = fixLine(line);
			nextSequence.append(line);
			line = in.readLine();
		}
		if(nextSequence.length()>0) {
			sequences.add(new DNAMaskedSequence(nextSequence.toString()));
			//Calculate average and largest sequence length
			avgSeqLength+=nextSequence.length();
			if(maxLengthSeq < nextSequence.length()) {
				maxLengthSeq = nextSequence.length();
			}
		}
		//Allocate memory for the alignment matrix
		int alignmentMax = (int) (1.2*maxLengthSeq+1);
		alignmentMatrix = new AlignmentCell [alignmentMax][alignmentMax];
		for(int i=0; i<alignmentMax;i++) {
			for(int j=0; j<alignmentMax;j++) {
				alignmentMatrix[i][j] = new AlignmentCell();
			}
		}
		System.out.println("Sequences loaded: "+sequences.size());
		if(sequences.size()>0) {
			avgSeqLength/=sequences.size();
			System.out.println("Average Sequence Length: "+avgSeqLength);
		}
		in.close();
	}
	/**
	 * Score function over two characters.
	 * @param c1
	 * @param c2
	 * @return int score(c1,c2) in this case 0 if the characters match and 1 otherwise
	 */
	public int getScore (char c1, char c2) {
		char upC1 = Character.toUpperCase(c1);
		char upC2 = Character.toUpperCase(c2);
		if (upC1!= upC2) {
			return 1;
		}
		return 0;
	}
	/**
	 * Performs a step in the dynamic programming solution for pairwise alignment
	 * Fills the matrix information for the cell <pos1,pos2>
	 * PRE: pos1 > 0 || pos2 > 0
	 * @param cs1 Character that appears in the first sequence in the position pos1-1. 
	 * 0 if pos1 = 0
	 * @param cs2 Character that appears in the second sequence in the position pos2-1
	 * 0 if pos2 = 0
	 * @param pos1 Row of alignmentMatrix to fill
	 * @param pos2 Column of alignmentMatrix to fill
	 */
	public void calculateScore(char cs1, char cs2, int pos1, int pos2) {
		int score = Integer.MIN_VALUE;
		if (pos1 > 0) {
			//Fill information if the best score is in the upper cell
			score = alignmentMatrix [pos1-1][pos2].getScore() + getScore (cs1,'-');
			alignmentMatrix [pos1][pos2].setSource('U');
			alignmentMatrix [pos1][pos2].setCs1(cs1);
			alignmentMatrix [pos1][pos2].setCs2('-');
			if (pos2 > 0 ) {
				int score2 = alignmentMatrix [pos1-1][pos2-1].getScore() + getScore (cs1,cs2);
				if (score >= score2) {
					//Fill information if the best score is in the diagonal. Preferred in case of ties
					score = score2;
					alignmentMatrix [pos1][pos2].setSource('D');
					alignmentMatrix [pos1][pos2].setCs1(cs1);
					alignmentMatrix [pos1][pos2].setCs2(cs2);
				}
			}
		} 
		if (pos2 > 0 ) {
			int score2 = alignmentMatrix [pos1][pos2-1].getScore() + getScore ('-',cs2);
			if (score == Integer.MIN_VALUE || score > score2) {
				//Fill information if the best score is in the left cell
				score = score2;
				alignmentMatrix [pos1][pos2].setSource('L');
				alignmentMatrix [pos1][pos2].setCs1('-');
				alignmentMatrix [pos1][pos2].setCs2(cs2);
			}
		}
		alignmentMatrix [pos1][pos2].setScore(score);
	}
	/**
	 * Performs the dynamic programming approach to align s1 and s2
	 * Stores the results in alignmentMatrix 
	 * @param s1 First string to align
	 * @param s2 Second string to align
	 */
	public void fillAlignmentMatrix (CharSequence s1, CharSequence s2) {
		//Fills the first position
		alignmentMatrix[0][0].setScore(0);
		//Fills the first column
		for(int i=0;i<s1.length();i++) {
			calculateScore(s1.charAt(i), '0', i+1, 0);
		}
		//Fills the first row
		for(int j=0;j<s2.length();j++) {
			calculateScore('0', s2.charAt(j),0, j+1);
		}
		//Fills the rest of the matrix row by row
		for(int i=0;i<s1.length();i++) {
			for(int j=0;j<s2.length();j++) {
				calculateScore(s1.charAt(i), s2.charAt(j), i+1, j+1);
			}
		}
	}
	/**
	 * Gets an optimum path in alignmentMatrix to the position <posS1,posS2>
	 * PRE: alignmentMatrix has the optimum alignment info for two sequences
	 * @param posS1 Row of the cell to calculate the optimum path
	 * @param posS2 Column of the cell to calculate the optimum path
	 * @param sequence 1 to get the first sequence characters. 
	 * 2 to get the second sequence characters
	 * @return String Aligned sequence according to the info stored in alignmentMatrix
	 */
	public String getPath (int posS1, int posS2, int sequence) {
		StringBuffer answer = new StringBuffer();
		//Iterates from the position to the cell in <0,0>
		int i= posS1;
		int j = posS2;
		while ( i>0 || j > 0 ) {
			//Stores the actual cell info
			if(sequence == 1) {
				answer.append(alignmentMatrix[i][j].getCs1());
			} else {
				answer.append(alignmentMatrix[i][j].getCs2());
			}
			//Performs the step to the next cell following optimum decisions
			if (alignmentMatrix[i][j].getSource() == 'L') {
				j--;
			} else if (alignmentMatrix[i][j].getSource() == 'U') {
				i--;
			} else {
				i--;
				j--;
			}
		}
		//Reverse the string because it was calculated from the last to the first character
		return answer.reverse().toString();
	}
	/**
	 * Calculates the sum of optimum scores of pairwise alignments between the sequence s in
	 * the position seqId and the other sequences.
	 * This score is the same as the sum of pairs of the star arrangement which center is the
	 * sequence in the position seqId
	 * @param seqId Position in sequences of the star center to calculate
	 * @return int Sum of pairwise alignment scores of the star arrangement
	 */
	public int getSumPairs(int seqId) {
		Sequence seqBase = sequences.get(seqId);
		int answer = 0;
		//Iterates over the sequences
		for(int i=0;i<sequences.size();i++) {
			if(i!=seqId) {
				//Calculates the pairwise alignment between i and seqId and 
				//adds the optimum score to the answer 
				Sequence seq = sequences.get(i);
				fillAlignmentMatrix(seqBase, seq);
				answer+=alignmentMatrix[seqBase.length()][seq.length()].getScore();
			}
		}
		return answer;
	}
	/**
	 * Calculates the best star arrangement center and best sum of pairs.
	 * Stores the answers in the corresponding attributes 
	 */
	public void calculateBestSumPairs() {
		//Calls for each sequence getSumPairs and calculates the minimum 
		for(int i=0;i<sequences.size();i++) {
			int next = getSumPairs(i);
			if(i==0 || bestSumPairs > next) {
				bestCenter = i;
				bestSumPairs = next;
			}
		}
	}
	/**
	 * Calculates the multiple sequence alignment aligning one sequence at a time
	 * with the best center. Stores the results in alignedSequences
	 */
	public void calculateAlignmentBestStar () {
		//Allocates aligned sequences
		alignedSequences = new ArrayList<StringBuilder>();
		//Variable with the aligned center.
		String centerSequence = sequences.get(bestCenter).toString();
		//Iterates over sequences adding one at a time
		for(int i=0;i<sequences.size();i++) {
			if (i!=bestCenter) {
				//Pairwise alignment between i-th sequence and the aligned center
				Sequence next = sequences.get(i);
				fillAlignmentMatrix(centerSequence, next);
				String alignedCenter = getPath(centerSequence.length(), next.length(), 1);
				String nextAligned = getPath(centerSequence.length(), next.length(), 2);
				if (!centerSequence.equals(alignedCenter)) {
					//If the aligned center change, update the new spaces in the previously aligned sequences
					updateNewSpaces (centerSequence, alignedCenter);
					centerSequence = alignedCenter;
				}
				//Finally, add the new alignment to the answer
				alignedSequences.add(new StringBuilder(nextAligned));
			} else {
				//Just add the aligned center because it is already aligned with the previous sequences
				alignedSequences.add(new StringBuilder(centerSequence));
			}
		}
	}
	/**
	 * Calculates the total sum of pairs of the alignment stored in alignedSequences 
	 * @return int Sum of pairwise alignment scores among all possible pairs in alignedSequences 
	 */
	public int calculateAlignmentCost() {
		int answer = 0;
		for(int i=0;i<alignedSequences.size();i++) {
			for(int j=i+1;j<alignedSequences.size();j++) {
				String s1 = alignedSequences.get(i).toString();
				String s2 = alignedSequences.get(j).toString();
				for(int k=0;k<s1.length();k++) {
					answer+=getScore(s1.charAt(k), s2.charAt(k));	
				}
			}
		}
		return answer;
	}
	/**
	 * Update the differences between centerSequence and alignedSequence in the sequences
	 * stored in alignedSequences
	 * PRE: Every sequence in alignedSequences is aligned with centerSequence
	 * POST: Every sequence in alignedSequences is aligned with alignedCenter
	 * @param centerSequence Center before a pairwise alignment
	 * @param alignedCenter Center after a pairwise alignment 
	 */
	public void updateNewSpaces(String centerSequence, String alignedCenter) {
		int i=0;
		int j;
		int spaces = 0;
		for(j = 0;i<centerSequence.length() && j<alignedCenter.length();j++ ) {
			//If there is a difference, it must be a new space
			if (centerSequence.charAt(i)!= alignedCenter.charAt(j)) {
				//Add the new space to all sequences
				for(StringBuilder aligSeq : alignedSequences) {
					aligSeq.insert(i+spaces, '-');
				}
				spaces++;
			} else {
				i++;
			}
		}
		//If centerSequence is over, there must be spaces at the end of alignedCenter  
		for(;j<alignedCenter.length();j++) {
			for(StringBuilder aligSeq : alignedSequences) {
				aligSeq.insert(i+spaces, '-');
			}
			spaces++;
		}
	}
	/**
	 * Main function. Just creates a new MultipleSequenceAlignment object
	 * @param args Fasta file with sequences to align
	 * @throws Exception
	 */
	public static void main(String [] args) throws Exception{
		if(args.length>0) {
			new MultipleSequenceAlignment(args[0]);
		} else {
			new MultipleSequenceAlignment("C:\\Documents and Settings\\Jorge\\My Documents\\uconn\\researchPrimerHunter\\inputFilesHV2\\H16.fas");
		}
		
	}
}
/**
 * Inner class to store the alignment matrix data
 * @author Jorge Duitama
 */
class AlignmentCell {
	//Optimum score of the cell
	private int score;
	//Where the optimim comes from. Can be L for left, U for up or D for diagonal
	private char source;
	//Character of the first sequence in the optimum path to this cell
	private char cs1;
	//Character of the second sequence in the optimum path to this cell
	private char cs2;
	
	public int getScore() {
		return score;
	}
	public char getCs1() {
		return cs1;
	}
	public char getCs2() {
		return cs2;
	}
	public char getSource() {
		return source;
	}
	public void setScore(int score) {
		this.score = score;
	}
	public void setSource(char source) {
		this.source = source;
	}
	public void setCs1(char cs1) {
		this.cs1 = cs1;
	}
	public void setCs2(char cs2) {
		this.cs2 = cs2;
	}
}