package edu.uconn.bioinformatics.programs;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;

import edu.uconn.ngstools.genome.GenomicRegionComparator;
import edu.uconn.ngstools.main.SNVSFinder;
import edu.uconn.ngstools.sequences.DNASequence;
import edu.uconn.ngstools.sequences.DefaultSequenceNames;
import edu.uconn.ngstools.sequences.DegenerateSequence;
import edu.uconn.ngstools.variants.CalledSNV;
import edu.uconn.ngstools.variants.PileupRecord;
import edu.uconn.ngstools.variants.filehandlers.CalledSNVSFileHandler;
import edu.uconn.ngstools.variants.filehandlers.PileupFileHandler;

public class PileupsMerger {
	public static PileupSNPRecord loadRecord (BufferedReader in, PileupFileHandler handler, SNVSFinder snpsFinder) throws IOException {
		PileupRecord record = handler.loadPileupRecord(in);
		if(record == null) return null;
		CalledSNV snp = (CalledSNV) snpsFinder.callSNV(record);
		return new PileupSNPRecord(record, snp);
		
	}
	public static PileupRecord mergeRecords(List<PileupSNPRecord> sourceRecords) {
		StringBuilder callsB = new StringBuilder();
		StringBuilder scoresB = new StringBuilder();
		short maxQualScore = 0;
		String seqName = null;
		int position = 0;
		char refBase = 'N';
		for(PileupSNPRecord rs:sourceRecords) {
			PileupRecord r = rs.getRecord();
			if(seqName ==null) {
				seqName = r.getSequenceName();
				position = r.getPosition();
				refBase = r.getRefBase();
			}
			callsB.append(r.getAlleleCalls());
			scoresB.append(r.getQualityScores());
			if(r.getMaxQualScore()>maxQualScore) {
				maxQualScore = r.getMaxQualScore();
			}
		}
		PileupRecord mergedRecord = new PileupRecord(seqName, position, refBase, callsB.toString(), scoresB.toString());
		mergedRecord.setMaxQualScore(maxQualScore);
		return mergedRecord;
	}
	private static boolean testAllele(List<PileupSNPRecord> sourceRecords, int minNumFiles, int index) {
		if(minNumFiles <=1) {
			return true;
		}
		int numFiles = 0;
		for(PileupSNPRecord sr:sourceRecords) {
			if(sr.getSnp()!=null && sr.getSnp().getAllCounts()[index]>0) {
				numFiles++;
			}
		}
		return numFiles>=minNumFiles;
	}
	public static void main(String[] args) throws IOException {
		GenomicRegionComparator cmp = new GenomicRegionComparator(DefaultSequenceNames.DEFAULT_SEQUENCE_NAMES);
		PileupFileHandler handler = new PileupFileHandler();
		CalledSNVSFileHandler snpsHandler = new CalledSNVSFileHandler();
		SNVSFinder snpsFinder = new SNVSFinder();
		DNASequence dummySequence = new DNASequence();
		snpsFinder.setAllVariantCalls(true);
		int minNumFiles = 1;
		boolean printPileup = false;
		boolean printLowerCaseRef = false;
		int k=0;
		for(;k<args.length && args[k].charAt(0)=='-';k++) {
			if("-h".equalsIgnoreCase(args[k])) {
				k++;
				snpsFinder.setHeterozygosityRate(Double.parseDouble(args[k]));
			} else if ("-m".equalsIgnoreCase(args[k])) {
				k++;
				minNumFiles = Integer.parseInt(args[k]);
			} else if ("-p".equalsIgnoreCase(args[k])) {
				printPileup = true;
			} else if("-keepLowerCaseRef".equalsIgnoreCase(args[k])) {
				printLowerCaseRef = true;
			} else {
				System.err.println("Unrecognized option: "+args[k]);
				return;
			}
		}
		BufferedReader [] streams = new BufferedReader [args.length-k];
		PileupSNPRecord [] records = new PileupSNPRecord[args.length-k];
		//Open buffers
		int j=0;
		for(;k<args.length;k++) {
			FileInputStream fis = new FileInputStream(args[k]);
			BufferedReader in = new BufferedReader(new InputStreamReader(fis));
			streams[j] = in;
			records[j] = loadRecord(in, handler, snpsFinder);
			j++;
		}
		//Merge files
		while(true) {
			List<PileupSNPRecord> recordsToMerge = new ArrayList<PileupSNPRecord>();
			int numWithVariants = 0;
			//Find next set of records to merge
			for(int i=0;i<records.length;i++) {
				if(records[i]!=null) {
					if(recordsToMerge.size() == 0) {
						recordsToMerge.add(records[i]);
						if(records[i].getSnp()!=null) numWithVariants++;
					} else {
						int resCmp = cmp.compare(recordsToMerge.get(0).getRecord(), records[i].getRecord());
						if(resCmp > 0) {
							recordsToMerge.clear();
							numWithVariants = 0;
							recordsToMerge.add(records[i]);
							if(records[i].getSnp()!=null) numWithVariants++;
						} else if (resCmp == 0) {
							recordsToMerge.add(records[i]);
							if(records[i].getSnp()!=null) numWithVariants++;
						}
					}
				}
			}
			if(recordsToMerge.size() == 0) {
				break;
			}
			//Merge records
			if(numWithVariants >0 && recordsToMerge.size()>=minNumFiles) {
				PileupRecord mergedRecord = mergeRecords(recordsToMerge);
				CalledSNV mergedSNP = snpsFinder.callSNV(mergedRecord);
				if(mergedSNP==null) System.err.println("Inconsistency in "+mergedRecord.getSequenceName()+":"+mergedRecord.getPosition()+". Num lanes: "+recordsToMerge.size()+". Num with varaints: "+numWithVariants);
				String genotype = mergedSNP.getGenotype();
				char upRef = Character.toUpperCase(mergedSNP.getRefBase());
				boolean accepted = !mergedSNP.isHomozygousReference();
				accepted = accepted && (printLowerCaseRef || Character.isUpperCase(mergedSNP.getRefBase()));
				if(accepted && genotype.charAt(0)!=upRef) {
					accepted = testAllele(recordsToMerge,minNumFiles, dummySequence.getAlphabetIndex(genotype.charAt(0)));
				}
				if(accepted && mergedSNP.isHeterozygous() && genotype.charAt(1)!=upRef) {
					accepted = testAllele(recordsToMerge,minNumFiles, dummySequence.getAlphabetIndex(genotype.charAt(1)));
				}
				mergedRecord.setGenotype(DegenerateSequence.getDegenerateBase(genotype.charAt(0), genotype.charAt(1)));
				mergedRecord.setGenotypeProbability(mergedSNP.getProbability());
				mergedRecord.setRefProbability(mergedSNP.getRefProbability()); 
				if(accepted) {
					if(printPileup) {
						handler.printRecord(mergedRecord,System.out);
					} else {
						snpsHandler.saveSNV(mergedSNP, System.out, false);
					}
					
				}
				System.out.flush();
			}
			//Step forward
			j=0;
			for(int i=0;i<records.length&&j<recordsToMerge.size();i++) {
				if(records[i]!=null) {
					PileupSNPRecord mergedRecord = recordsToMerge.get(j);
					if(records[i]==mergedRecord) {
						records[i] = loadRecord(streams[i], handler, snpsFinder);
						j++;
					}
				}
			}
		}
		//Close streams
		for(int i=0;i<streams.length;i++) {
			streams[i].close();
		}
	}
}
class PileupSNPRecord {
	private PileupRecord record;
	private CalledSNV snp;
	public PileupSNPRecord(PileupRecord record, CalledSNV snp) {
		super();
		this.record = record;
		this.snp = snp;
	}
	
	public PileupRecord getRecord() {
		return record;
	}

	public void setRecord(PileupRecord record) {
		this.record = record;
	}

	public CalledSNV getSnp() {
		return snp;
	}
}
